# -*- coding: utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Ildar Nikolaev"
__email__ = "nildar@users.sourceforge.net"

import os
import sys
import subprocess
import threading
import signal
import time
import datetime
import traceback

import bpy
import bl_ui

from . import BMaxwellPreferences

if sys.platform == 'darwin':
    _STUDIO = "Studio.app"
    _MAXWELL = "Maxwell.app"
    _MXNETWORK = "Mxnetwork.app"
elif sys.platform == 'win32':
    _STUDIO = "studio.exe"
    _MAXWELL = "maxwell.exe"
    _MXNETWORK = "mxnetwork.exe"
else:
    _STUDIO = "studio"
    _MAXWELL = "maxwell"
    _MXNETWORK = "mxnetwork"


def abspath(path):
    if path:
        path = bpy.path.abspath(path)
        return os.path.normpath(path)
    return None


def get_maxwell_path(program=None):
    prefs = BMaxwellPreferences.get_preferences()
    path = abspath(prefs.maxwell_path)
    if program is not None:
        path = os.path.join(path, program)
    return path


class Logger:
    def __init__(self, verbose=None, report=None):
        if verbose is None:
            verbose = BMaxwellPreferences.get_preferences().verbose_output
        self.verbose = verbose
        if report is None:
            report = bpy.ops.maxwell.report
        self.report = report

    def __call__(self, text, type=None):
        if self.verbose:
            print("[BMAXWELL: {0}] {1}".format(time.strftime('%Y-%b-%d %H:%M:%S'), text))
        if type is not None:
            self.report(type=type, text=text)


def open_app(executable, *args):
    if sys.platform == 'darwin':
        _args = ["open", "-a", executable]
        if args:
            _args.append("--args")
    else:
        _args = [executable]
    _args.extend(args)
    if subprocess.mswindows:
        # TODO: find a better method to run app
        subprocess.Popen(_args, creationflags=0x8)
    else:
        subprocess.Popen(_args)


def run_studio(mxs):
    open_app(get_maxwell_path(_STUDIO), mxs)


def run_render_script(ms):
    open_app(get_maxwell_path(_MAXWELL), "-script:'%s'" % ms)


def run_network(mxs, run_manager, run_node, manager):
    executable = get_maxwell_path(_MXNETWORK)
    if run_manager:
        open_app(executable, "-manager")
    if run_node:
        arg = "-node"
        if manager:
            arg += ":" + manager
        open_app(executable, arg)
    arg = "-monitor"
    if manager:
        arg += ":" + manager
    open_app(executable, arg, "-addjob:'%s'" % mxs)


from .maxwell import TexturePreview
from .maxwell import MxiReader
from .maxwell import MxRender

from .export import Exporter


class MaxwellRenderEngine(bpy.types.RenderEngine):
    bl_idname = "maxwell_render"
    bl_label = "Maxwell Render"
    
    bl_use_preview = True
    bl_use_texture_preview = True

    # bl_use_exclude_layers = True
    # bl_use_shading_nodes = True

    _COMPATIBLE_PANELS = (
        ("properties_render", ((
            "RENDER_PT_render",
            "RENDER_PT_dimensions",
            "RENDER_PT_output",
            "RENDER_PT_post_processing",
        ), False)),
        ("properties_world", ((
            "WORLD_PT_context_world",
            "WORLD_PT_custom_props",
        ), False)),
        ("properties_material", ((
            "MATERIAL_PT_custom_props",
        ), False)),
        ("properties_texture", None),
        #("properties_texture", (
        #    "TEXTURE_PT_context_texture",
        #    "TEXTURE_PT_preview",
        #    "TEXTURE_PT_image",
        #    #"TEXTURE_PT_image_sampling",
        #    #"TEXTURE_PT_image_mapping",
        #    "TEXTURE_PT_mapping",
        #    #"TEXTURE_PT_influence",
        #), False)),
        ("properties_render_layer", None),
        ("properties_scene", None),
        ("properties_data_camera", None),
        ("properties_data_mesh", None),
        ("properties_particle", None),
    )

    _lock = threading.Lock()
    _tex = None

    def __init__(self):
        self._log = Logger()

        # Texture preview
        self._mxtex = None
        # Render
        self._export_result = False
        # FIRE
        self._timer_update = None
        self._mxr = None

    def __del__(self):
        if hasattr(self, "_mxr"):
            if self._mxr:
                self._mxr.Stop()
            del self._mxr
        else:  # FIXME: bug in blender???
            del MaxwellRenderEngine._lock
            del MaxwellRenderEngine._tex

    def update(self, data, scene):
        try:
            if self.is_preview:
                ob_tex = scene.objects.get("texture")
                if ob_tex and ob_tex.is_visible(scene) and not ob_tex.hide_render:
                    #tex = ob_tex.material_slots[0].material.texture_slots[0].texture
                    tex = ob_tex.active_material.active_texture
                    if tex.type == 'IMAGE' and tex.image:
                        image = tex.image

                        r = self.render
                        width = r.resolution_x
                        height = r.resolution_y

                        tex_preview = MaxwellRenderEngine._tex
                        if tex_preview is None:
                            mxtex = TexturePreview(image, tex.maxwell.as_pointer())
                        else:
                            _image, mxtex, _width, _height = tex_preview
                            if image != _image:
                                mxtex = TexturePreview(image, tex.maxwell.as_pointer())
                            elif mxtex.Precompute(tex.maxwell.as_pointer()):
                                pass
                            elif width != _width or height != _height:
                                pass
                            else:
                                return

                        tex_preview = image, mxtex, width, height
                        MaxwellRenderEngine._tex = tex_preview
                        self._mxtex = mxtex
            else:
                exporter = Exporter(scene, self)
                self._export_result = exporter.export()
        except:
            self.report({'ERROR'}, traceback.format_exc())

    def render(self, scene):
        """
        Args:
            scene (bpy.types.Scene): .
        """
        with MaxwellRenderEngine._lock:
            if self.is_preview:
                mxtex = self._mxtex
                if mxtex is not None:
                    r = self.render
                    rr = self.begin_result(0, 0, r.resolution_x, r.resolution_y)
                    mxtex.UpdateResult(rr.as_pointer())
                    self.end_result(rr)
            elif self._export_result:
                run, args = self._export_result;
                if run == 'studio':
                    run_studio(*args)
                elif run == 'network':
                    run_network(*args)
                elif run == 'render_script':
                    if args[0]:
                        run_render_script(args[1])
                elif run == 'render':
                    self._render(*args)

    def _render(self, mxs, mxi, cmd_args, xres, yres, offset):
        self._log("--- start render ---")
        if cmd_args:
            self._log("args: " + " ".join(cmd_args))

        maxwell_path = get_maxwell_path()
        if sys.platform == 'darwin':
            cwd = maxwell_path
            executable = os.path.join(cwd, _MAXWELL, "Contents/MacOS/Maxwell")
        else:
            cwd = os.path.dirname(mxs)
            executable = os.path.join(maxwell_path, _MAXWELL)
        args = [executable, "-mxs:%s" % mxs] + cmd_args
        if subprocess.mswindows:
            process = subprocess.Popen(args, cwd=cwd, creationflags=subprocess.CREATE_NEW_PROCESS_GROUP)
        else:
            process = subprocess.Popen(args, cwd=cwd)

        reader = MxiReader()

        def update_render_result():
            self._log("Render update: %s" % mxi)
            try:
                reader.Read(mxi)
                rr = self.begin_result(0, 0, xres, yres)
                reader.UpdateRenderResult(rr, *offset)
                self.update_result(rr)
                self.update_stats('', "Maxwell Render: '%s' SL=%.4f RT=%s" % (
                    reader.CameraName,
                    reader.SamplingLevel,
                    datetime.timedelta(seconds=reader.RenderTime)
                ))
                self.end_result(rr)
            except Exception as e:
                self.report({'WARNING'}, str(e))

        file_time = False
        while process.poll() is None:
            if self.test_break():
                if "-nogui" in cmd_args or "-hide" in cmd_args:
                    if subprocess.mswindows:
                        process.send_signal(signal.CTRL_BREAK_EVENT)
                    else:
                        process.terminate()
                    timer = threading.Timer(5, lambda p: p.kill(), [process])
                    timer.start()
                    process.wait()
                    timer.cancel()
                break
            try:
                t = os.stat(mxi).st_mtime
                if t != file_time:
                    if file_time:
                        update_render_result()
                    file_time = t
            except IOError:
                file_time = True
            time.sleep(1)
            
        if file_time:
            update_render_result()
        self._log("--- end render ---")

    def view_update(self, context=None):
        try:
            with MaxwellRenderEngine._lock:
                if self._timer_update:
                    if self._timer_update.is_alive():
                        self._timer_update.cancel()
                    else:
                        if self._mxr:
                            self._mxr.Stop()
                        self._mxr = MxRender(
                            data=context.blend_data.as_pointer(),
                            scene=context.scene.as_pointer())
                        self.InteractiveModeEnabled = True
                        self._mxr.UpdateView(
                            region=context.region.as_pointer(),
                            v3d=context.space_data.as_pointer(),
                            rv3d=context.region_data.as_pointer())
                        self._timer_update = None
                        return
                self._timer_update = threading.Timer(0.1, self.tag_update)
                self._timer_update.start()
        except Exception as e:
            # TODO: find a better way to show error
            bpy.ops.maxwell.report(type={'WARNING'}, text=str(e))
            #bpy.ops.maxwell.report(type={'ERROR'}, text=str(e))
            #self.report({'ERROR'}, traceback.format_exc())
            #self.report({'WARNING'}, str(e))

    def view_draw(self, context=None):
        with MaxwellRenderEngine._lock:
            if self._mxr and self._timer_update is None:
                self._mxr.UpdateView(
                    region=context.region.as_pointer(),
                    v3d=context.space_data.as_pointer(),
                    rv3d=context.region_data.as_pointer(),
                    draw=True)

    @classmethod
    def _compatible(cls, mod, panels, remove=False):
        mod = getattr(bl_ui, mod)
        if panels is None:
            for c in mod.__dict__.values():
                ce = getattr(c, "COMPAT_ENGINES", None)
                if ce is not None:
                    if remove:
                        ce.remove(cls.bl_idname)
                    else:
                        ce.add(cls.bl_idname)
        else:
            classes, exclude = panels
            if exclude:
                for c in mod.__dict__.values():
                    if c.__name__ not in classes:
                        ce = getattr(c, "COMPAT_ENGINES", None)
                        if ce is not None:
                            if remove:
                                ce.remove(cls.bl_idname)
                            else:
                                ce.add(cls.bl_idname)
            else:
                for c in classes:
                    ce = getattr(mod, c).COMPAT_ENGINES
                    if remove:
                        ce.remove(cls.bl_idname)
                    else:
                        ce.add(cls.bl_idname)

    @classmethod
    def register(cls):
        for mod, panels in cls._COMPATIBLE_PANELS:
            cls._compatible(mod, panels)

    @classmethod
    def unregister(cls):
        for mod, panels in cls._COMPATIBLE_PANELS:
            cls._compatible(mod, panels, True)
