# -*- coding: utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Ildar Nikolaev"
__email__ = "nildar@users.sourceforge.net"

import os
import time
import weakref

import bpy
from mathutils import Matrix

from .. import bl_info
from .. import maxwell
from .. import BMaxwellPreferences

from ..engine import abspath, Logger

from .material import Materials
from .camera import Camera  # for proper initialization
from .object import Object


SCRIPT_HEADER = """// Animation script, created with B-Maxwell
var i = 0;
var mxsList = [
"""

SCRIPT_FOOTER = """];
if (i < mxsList.length)
{
  renderScene();
}
function renderScene()
{
  var mxsFile = mxsList[i];
  Maxwell.print("rendering Mxs file:" + mxsFile);
  Maxwell.openMxs(mxsFile);
  Maxwell.startRender();
}
function renderHasFinished()
{
  Maxwell.print("finished")
  if (++i < mxsList.length)
    renderScene();
}
RenderEvents["renderFinished()"].connect(renderHasFinished);
"""


def dupli_group_objects(dupli_group):
    for obj in dupli_group.objects:
        if obj.dupli_type == 'GROUP':
            yield from dupli_group_objects(obj.dupli_group)
        else:
            yield obj


class Exporter:
    def __init__(self, scene, engine):
        """
        scene (bpy.types.Scene): Blender scene for export.
        engine (MaxwellRenderEngine): engine.
        """
        self.is_animation = engine.is_animation
        self.scene = scene

        self.log = Logger()

        render = scene.render

        ### output paths ##################################
        # scene.render.filepath => scene.render.frame_path()
        # bpy.data.filepath = '' | bpy.data.filepath != ''
        # <None> => 0001.ext
        ######
        # a   => C:\a0001.ext
        # a/  => C:\a\0001.ext
        # a/b => C:\a\b0001.ext
        ######
        # /   => C:\0001.ext
        # /a  => C:\a0001.ext
        ## absolute from root, on windows from C: drive
        ######
        # //  => 0001.ext | <abs scene path>\0001.ext
        # //a => a0001.ext | <abs scene path>\a0001.ext
        ######
        # self.out_dir       # output directory
        # self.out_mxname    # mxs, mxi and ms names
        # self.out_name      # image name
        # self.out_ext       # image extension
        # self.out_suffix    # if an animation append to the name a frame number
        ######
        out_dir, out_name = os.path.split(render.frame_path())
        if not render.filepath:
            out_dir = os.path.dirname(bpy.data.filepath)
        elif not os.path.isabs(out_dir):
            out_dir = os.path.join(os.path.dirname(bpy.data.filepath), out_dir)
        self.out_dir = os.path.normpath(out_dir)
        self.out_name = os.path.splitext(out_name)[0]
        self.out_ext = render.file_extension
        if bpy.data.filepath:
            self.out_mxname = os.path.splitext(os.path.basename(bpy.data.filepath))[0]
        else:
            self.out_mxname = "untitled"
        self.frame_digits = max(6, len(str(scene.frame_end)))
        if self.is_animation:
            self.out_suffix = "_{0:0{1}}".format(scene.frame_current, self.frame_digits)
        else:
            self.out_suffix = ""

        ### dimensions ####################################
        res_scale = render.resolution_percentage / 100
        self.xres = int(render.resolution_x * res_scale)
        self.yres = int(render.resolution_y * res_scale)
        if render.use_border:
            self.border = (
                int(self.xres * render.border_min_x), int(self.yres * (1.0 - render.border_max_y)),
                int(self.xres * render.border_max_x), int(self.yres * (1.0 - render.border_min_y))
            )
        else:
            self.border = None

        self.aspect = 1.0  # pixel aspect ratio
        self.scale = 1.0  # global scene scale
        self.fps = 24

        self.layers = None  # list of indices for enabled render layers

        self.mxs = None
        self.materials = None  # bmaxwell.export.Materials

        self.objects = {}  # {bpy.types.Object : export.Object}
        self.shared_objects = {}  # {bpy.types.Object.data : maxwell.Object}

        # motion blur
        self.mb_steps = 0
        self.mb_step = 0.0
        self.mb_frames = 0.0

    def _render_params(self, rp, mxi_path):
        """
        _render_params(self, rp, mxi_path)

        Args:
            rp (bmaxwell.properties.MaxwellParamsProperties): Render parameters.
            mxi_path (str): Full path to MXI file.
        """
        mxs_rp = self.mxs.RenderParams
        ## render params #################################
        mxs_rp['MXI FULLNAME'] = mxi_path
        mxs_rp['ENGINE'] = rp.quality
        mxs_rp['NUM THREADS'] = rp.threads
        mxs_rp['STOP TIME'] = rp.time * 60
        mxs_rp['SAMPLING LEVEL'] = rp.sl
        # Multilight
        mxs_rp['USE MULTILIGHT'] = int(rp.multilight)
        mxs_rp['SAVE LIGHTS IN SEPARATE FILES'] = int(rp.separate)

        # Globals
        if rp.motionblur and self.scene.camera.type == 'CAMERA':
            self.mb_steps = rp.mbsteps
            mx_cam = self.scene.camera.data.maxwell
            if mx_cam.shtype == 'ANGLE':
                shutter = 360 * self.fps / mx_cam.angle
            else:  # elif mx_cam.shtype == 'SPEED':
                shutter = mx_cam.shutter
            self.mb_frames = self.fps / shutter
            self.mb_step = self.mb_frames / self.mb_steps
            # only set if we have active camera
            mxs_rp['DO MOTION BLUR'] = rp.motionblur
        mxs_rp['DO DISPLACEMENT'] = rp.displacement
        mxs_rp['DO DISPERSION'] = rp.dispersion

        # Illumination & Caustics
        illumination = int(rp.illumination)
        mxs_rp['DO DIRECT LAYER'] = bool(illumination & 1)
        mxs_rp['DO INDIRECT LAYER'] = bool(illumination & 2)
        refl_caustics = int(rp.reflcaustics)
        mxs_rp['DO DIRECT REFLECTION CAUSTIC LAYER'] = bool(refl_caustics & 1)
        mxs_rp['DO INDIRECT REFLECTION CAUSTIC LAYER'] = bool(refl_caustics & 2)
        refr_caustics = int(rp.refrcaustics)
        mxs_rp['DO DIRECT REFRACTION CAUSTIC LAYER'] = bool(refr_caustics & 1)
        mxs_rp['DO INDIRECT REFRACTION CAUSTIC LAYER'] = bool(refr_caustics & 2)

        # Materials
        self.mxs.EnableOverrideMaterial = bool(rp.ovrmat)
        self.mxs.OverrideMaterial = abspath(rp.ovrmat)
        self.mxs.EnableDefaultMaterial = bool(rp.defmat)
        self.mxs.DefaultMaterial = abspath(rp.defmat)
        for p in rp.searching_paths:
            p = abspath(p.name)
            if p:
                self.mxs.AddSearchingPath(p)

        # ToneMapping
        self.mxs.ColorSpace = int(rp.colorspace)
        self.mxs.ToneMapping = {"gamma": rp.gamma, "burn": rp.burn}
        mxs_rp['DO SHARPNESS'] = rp.doSharpness
        mxs_rp['SHARPNESS'] = rp.sharpness / 100

        # Simulens
        self.mxs.Diffraction = {
            "enabled": rp.doDiffraction,
            "intensity": rp.diffraction / 2500.0,
            "frequency": rp.frequency / 2500.0,
            "aperture": rp.aperture,
            "obstacle": rp.obstacle
        }
        mxs_rp['DO SCATTERING_LENS'] = rp.doScattering
        mxs_rp['SCATTERING_LENS'] = rp.scattering / 2500.0
        mxs_rp['DO DEVIGNETTING'] = rp.doDevignetting
        mxs_rp['DEVIGNETTING'] = rp.devignetting / 100.0

        ### Channels ######################################
        channels = rp.channels

        def channel_path(channel, name):
            fmt = channel.format.split("_", 1)
            if channels.output == '0':
                filename = "%s%s_%s.%s" % (self.out_name, self.out_suffix, name, fmt[0])
            else:
                filename = "%s%s.%s" % (self.out_name, self.out_suffix, fmt[0])
            return {
                "path": os.path.join(self.out_dir, filename),
                "depth": 32 if len(fmt) < 2 else int(fmt[1])
            }

        render_type = channels.render_type
        if render_type == '1':
            mxs_rp['DO DIFFUSE LAYER'] = True
        elif render_type == '2':
            mxs_rp['DO REFLECTION LAYER'] = True
        mxs_rp['EMBED CHANNELS'] = (channels.output == '1')
        # Output paths of the channels
        # "RENDER", "ALPHA", "SHADOW", "OBJECT", "MATERIAL", "MOTION", "Z",
        # "ROUGHNESS", "FRESNEL", "NORMALS", "POSITION", "DEEP", "UV", "ALPHA_CUSTOM".
        mxs_paths = self.mxs.Paths
        # Render
        mxs_paths['RENDER'] = {
            "path": os.path.join(self.out_dir, self.out_name + self.out_ext),
            "depth": int(self.scene.render.image_settings.color_depth)
        }
        mxs_rp['DO RENDER CHANNEL'] = channels.render_enabled
        # Alpha
        mxs_paths['ALPHA'] = channel_path(channels.alpha, "alpha")
        mxs_rp['DO ALPHA CHANNEL'] = channels.alpha.enabled
        mxs_rp['OPAQUE ALPHA'] = channels.alpha_opaque
        # Z-Buffer
        mxs_paths['Z'] = channel_path(channels.z_buffer, "zbuffer")
        mxs_rp['DO ZBUFFER CHANNEL'] = channels.z_buffer.enabled
        mxs_rp['ZBUFFER RANGE'] = channels.z_range
        # Shadow
        mxs_paths['SHADOW'] = channel_path(channels.shadow, "shadow")
        mxs_rp['DO SHADOW PASS CHANNEL'] = channels.shadow.enabled
        # Material ID
        mxs_paths['MATERIAL'] = channel_path(channels.material_id, "material")
        mxs_rp['DO IDMATERIAL CHANNEL'] = channels.material_id.enabled
        # Object ID
        mxs_paths['OBJECT'] = channel_path(channels.object_id, "object")
        mxs_rp['DO IDOBJECT CHANNEL'] = channels.object_id.enabled
        # Motion Vector
        mxs_paths['MOTION'] = channel_path(channels.motion, "motion")
        mxs_rp['DO MOTION CHANNEL'] = channels.motion.enabled
        # Roughness
        mxs_paths['ROUGHNESS'] = channel_path(channels.roughness, "roughness")
        mxs_rp['DO ROUGHNESS CHANNEL'] = channels.roughness.enabled
        # Fresnel
        mxs_paths['FRESNEL'] = channel_path(channels.fresnel, "fresnel")
        mxs_rp['DO FRESNEL CHANNEL'] = channels.fresnel.enabled
        # Normals
        mxs_paths['NORMALS'] = channel_path(channels.normals, "normals")
        mxs_rp['NORMALS CHANNEL SPACE'] = int(channels.normals_space)
        mxs_rp['DO NORMALS CHANNEL'] = channels.normals.enabled
        # Position
        mxs_paths['POSITION'] = channel_path(channels.position, "position")
        mxs_rp['POSITION CHANNEL SPACE'] = int(channels.position_space)
        mxs_rp['DO POSITION CHANNEL'] = channels.position.enabled
        # UV
        mxs_paths['UV'] = channel_path(channels.uv, "uv")
        mxs_rp['DO UV CHANNEL'] = channels.uv.enabled

        ### Overlay Text ##################################
        ot = rp.get("overlay_text")
        if ot:
            self.mxs.OverlayText = ot.to_dict()

    def _environment(self, env):
        """
        Args:
            env (properties.maxwell_environment): Environment settings.
        """
        mxs_environ = self.mxs.Environment
        # ## sky ###########################################
        mxs_environ.PhysicalSkyAtmosphere = env.physical()
        mxs_environ.SkyConstant = env.constant()
        mxs_environ.ActiveSky = env.sky if env.type != 'none' else None
        ### sun ###########################################
        sun = env.sun
        mxs_environ.SunProperties = {
            "type": int(sun.type),
            "power": sun.power,
            "radius": sun.radius,
            "temperature": sun.temperature,
            "color": sun.color
        }
        mxs_environ.SunLongitudeAndLatitude = {
            'latitude': sun.latitude,
            'longitude': sun.longitude,
            'sm': sun.sm,
            'dayOfYear': sun.day,
            'timeOfDay': sun.time
        }
        mxs_environ.SunRotation = sun.rotation
        mxs_environ.SunAngles = {
            'zenith': sun.zenith,
            'azimuth': sun.azimuth
        }
        obj = self.scene.objects.get(sun.dir_object, None)
        #mxs_environ.SunDirection = obj.location if obj else sun.direction
        mxs_environ.SunDirection = obj.matrix_world.col[2].xyz if obj else sun.direction
        mxs_environ.SunPositionType = int(sun.position)
        ### ibl ###########################################
        ibl = env.ibl
        mxs_environ.EnvironmentEnabled = (env.type == 'ibl')
        mxs_environ.EnvironmentWeight = ibl.intensity
        ### layers ########################################
        mx_layers = mxs_environ.Layers
        interpolation = ibl.interpolation
        background = ibl.background(interpolation)
        bkg = background.copy()
        bkg["state"] = 3
        if_bkg = lambda l, i: bkg if l.type == '3' else l(i)
        background["spherical"] = not ibl.screen_mapping
        mx_layers[mx_layers.BACKGROUND] = background
        mx_layers[mx_layers.REFLECTION] = if_bkg(ibl.reflection, interpolation)
        mx_layers[mx_layers.REFRACTION] = if_bkg(ibl.refraction, interpolation)
        mx_layers[mx_layers.ILLUMINATION] = if_bkg(ibl.illumination, interpolation)

    def export(self):
        self.log("--- begin export (v{0}.{1}.{2}) ---".format(*bl_info["version"]))

        result = False
        if not os.path.isdir(self.out_dir):
            self.log("Output directory not found or not writable: '%s'" % self.out_dir, {'ERROR'})
            return result

        scene = self.scene
        mx = scene.maxwell
        render = scene.render
        frame_current = scene.frame_current

        mx_path = os.path.join(self.out_dir, self.out_mxname + self.out_suffix)
        mxs_path = mx_path + ".mxs"
        mxi_path = mx_path + ".mxi"

        cmd_args = mx.cmdline.split()
        if self.is_animation and (mx.run != 'render' or "-nogui" not in cmd_args):
            ms_path = mx_path + ".ms"
            run_render_script = False
            if frame_current == scene.frame_start:
                with open(ms_path, "w") as sf:
                    sf.write(SCRIPT_HEADER)
                    sf.write("\t\"%s\",\n" % mxs_path)
            else:
                with open(ms_path, "a") as sf:
                    sf.write("\t\"%s\",\n" % mxs_path)
                    if frame_current == scene.frame_end:
                        sf.write(SCRIPT_FOOTER)
                        if mx.run == 'render':
                            run_render_script = True
            result = ('render_script', (run_render_script, ms_path))
        elif mx.run == 'studio':
            result = ('studio', (mxs_path, ))
        elif mx.run == 'network':
            result = ('network', (mxs_path, mx.run_manager, mx.run_node, mx.manager ))
        elif mx.run == 'render':
            offset = self.border[:2] if self.border else (0, 0)
            result = ('render', (mxs_path, mxi_path, cmd_args, self.xres, self.yres, offset))

        self.log("MXS: %s" % mxs_path)
        if not render.use_overwrite:
            if os.path.isfile(mxs_path):
                self.log("--- end export (mxs exists) ---")
                if mx.run == 'render' and \
                        "-trytoresume" not in cmd_args and \
                        os.path.isfile(mxi_path):
                    result[1][2].append("-resume:'%s'" % mxi_path)
                return result

        ### render layers #################################
        self.layers = [i for i, j in enumerate(scene.layers) if j]

        ### global scale ##################################
        us = scene.unit_settings
        if us.system != 'NONE':
            self.scale = us.scale_length

        self.aspect = render.pixel_aspect_x / render.pixel_aspect_y
        self.fps = render.fps

        ### maxwell scene #################################
        self.mxs = maxwell.Scene()
        self.mxs.AxisConversion = "ZXY"
        # blender uses floats for geometry
        self.mxs.SetSinglePrecisionOfGeometry()
        self.mxs.ProtectionEnabled = mx.protected

        ### materials #####################################
        self.materials = Materials(self.mxs)
        #self.materials.export_all(bpy.data.materials)

        self._render_params(mx.rp, mxi_path)
        world = self.scene.world
        if world:
            self._environment(world.maxwell)

        self.log("--- processing objects ---")
        st = time.perf_counter()

        ### objects #######################################
        dupli_objects = set()
        dupli_groups = set()
        duplicators = set()
        cameras = set()
        particles = []
        extensions = []
        render_objects = []
        for bl_obj in self.scene.objects:
            self.log("{0:<7}: '{1}'".format(bl_obj.type, bl_obj.name))

            ex_obj = Object(bl_obj, self)
            self.objects[bl_obj] = ex_obj

            if ex_obj.hidden:
                continue

            if bl_obj.type == 'CAMERA':
                cameras.add(ex_obj)
                continue

            if bl_obj.particle_systems:
                use_render_emitter = False
                mps = {m.particle_system: m for m in bl_obj.modifiers if m.type == 'PARTICLE_SYSTEM'}
                for ps in bl_obj.particle_systems:
                    show_render = mps[ps].show_render
                    pss = ps.settings
                    if pss.use_render_emitter and (
                            bl_obj.dupli_type != 'FRAMES' or
                            pss.render_type in ('NONE', 'HALO') or (
                            show_render and (
                                (pss.render_type == 'OBJECT' and pss.dupli_object) or
                                (pss.render_type == 'GROUP' and pss.dupli_group)))):
                        use_render_emitter = True
                        ex_obj.flags |= 0x10
                    if show_render:
                        pss_mx = pss.maxwell
                        if pss.render_type == 'OBJECT':
                            if pss.dupli_object:
                                if pss.type == 'HAIR' and pss_mx.scatter.enabled:
                                    extensions.append((ex_obj, ps, 'SCATTER'))
                                    use_render_emitter = True
                                else:
                                    dupli_objects.add(pss.dupli_object)
                        elif pss.render_type == 'GROUP':
                            if pss.dupli_group:
                                dupli_groups.add(pss.dupli_group)
                        elif pss.type == 'HAIR':
                            if pss.render_type == 'PATH':
                                extensions.append((ex_obj, ps, pss_mx.h_type))
                        elif pss.render_type != 'NONE':
                            particles.append(ps)
                if use_render_emitter:
                    ex_obj.flags |= 0x20
                    render_objects.append(ex_obj)
                if bl_obj.is_duplicator:
                    # TODO: add only if 'bl_obj' have a particle_system duplicator
                    dt = 1 if bl_obj.dupli_type == 'FRAMES' else 3
                    ex_obj.duplicator_type = dt
                    duplicators.add(ex_obj);
            elif bl_obj.dupli_type == 'GROUP':
                if bl_obj.dupli_group:
                    dupli_groups.add(bl_obj.dupli_group)
                    if not ex_obj.flags & 0x2:
                        ex_obj.duplicator_type = 0
                        duplicators.add(ex_obj)
            elif bl_obj.dupli_type == 'FRAMES':
                if not ex_obj.flags & 0x2:
                    ex_obj.duplicator_type = 1
                    duplicators.add(ex_obj)
            elif bl_obj.dupli_type in ('VERTS', 'FACES'):
                if not ex_obj.flags & 0x2:
                    ex_obj.duplicator_type = 2
                    duplicators.add(ex_obj)
                ex_obj.flags |= 0x1
            elif not ex_obj.flags & 0x3:
                ex_obj.flags |= 0x20
                render_objects.append(ex_obj)

        ### dupli objects ###
        for obj in dupli_objects:
            self.log("DupliObject: '%s'" % obj.name)
            # object may belong to another scene
            ex_obj = self.objects.get(obj)
            if ex_obj is None:
                self.objects[obj] = Object(obj, self)
            else:
                ex_obj.flags &= ~0x20
                try:
                    render_objects.remove(ex_obj)
                except ValueError:
                    pass

        ### dupli groups ###
        for group in dupli_groups:
            self.log("DupliGroup: '{0}'".format(group.name))
            for obj in dupli_group_objects(group):
                self.log("  {0:<7}: '{1}'".format(obj.type, obj.name))
                ex_obj = self.objects.get(obj)
                if ex_obj is None:
                    self.objects[obj] = Object(obj, self)

        self.log("--- exporting ---")
        ### maxwell objects ###############################
        for ex_obj in render_objects:
            self.log(ex_obj)
            ex_obj.export()

        ### instances ###
        if duplicators:
            fmt = BMaxwellPreferences.get_preferences().format_instances
            if fmt:
                Object.INSTANCE_NAME_FORMAT = fmt
            for ex_obj in sorted(duplicators, key=lambda t: t.duplicator_type):
                ex_obj.create_instances()

        ### extensions (grass, hair, scatter) ###
        for ex_obj, ps, et in extensions:
            ex_obj.create_extension(ps, et)

        ### particles ###
        ps_files = {}
        for ps in particles:
            pss = ps.settings
            name = ps.name
            # particles .bin file
            filepath = ps_files.get(ps)
            if filepath is None:
                _n = "{0}-{1}_{2:0{3}}.bin".format(self.out_mxname, name, frame_current, self.frame_digits)
                filepath = os.path.join(self.out_dir, _n)
                ret = self.mxs.CreateParticlesFile(
                    psys=ps,
                    file=filepath,
                    radius=pss.particle_size,
                    scene_scale=self.scale,
                    emitter_scale=(self.scale, ) * 3,
                    frame=frame_current,
                    fps=self.fps
                )
                if ret:
                    ps_files[ps] = filepath
                else:
                    ps_files[ps] = False
                    continue
            elif filepath == False:
                continue
            # extension
            mx = pss.maxwell
            self.log("Particles: '%s' [%s]\n    %s" % (name, mx.p_type, filepath))
            if mx.p_type == 'MESHER':
                mxobj = self.mxs.CreateMaxwellMesher(
                    name=name,
                    params=mx.mesher,
                    file=filepath,
                    frame=frame_current,
                    fps=self.fps
                )
            elif mx.p_type == 'PARTICLES':
                mxobj = self.mxs.CreateMaxwellParticles(name, mx.particles, filepath)
            elif mx.p_type == 'VOLUMETRIC':
                mxobj = self.mxs.CreateVolumetric(
                    name=name,
                    type=0,  # particle based
                    params=mx.volumetric,
                    file=filepath,
                    frame=frame_current,
                    fps=self.fps
                )
            # material and base
            obj = ps.id_data
            mxm = self.materials.get(obj=obj, slot=pss.material - 1)
            if mxm:
                mxobj.Material = mxm
            ex_obj = self.objects.get(obj)
            if ex_obj and ex_obj.mx_object:
                mxobj.Parent = ex_obj.mx_object
                mxobj.BaseAndPivot = {"base": obj.matrix_world.inverted()}
            else:
                mxobj.BaseAndPivot = {"base": Matrix.Scale(self.scale, 4)}

        ### motion blur ###################################
        if self.mb_steps:
            self.log("--- motion blur ---")
            pcmb = time.perf_counter()
            frame = frame_current
            for i in range(1, self.mb_steps + 1):
                step_time = i / self.mb_steps
                frame += self.mb_step
                f = int(frame)
                self.scene.frame_set(f, frame - f)
                for ex_obj in self.objects.values():
                    ex_obj.update(step_time)
                    if hasattr(ex_obj, "instances"):
                        ex_obj.update_instances(step_time)
                ### calculate motion blur steps for cameras ###
                for c in cameras:
                    if step_time != 1.0:
                        c.add_step(step_time, True)
                    elif c.step_time != 0.0:
                        c.add_step(1.0)
            self.scene.frame_set(frame_current)
            self.log("steps=%d frames=%f (%f)" % (self.mb_steps, self.mb_frames, time.perf_counter() - pcmb))

        ### cameras ###
        for cam in cameras:
            self.log(cam)
            cam.export()

        ### write mxs file ################################
        self.mxs.WriteMXS(mxs_path)

        self.log("--- end export (%f) ---" % (time.perf_counter() - st))
        return result
