# -*- coding: utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Ildar Nikolaev"
__email__ = "nildar@users.sourceforge.net"

from mathutils import Vector, geometry
from .object import Object


def _dof(camera, mw):
    if camera.dof_object:
        loc = mw.to_translation()
        n = mw * Vector([0, 0, 1])
        t = camera.dof_object.matrix_world.to_translation()
        dof = geometry.distance_point_to_plane(loc, t, n)
    elif camera.dof_distance:
        dof = camera.dof_distance
    else:
        dof = 1.0
    return dof


class Camera(Object):
    def __init__(self, bl_object, exporter):
        super().__init__(bl_object, exporter)
        self.flags |= 0x8

        self.steps = []
        self.step_time = 0.0

        mw = bl_object.matrix_world.copy()
        
        camera = bl_object.data
        mx = camera.maxwell

        self.lens = camera.lens

        self.fstop = mx.fstop
        self.correction = mx.correction

        self.matrix_world = mw
        self.dof = _dof(camera, mw)

        self.add_step()

    def __str__(self):
        return "Camera:" + super().__str__()

    def add_step(self, step_time=0.0, if_changed=False):
        if if_changed:
            bl_object = self.bl_object
            camera = bl_object.data
            mx = camera.maxwell
            mw = bl_object.matrix_world
            dof = _dof(camera, mw)

            changed = False
            if self.lens != camera.lens:
                self.lens = camera.lens
                changed = True
            if self.fstop != mx.fstop:
                self.fstop = mx.fstop
                changed = True
            if self.correction != mx.correction:
                self.correction = mx.correction
                changed = True
            if self.matrix_world != mw:
                self.matrix_world = mw.copy()
                changed = True
            if self.dof != dof:
                self.dof = dof
                changed = True
            if not changed:
                return

        self.steps.append({
            "time": step_time,
            "lens": self.lens,
            "fstop": self.fstop,
            "correction": self.correction,
            "matrix": self.matrix_world,
            "dof": self.dof
        })
        self.step_time = step_time

    def export(self):
        exporter = self.exporter
        bl_object = self.bl_object
        camera = bl_object.data
        mx = camera.maxwell

        if camera.sensor_fit == 'VERTICAL':
            height = camera.sensor_height
        elif camera.sensor_fit == 'HORIZONTAL' or (exporter.yres / exporter.xres) < exporter.aspect:
            height = (camera.sensor_width * exporter.yres) / (exporter.aspect * exporter.xres)
        else:
            height = camera.sensor_width
        width = 0.0

        if mx.shtype == 'ANGLE':
            shutter = 360 * exporter.fps / mx.rotary
        else:
            assert(mx.shtype == 'SPEED')
            shutter = mx.shutter

        mx_cam = exporter.mxs.AddCamera(
            name=bl_object.name,
            xres=exporter.xres,
            yres=exporter.yres,
            film_width=width / 1000.0,
            film_height=height / 1000.0,
            aspect=exporter.aspect,
            iso=mx.iso,
            shutter=1 / shutter,
            diaphragm=mx.diaphragm,
            blades=mx.blades,
            angle=mx.angle,
            steps=len(self.steps),
            fps=exporter.fps,
            #lens_type=TYPE_THIN_LENS  # TODO: Implement different camera types
        )

        mx_cam.CutPlanes = (
            camera.clip_start * exporter.scale,
            camera.clip_end * exporter.scale,
            True
        )

        mx_cam.ShiftLens = (
            camera.shift_x * 100,
            camera.shift_y * -100
        )

        if exporter.border:
            mx_cam.ScreenRegion = ('REGION', exporter.border)

        for step, params in enumerate(self.steps):
            matrix = params["matrix"] * exporter.scale
            mx_cam.SetStep(
                step=step,
                origin=matrix.to_translation(),
                focal_point=matrix * Vector([0, 0, -params["dof"]]),
                up=matrix.col[1][:3],  #(matrix * Vector([0, 1, 0, 0]))[:3]
                focal_length=params["lens"] / 1000,
                f_stop=params["fstop"],
                step_time=params["time"],
                fl_correction=params["correction"]
            )

        if bl_object == exporter.scene.camera:
            mx_cam.SetActive()
