# -*- coding: utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Ildar Nikolaev"
__email__ = "nildar@users.sourceforge.net"

from numbers import Number

from .. import maxwell
from ..engine import abspath


class Materials:
    def __init__(self, mxs=None):
        """
        Args:
            mxs (maxwell.Scene): .
        """
        self.mxs = mxs if mxs else maxwell.Scene()
        self.materials = {}
        self.textures = {}

    def get(self, **kwargs):
        """
        .get(mesh=mesh) -> list
        .get(obj=object, slot=slot) -> maxwell.Material

        Args:
            mesh (bpy.types.Mesh): get all materials assigned to mesh.
            or
            obj (bpy.types.Object): get material assigned to object by slot number or by name.
            slot (int or str): slot number [1, 32767] or name.

        Returns (list of maxwell.Material, maxwell.Material or None):
            list of maxwell materials for a specified mesh
            or maxwell material for a slot in an object
        """
        mesh = kwargs.get("mesh", None)
        if mesh:
            return [self._get(m) for m in mesh.materials]

        material = None
        obj = kwargs["obj"]
        slot = kwargs["slot"]
        try:
            material = obj.material_slots[slot].material
        except KeyError:
            pass
        except IndexError:
            pass
        return self._get(material)

    def _get(self, material):
        if material:
            mxm = self.materials.get(material, False)
            if mxm == False:
                mxm = self.export(material)
                # TODO: add searching paths to mxs???
                self.materials[material] = mxm
            return mxm
        return None

    def get_texture(self, id_data, name):
        if name:
            mx_tex = self.textures.get(name, False)
            if mx_tex:
                return mx_tex

            slot = id_data.texture_slots.get(name, False)
            if slot:
                tex = slot.texture
                mx = getattr(tex, "maxwell", None)
                if mx is not None:
                    mx_tex = maxwell.Texture(
                        # Bitmap Properties
                        path=abspath(tex.image.filepath),
                        # Projection Properties
                        globalMap=mx.override,
                        channel=mx.channel,
                        uIsTiled=mx.tile_x,
                        vIsTiled=mx.tile_y,
                        absolute=mx.get("units", False),
                        scale=mx.repeat,
                        uIsMirrored=mx.mirror_x,
                        vIsMirrored=mx.mirror_y,
                        offset=mx.offset,
                        rotation=mx.rotation,
                        # Image Properties
                        flipRed=mx.flip_x,
                        flipGreen=mx.flip_y,
                        fullRange=mx.wide,
                        #gamma=mx.gamma,
                        invert=mx.invert,
                        alpha=mx.alpha_only,
                        interpolation=mx.interpolation,
                        brightness=mx.brightness / 100,
                        saturation=mx.saturation / 100,
                        contrast=mx.contrast / 100,
                        hue=mx.hue / 100,
                        clampMin=mx.clamp[0] / 255,
                        clampMax=mx.clamp[1] / 255,
                        #cosA=mx.cosA,
                        #sinA=mx.sinA
                    )
                    self.textures[name] = mx_tex
                    return mx_tex
        return maxwell.Texture()

    def get_attribute(self, id_data, name, value):
        """
        Args:
            id_data (bpy.types.ID): ID with 'texture_slots' attribute
                (bpy.types.Material or bpy.types.ParticleSettings).
            name (str): texture slot name.
            value (float or tuple): value or color if texture does't exist.

        Returns (maxwell.Attribute): .
        """
        params = {}

        if isinstance(value, Number):
            params["type"] = 0  # MAP_TYPE_VALUE
            params["value"] = value
        elif hasattr(value, "__len__"):
            params["type"] = 1  # MAP_TYPE_RGB
            params["color"] = value

        mx_tex = self.get_texture(id_data, name)
        if not mx_tex.isEmpty:
            params["type"] = 2  # MAP_TYPE_VALUE
            params["texture"] = mx_tex

        return maxwell.Attribute(**params)

    def export(self, mat):
        """
        Args:
            mat (bpy.types.Material): Blender material.

        Returns (maxwell.Material): .
        """
        mx_mat = mat.maxwell
        mx_type = mx_mat.type
        if mx_type == 'file':
            mxm = None
            path = abspath(mx_mat.mxm)
            if mx_mat.embed:
                mxm = self.mxs.ReadMaterial(path)
                if mxm.isNull:
                    mxm = None
            else:
                mxm = self.mxs.CreateMaterial(mat.name, True)
                mxm.Reference = (path, True)
            return mxm

        mxm = self.mxs.CreateMaterial(mat.name, True)

        mx_map = mx_mat.override_map
        mxm.GlobalMap = maxwell.Texture(
            channel=mx_map.channel,
            uIsTiled=mx_map.tile_x,
            vIsTiled=mx_map.tile_y,
            absolute=mx_map.get("units", False),
            scale=mx_map.repeat,
            uIsMirrored=mx_map.mirror_x,
            vIsMirrored=mx_map.mirror_y,
            offset=mx_map.offset,
            rotation=mx_map.rotation
        )

        mxm.Bump = self.get_attribute(mat, mx_mat.bump_tex, mx_mat.bump)
        mxm.NormalMapState = mx_mat.bump_normal

        mxm.Matte = mx_mat.matte
        mxm.MatteShadow = mx_mat.shadow
        mxm.Dispersion = mx_mat.dispersion
        mxm.ColorID = mx_mat.pi_color if mx_mat.pi_as_color else mx_mat.color_id

        if mx_type == 'custom':
            # Displacement
            d = mx_mat.displacement
            if d.enabled:
                mxm_d = mxm.Displacement
                mxm_d.Enabled = d.enabled
                mxm_d.Map = self.get_texture(mat, d.map)
                mxm_d.Parameters = {
                    "type": d.get("type", 1),
                    "level": d.level,
                    "smooth": d.smooth,
                    "offset": d.offset,
                    "method": d.get("method", 0),
                    "uvi": d.get("uvi", 2)
                }
                mxm_d.HeightMap = {
                    "absolute": d.absolute,
                    "height": d.height / 100 if d.absolute else d.height,
                    "adaptive": d.adaptive
                }
                mxm_d.Vector = {
                    "scale": d.scale,
                    "transform": d.get("transform", 0),
                    "mapping": d.get("mapping", 0),
                    "preset": d.get("preset", 0)
                }

            # Layers
            for layer in mx_mat.layers:
                mxm_layer = mxm.AddLayer()
                mxm_layer.Name = layer.name
                mxm_layer.Enabled = layer.enabled
                mxm_layer.Blending = layer.get("blend", 0)
                mxm_layer.Opacity = self.get_attribute(mat, layer.opacity_tex, layer.opacity)

                # Emitter
                emitter = layer.emitter
                if emitter.enabled:
                    mxm_emitter = mxm_layer.CreateEmitter()

                    mxm_emitter.Enabled = True
                    mxm_emitter.Type = emitter.get("type", 0)

                    lobe = mxm_emitter.Lobe
                    lobe.Type = emitter.get("lobe", 0)
                    lobe.IES = emitter.ies
                    lobe.Intensity = emitter.intensity
                    lobe.Image = emitter.type == '2', self.get_texture(mat, emitter.image)

                    mxm_emitter.Temperature = emitter.temperature
                    mxm_emitter.MXI = self.get_attribute(mat, emitter.hdri_tex, emitter.hdri_intensity)

                    mxm_emitter.ActivePair = (
                        # colorType: 0 = EMISSION_RGB, 2 = EMISSION_COLOR_TEMPERATURE
                        2 if emitter.color_as_temp else 0,  # Luminance Units
                        emitter.get("luminance", 0)
                    )
                    luminance = emitter.luminance
                    if luminance == '0':
                        pair = {
                            "watts": emitter.power,
                            "efficacy": emitter.efficacy
                        }
                    else:
                        output = emitter.output
                        pair = {
                            "luminous": output,  # In really need only this
                            "illuminance": output,
                            "intensity": output,
                            "luminance": output
                        }
                    pair["rgb"] = emitter.color
                    pair["temperature"] = emitter.color_temp
                    mxm_emitter.Pair = pair

                # BSDFs
                for bsdf in layer.bsdfs:
                    mxm_bsdf = mxm_layer.AddBSDF()
                    mxm_bsdf.Name = bsdf.name
                    mxm_bsdf.Enabled = bsdf.enabled
                    mxm_bsdf.Weight = self.get_attribute(mat, bsdf.weight_tex, bsdf.weight)

                    refl = mxm_bsdf.Reflectance
                    attrs = refl.Attributes

                    # BSDF Properties
                    refl.ComplexIor = abspath(bsdf.ior_file)
                    refl.ActiveIorMode = bsdf.get("ior_type", 0)
                    attrs["color"] = self.get_attribute(mat, bsdf.refl_0_tex, bsdf.refl_0)
                    attrs["color.tangential"] = self.get_attribute(mat, bsdf.refl_90_tex, bsdf.refl_90)
                    attrs["transmittance.color"] = self.get_attribute(mat, bsdf.trans_tex, bsdf.trans)
                    refl.AbsorptionDistance = (bsdf.get("atten_units", 0), bsdf.attenuation)
                    refl.IOR = (bsdf.nd, bsdf.abbe)
                    refl.Conductor = bsdf.k
                    refl.ForceFresnel = bsdf.force_fresnel
                    refl.FresnelCustom = (bsdf.cf_angle, bsdf.cf_roughness, bsdf.cf_enabled)

                    # Surface Properties
                    mxm_bsdf.Roughness = self.get_attribute(mat, bsdf.roughness_tex, bsdf.roughness)
                    mxm_bsdf.Anisotropy = self.get_attribute(mat, bsdf.anisotropy_tex, bsdf.anisotropy)
                    mxm_bsdf.Angle = self.get_attribute(mat, bsdf.angle_tex, bsdf.angle)
                    mxm_bsdf.Bump = self.get_attribute(mat, bsdf.bump_tex, bsdf.bump)
                    mxm_bsdf.NormalMapState = bsdf.bump_normal

                    # Subsurface Properties
                    attrs["scattering"] = self.get_attribute(mat, bsdf.scattering_tex, bsdf.scattering)
                    refl.ScatteringParameters = (bsdf.coef, bsdf.asymmetry, bsdf.single_sided)
                    refl.ScatteringThickness = self.get_attribute(mat, bsdf.thickness_tex, bsdf.thickness / 1000.0)
                    refl.ScatteringThicknessRange = (bsdf.thickness_min / 1000.0, bsdf.thickness_max / 1000.0)

                    # Coating
                    coating = bsdf.coating
                    if coating.enabled:
                        mxm_coating = mxm_bsdf.AddCoating()
                        mxm_coating.Enabled = True
                        mxm_coating.Thickness = self.get_attribute(
                            mat, coating.thickness_tex, coating.thickness / 1000000000.0)
                        mxm_coating.ThicknessRange = (
                            coating.thickness_min / 1000000000.0,
                            coating.thickness_max / 1000000000.0
                        )

                        refl = mxm_coating.Reflectance
                        refl.ComplexIor = abspath(coating.ior_file)
                        refl.ActiveIorMode = coating.get("ior_type", 0)
                        refl.IOR = (coating.nd, 0.0)
                        refl.Conductor = coating.k
                        refl.ForceFresnel = coating.force_fresnel
                        refl.FresnelCustom = (coating.cf_angle, 0.0, coating.cf_enabled)

                        attrs = refl.Attributes
                        attrs["color"] = self.get_attribute(mat, coating.refl_0_tex, coating.refl_0)
                        attrs["color.tangential"] = self.get_attribute(mat, coating.refl_90_tex, coating.refl_90)
        elif mx_type == 'ags':
            mxm.ApplyModifier(0, mx_mat.ags)
        elif mx_type == 'opaque':
            params = mx_mat.opaque
            _maps = {}
            for m, n in [(params.color_map, "color"),
                         (params.shininess_map, "shininess"),
                         (params.roughness_map, "roughness")]:
                if m:
                    t = self.get_texture(mat, m)
                    if t:
                        _maps[n] = t
            mxm.ApplyModifier(1, params, _maps)
        elif mx_type == 'transparent':
            params = mx_mat.transparent
            _maps = {}
            for m, n in [(params.color_map, "color"),
                         (params.roughness_map, "roughness")]:
                if m:
                    t = self.get_texture(mat, m)
                    if t:
                        _maps[n] = t
            mxm.ApplyModifier(2, params, _maps)
        elif mx_type == 'metal':
            params = mx_mat.metal
            _maps = {}
            for m, n in [(params.color_map, "color"),
                         (params.roughness_map, "roughness"),
                         (params.anisotropy_map, "anisotropy"),
                         (params.angle_map, "angle"),
                         (params.dust_map, "dust"),
                         (params.perforation_map, "perforation")]:
                if m:
                    t = self.get_texture(mat, m)
                    if t:
                        _maps[n] = t
            mxm.ApplyModifier(3, params, _maps)
        elif mx_type == 'translucent':
            params = mx_mat.translucent
            _maps = {}
            for m, n in [(params.color_map, "color"),
                         (params.roughness_map, "roughness")]:
                if m:
                    t = self.get_texture(mat, m)
                    if t:
                        _maps[n] = t
            mxm.ApplyModifier(4, params, _maps)
        elif mx_type == 'carpaint':
            mxm.ApplyModifier(5, mx_mat.carpaint)
        elif mx_type == 'hair':
            params = mx_mat.hair
            _maps = {}
            for m, n in [(params.color_map, "color"),
                         (params.rt_map, "rt"),
                         (params.rt_weight_map, "rt_weight")]:
                if m:
                    t = self.get_texture(mat, m)
                    if t:
                        _maps[n] = t
            mxm.ApplyModifier(6, params, _maps)
        return mxm
