# -*- coding: utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Ildar Nikolaev"
__email__ = "nildar@users.sourceforge.net"

import bpy
from math import radians
from mathutils import Matrix, Euler

from ..engine import abspath

ROTATION_MATRIX = Matrix.Rotation(radians(90.0), 4, 'X')
PART_ID = lambda p: tuple(iter(iter(p).__next__, 0x7FFFFFFF))


def set_object_properties(mxobj, mx):
    mxobj.Opacity = mx.opacity / 100.0
    mxobj.ColorID = mx.pi_color if mx.pi_as_color else mx.color_id
    mxobj.HideToCamera = mx.hide_to_camera
    mxobj.HideToReflRefr = mx.hide_to_rr
    mxobj.HideToGI = mx.hide_to_gi
    mxobj.ExcludedOfCutPlanes = mx.ecp
    mxobj.SetCausticsProperties(mx.cd_refl, mx.cd_refr, mx.ci_refl, mx.ci_refr)


class _types(type):
    class _no_impl:
        def __new__(cls, bl_object, exporter):
            raise NotImplementedError

    def __new__(cls, name, bases, attrs):
        if '__new__' not in attrs:
            bases = (cls._no_impl, ) + bases
        t = type.__new__(cls, name, bases, attrs)
        setattr(cls, name, t)
        return t


class Object(metaclass=_types):
    __slots__ = [
        "bl_object",
        "exporter",
        "flags",
        "parent",
        "mx_object",
        "mx_instance",
        "matrix",
        "step_time",
        "duplicator_type",
        "instance_name",
        "instance_id",
        "instances",
    ]

    INSTANCE_NAME_FORMAT = "[I:{1:07}] {0}"

    def __new__(cls, bl_object, exporter_ref):
        t = bl_object.type
        if t == 'CAMERA':
            return object.__new__(_types.Camera)
        elif t == 'EMPTY':
            mx = bl_object.maxwell
            if mx.type == 'REFERENCE':
                return object.__new__(_types.ObjectReference)
            elif mx.type == 'VOLUMETRIC':
                return object.__new__(_types.ObjectVolumetric)
            else:
                return object.__new__(cls)
        elif t == 'MESH':
            if bl_object.data.maxwell.sea.enabled:
                return object.__new__(_types.ObjectSea)
        elif not t in ('CURVE', 'SURFACE', 'META', 'FONT'):
            return object.__new__(cls)
        return object.__new__(_types.ObjectMesh)

    def __init__(self, bl_object, exporter):
        """
        Args:
            bl_object (bpy.types.Object): Blender object.
            exporter_ref (weakref.ref): WeakRef to Exporter object.
        """
        self.bl_object = bl_object
        self.exporter = exporter

        # 0x01 - bl_object is duplicator (bl_object.dupli_type in ['VERTS', 'FACES'])
        # 0x02 - in parents bl_obj have a duplicator (with flag 0x1)
        # 0x04 - mx_object is an empty M~R object, group
        # 0x08 - bl_object can't be exported as a maxwell object
        # 0x10 - don't hide object. used by Scatter extension
        # 0x20 - 1: mx_object, 0: mx_instance
        self.flags = 0

        self.parent = None          # bmaxwell.export.Object
        self.mx_object = None       # maxwell.Object
        self.mx_instance = None     # maxwell.Object for dupli objects

        bl_parent = bl_object.parent

        matrix = bl_object.matrix_world
        if bl_parent is None:
            matrix *= exporter.scale
        else:
            matrix = bl_parent.matrix_world.inverted() * matrix
            parent = exporter.objects[bl_parent]
            if parent.flags & 0x3:
                if parent.flags & 0x2:
                    parent = parent.parent
                self.flags |= 0x2
            self.parent = parent

        # cached local (related to parent) matrix,
        # matrix_local not suitable its has own logic
        self.matrix = matrix  # bl_object.matrix_local.copy()

        # 0 - GOURP
        # 1 - FRAMES
        # 2 - VERTS, FACES
        # 3 - ParticleSystem
        # self.duplicator_type

        self.instance_id = 0
        #self.instance_name = None  # base name for instances
        #self.instances = {}        # {persistent_id: maxwell.Object (mesh or instance)}

        # for motion blur
        self.step_time = 0.0

    def __str__(self):
        if self.bl_object:
            return "'{0.name}' ({0.type}:{1:b})".format(self.bl_object, self.flags)
        return "None"

    @property
    def hidden(self):
        bl_obj = self.bl_object
        if not bl_obj.hide_render:
            layers = bl_obj.layers
            for i in self.exporter.layers:
                if layers[i]:
                    return False
        return True

    @property
    def mxobject(self):
        return self.mx_object

    @mxobject.setter
    def mxobject(self, mxobj):
        parent = self.parent
        if parent:
            mxp = parent.mx_object
            if not mxp:
                mxp = self.exporter.mxs.CreateMesh(parent.bl_object.name)
                mxp.Hide = True
                parent.flags |= 0x4
                parent.mxobject = mxp
            mxobj.Parent = mxp
        mxobj.BaseAndPivot = {"base": self.matrix}
        self.mx_object = mxobj

    @property
    def matrix_inverse(self):
        return self.bl_object.matrix_world.inverted()

    def _create(self, name):
        return None

    def export(self):
        bl_object = self.bl_object
        mxobj = self._create(bl_object.name)
        if mxobj:
            set_object_properties(mxobj, bl_object.maxwell)
            self.instance_name = mxobj.Name
            self.mxobject = mxobj
        else:
            self.flags |= 0x8
        return mxobj

    def update(self, step_time):
        mx_object = self.mx_object
        if mx_object is None:
            return

        parent = self.parent
        matrix = self.bl_object.matrix_world
        #matrix = self.bl_object.matrix_local.copy()
        if parent is not None:
            matrix *= self.exporter.scale
        else:
            matrix = parent.bl_object.matrix_world.inverted() * matrix

        if self.matrix != matrix:
            pass
        elif step_time == 1.0 and self.step_time != 0.0:
            pass
        #elif parent is not None and parent.step_time == step_time:
        #    pass
        else:
            return

        self.matrix = matrix
        self.step_time = step_time
        mx_object.BaseAndPivot = {"base": matrix, "time": step_time}

    def create_group(self):
        mxobj = self.exporter.mxs.CreateMesh(self.bl_object.name)
        mxobj.BaseAndPivot = {"base": self.matrix}
        mxobj.Hide = True
        self.flags |= 0x4
        self.mxobject = mxobj
        return mxobj

    def create_instance(self):
        inst = None
        fmt = self.INSTANCE_NAME_FORMAT
        if not self.flags & 0x20:
            if self.mx_instance:
                name = fmt.format(self.instance_name, self.instance_id)
                inst = self.exporter.mxs.CreateInstance(name, self.mx_instance)
                self.instance_id += 1
            else:
                bl_object = self.bl_object
                name = bl_object.name
                inst = self._create(fmt.format(name, self.instance_id))
                if inst:
                    set_object_properties(inst, bl_object.maxwell)
                    self.mx_instance = inst
                    self.instance_name = name
                    self.instance_id += 1
        elif self.mx_object and not (self.flags & 0x4):
            name = fmt.format(self.instance_name, self.instance_id)
            inst = self.exporter.mxs.CreateInstance(name, self.mx_object)
            self.instance_id += 1
        return inst

    def create_instances(self):
        ex_obj = None
        mx_parent = None
        bl_current = None
        matrix_inverse = None
        exporter = self.exporter
        bl_duplicator = self.bl_object

        exporter.log("Duplicator<{0.duplicator_type}:{1}> {0}".format(self, bl_duplicator.dupli_type))

        mb = bool(exporter.mb_steps)
        if mb:
            self.instances = {}

        bl_duplicator.dupli_list_create(exporter.scene, 'RENDER')
        try:
            for d in bl_duplicator.dupli_list:
                bl_obj = d.object
                if bl_obj.dupli_type in ('VERTS', 'FACES', 'GROUP'):
                    continue

                ps = d.particle_system
                if ps and ps.settings.maxwell.scatter.enabled:
                    continue

                if bl_current != bl_obj:
                    bl_current = bl_obj
                    ex_obj = exporter.objects[bl_obj]

                if ex_obj.flags & 0x8:
                    continue

                mx_inst = ex_obj.create_instance()
                if mx_inst:
                    if not mx_parent:
                        parent = self.parent if d.type == 'FRAMES' else self
                        mx_parent = parent.mx_object
                        if mx_parent is None:
                            mx_parent = parent.create_group()
                        matrix_inverse = parent.matrix_inverse
                    mx_inst.BaseAndPivot = {"base": matrix_inverse * d.matrix}
                    mx_inst.Parent = mx_parent
                    
                    if mb:  # store created instances for motion blur
                        self.instances[PART_ID(d.persistent_id)] = mx_inst
                else:
                    ex_obj.flags |= 0x8
        finally:
            bl_duplicator.dupli_list_clear()

    def update_instances(self, step_time):
        bl_duplicator = self.bl_object
        bl_duplicator.dupli_list_create(self.exporter.scene, 'RENDER')
        try:
            matrix = None
            for d in bl_duplicator.dupli_list:
                bl_obj = d.object
                if bl_obj.dupli_type in ('VERTS', 'FACES', 'GROUP'):
                    continue

                mx_inst = self.instances.get(PART_ID(d.persistent_id))
                if mx_inst is None:
                    continue

                if matrix is None:
                    if d.type == 'FRAMES':  # and not d.particle_system:
                        matrix = self.parent.bl_object.matrix_world.inverted()
                    else:
                        matrix = bl_duplicator.matrix_world.inverted()
                mx_inst.BaseAndPivot = {"base": matrix * d.matrix, "time": step_time}
        finally:
            bl_duplicator.dupli_list_clear()

    def create_extension(self, ps, et):
        """
        Args:
            ps (bpy.ParticleSystem): particle system.
            et (str): extension type ('GRASS', 'HAIR', 'SCATTER').
        """
        pss = ps.settings
        mx = pss.maxwell

        exporter = self.exporter
        mx_obj = self.mx_object
        bl_obj = self.bl_object

        if et == 'GRASS':
            exporter.log("Grass: '%s'" % ps.name)
            # TODO: export grass from a hidden mesh
            if mx_obj:
                maps = {}
                g = mx.grass
                for _map, _name in [(g.density_map, "density"),
                                    (g.length_map, "length"),
                                    (g.angle_map, "angle"),
                                    (g.start_bend_map, "start_bend"),
                                    (g.bend_radius_map, "bend_radius"),
                                    (g.bend_angle_map, "bend_angle"),
                                    (g.cut_off_map, "cut_off")]:
                    if _map:
                        _tex = exporter.materials.get_texture(pss, _map)
                        if _tex:
                            maps[_name] = _tex
                mxm = exporter.materials.get(obj=bl_obj, slot=pss.material - 1)
                mx_obj.ApplyGrass(g, None if not mxm else mxm.Name, maps)
        elif et == 'HAIR':
            exporter.log("Hair: '%s'" % ps.name)

            bl_scene = exporter.scene
            ps.set_resolution(bl_scene, bl_obj, 'RENDER')
            mx_hair = exporter.mxs.ImportHair(ps.as_pointer(), bl_obj.as_pointer())
            ps.set_resolution(bl_scene, bl_obj, 'PREVIEW')

            mxm = exporter.materials.get(obj=bl_obj, slot=pss.material - 1)
            if mxm:
                mx_hair.Material = mxm
            if mx_obj:
                mx_hair.Parent = mx_obj
                mx_hair.BaseAndPivot = {"base": bl_obj.matrix_world.inverted()}
            else:
                mx_hair.BaseAndPivot = {"base": Matrix.Scale(exporter.scale, 4)}
        elif et == 'SCATTER':
            exporter.log("Scatter: '%s'" % ps.name)

            ex_dobj = exporter.objects[pss.dupli_object]

            mx_name = ""
            if ex_dobj.mx_object:
                if ex_dobj.flags & 0x20:
                    mx_name = ex_dobj.mx_object.Name
                else:
                    # mx_object is group
                    # TODO: find a simple way to create mxobj and don't break things
                    pass
            elif ex_dobj.mx_instance:
                mxobj = exporter.mxs.CreateInstance(
                    ex_dobj.bl_object.name,
                    ex_dobj.mx_instance
                )
                mxobj.Hide = True
                ex_dobj.mxobject = mxobj
                mx_name = mxobj.Name
            elif not ex_dobj.flags & 0x8:
                mxobj = ex_dobj.export()
                if mxobj:
                    mxobj.Hide = True
                    ex_dobj.flags |= 0x20
                    mx_name = mxobj.Name

            maps = {}
            s = mx.scatter
            for _map, _name in [(s.density_map, "density"),
                                (s.scale_map, "scale"),
                                (s.rotation_map, "rotation")]:
                if _map:
                    _tex = exporter.materials.get_texture(pss, _map)
                    if _tex:
                        maps[_name] = _tex

            mx_obj.ApplyScatter(ps.as_pointer(), mx_name, maps)
            if not (self.flags & 0x10):
                mx_obj.Hide = True


class ObjectMesh(Object):
    def __str__(self):
        return "Mesh:" + super().__str__()

    def _create(self, name):
        exporter = self.exporter
        bl_object = self.bl_object

        id_data = bl_object.data
        if id_data.users < 2 or bl_object.is_modified(exporter.scene, 'RENDER'):
            return self._to_mesh(name)

        sh_obj = exporter.shared_objects.get(id_data)
        if sh_obj:
            return exporter.mxs.CreateInstance(name, sh_obj)

        mxobj = self._to_mesh(name)
        if mxobj:
            exporter.shared_objects[id_data] = mxobj
        return mxobj

    def _to_mesh(self, name):
        exporter = self.exporter
        bl_scene = exporter.scene
        bl_object = self.bl_object
        mesh = bl_object.to_mesh(bl_scene, True, 'RENDER')
        try:
            # TODO: make more readable
            uvsets = tuple()
            mx = getattr(bl_object.data, "maxwell", None)
            if mx and mx.use_projections:
                uvsets = []
                for p in mx.projections:
                    if p.type == 'LOCKED':
                        uvsets.append((p.as_pointer(), None))
                    else:
                        bb = bl_object.bound_box
                        s = bl_object.matrix_world.to_scale() * exporter.scale
                        x = s[0] * (bb[6][0] - bb[0][0])
                        y = s[1] * (bb[6][1] - bb[0][1])
                        z = s[2] * (bb[6][2] - bb[0][2])
                        if p.object:
                            o = bl_scene.objects.get(p.object)
                            if o:
                                base = self.matrix_inverse * o.matrix_world
                                base *= ROTATION_MATRIX
                            else:
                                base = Matrix()
                        else:
                            x *= p.scale[0]
                            y *= p.scale[1]
                            z *= p.scale[2]
                            if x:
                                x = 1 / x
                            if y:
                                x = 1 / y
                            if z:
                                z = 1 / z
                            base = Matrix([
                                [x, 0, 0],
                                [0, y, 0],
                                [0, 0, z]
                            ])
                            base.rotate(Euler(p.rotation))
                            base = base.to_4x4()
                            base.translation = p.position
                        uvsets.append((p.as_pointer(), base))
                uvsets = (uvsets, )

            mesh.calc_normals()
            mats = exporter.materials.get(mesh=mesh)
            mxobj = exporter.mxs.ImportMesh(name, mesh, mats, *uvsets)
            # can be None if mesh dont have any triangles
            if mxobj:
                bfm = bl_object.maxwell.backface_material
                if bfm:
                    mxm = exporter.materials.get(obj=bl_object, slot=bfm)
                    if mxm:
                        mxobj.BackfaceMaterial = mxm
            return mxobj
        finally:
            bpy.data.meshes.remove(mesh)
        return None


class ObjectReference(Object):
    def __str__(self):
        return "Reference:" + super().__str__()

    def _create(self, name):
        bl_object = self.bl_object

        mx = bl_object.maxwell
        reference = mx.reference

        path = abspath(reference.path)
        if not path:
            return None

        mxobj = self.exporter.mxs.CreateMesh(name)  # create empty (group) object
        mxobj.ReferencedFlags = reference.get("flags", 0)
        mxobj.ReferencedPath = path
        return mxobj


class ObjectVolumetric(Object):
    def __str__(self):
        return "Volumetric:" + super().__str__()

    def __init__(self, bl_obj, exporter):
        super().__init__(bl_obj, exporter)
        self.matrix *= Matrix.Scale(2 * bl_obj.empty_draw_size, 4)

    def _create(self, name):
        exporter = self.exporter
        bl_object = self.bl_object

        mx = bl_object.maxwell
        volumetric = mx.volumetric

        mxobj = exporter.mxs.CreateVolumetric(
            name=name, type=volumetric.get("field_type", 1), params=volumetric)
        if bl_object.parent:
            material = volumetric.material
            if material:
                mxm = exporter.materials.get(obj=bl_object.parent, slot=material)
                if mxm is not None:
                    mxobj.Material = mxm
        return mxobj

    def update(self, step_time):
        # TODO: implement?
        pass


class ObjectSea(Object):
    def __str__(self):
        return "Sea:" + super().__str__()

    def _create(self, name):
        exporter = self.exporter
        bl_object = self.bl_object

        sea = bl_object.data.maxwell.sea

        mxobj = exporter.mxs.CreateSea(name, sea)
        if mxobj:
            material = sea.material
            if material:
                mxm = exporter.materials.get(obj=bl_object, slot=material)
                if mxm:
                    mxobj.Material = mxm
        return mxobj
