# -*- coding: utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Ildar Nikolaev"
__email__ = "nildar@users.sourceforge.net"

__all__ = (
    "Rgb",
    "Point2D",
    "Texture",
    "Attribute",
    "Scene",
    "ObjectIterator",
    "MxiReader",
    "MxImage",
    "MxRender",
    "TexturePreview",
)


import os
import sys
from contextlib import contextmanager

from ..engine import get_maxwell_path


class _MaxwellLoaderMeta(type):
    def __new__(cls, name, bases, attrs):
        if '_loader' not in attrs:
            if sys.platform == 'win32':
                attrs['_loader'] = cls._loader_windows
            else:
                attrs['_loader'] = cls._loader_posix
        return super().__new__(cls, name, bases, attrs)

    @contextmanager
    def _loader_windows(cls, path):
        from ctypes import windll, create_unicode_buffer

        k32 = windll.kernel32
        size = k32.GetDllDirectoryW(0, None)
        if size > 0:
            directory = create_unicode_buffer(size + 1)
            k32.GetDllDirectoryW(size + 1, directory)
            old = directory
        else:
            old = None
        k32.SetDllDirectoryW(path)
        yield
        k32.SetDllDirectoryW(old)

    @contextmanager
    def _loader_posix(cls, path):
        lib = "libmx_common.dylib" if sys.platform == "darwin" else "libmxcommon.so"
        link_path = os.path.join(os.path.dirname(__file__), lib)
        lib_path = os.path.join(path, lib)
        if os.path.islink(link_path):
            if os.path.realpath(link_path) == lib_path:
                yield
                return
            os.unlink(link_path)
        os.symlink(lib_path, link_path)
        yield


class _MaxwellMeta(type, metaclass=_MaxwellLoaderMeta):
    def _import_pymaxwell(cls):
        path = get_maxwell_path()
        with cls._loader(path):
            from . import pymaxwell

            pymaxwell.init_extensions(os.path.join(path, "extensions"))

    def __getattr__(cls, name):
        global pymaxwell
        if "pymaxwell" not in globals():
            cls._import_pymaxwell()
        return getattr(getattr(pymaxwell, cls.__name__), name)


class _Maxwell(metaclass=_MaxwellMeta):
    def __new__(cls, *args, **kwargs):
        global pymaxwell
        if "pymaxwell" not in globals():
            cls._import_pymaxwell()
        return getattr(pymaxwell, cls.__name__)(*args, **kwargs)


class Rgb(_Maxwell):
    pass


class Point2D(_Maxwell):
    pass


class Texture(_Maxwell):
    pass


class Attribute(_Maxwell):
    pass


class Scene(_Maxwell):
    pass


class ObjectIterator(_Maxwell):
    pass


class MxiReader(_Maxwell):
    pass


class MxImage(_Maxwell):
    pass


class MxRender(_Maxwell):
    pass


class TexturePreview(_Maxwell):
    pass
