# -*- coding: utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Ildar Nikolaev"
__email__ = "nildar@users.sourceforge.net"

import bpy
from bpy.types import Operator
from bpy.props import StringProperty
from bpy_extras.io_utils import ImportHelper, axis_conversion

from mathutils import Matrix

from . import world
from . import material
from . import texture
from . import object
from . import uvlayout
from . import render


CONVERSION_MATRIX = axis_conversion('-Z', 'Y').to_4x4()
CONVERSION_MATRIX_INVERSE = CONVERSION_MATRIX.inverted()


class Report(Operator):
    bl_idname = "maxwell.report"
    bl_label = "Show UI Message"
    bl_options = {'INTERNAL'}

    _TYPES = {'DEBUG', 'INFO', 'OPERATOR', 'PROPERTY', 'WARNING', 'ERROR',
              'ERROR_INVALID_INPUT', 'ERROR_INVALID_CONTEXT', 'ERROR_OUT_OF_MEMORY'}
    type = bpy.props.EnumProperty(
        items=[(t, t, "") for t in _TYPES],
        options={'ENUM_FLAG'},
        default={'INFO'}
    )
    text = StringProperty()

    def execute(self, context):
        self.report(self.type, self.text)
        return {'FINISHED'}


class ExpandUI(Operator):
    bl_idname = "maxwell.expand_ui"
    bl_label = "Expand UI"
    bl_options = {'INTERNAL'}

    attr = StringProperty(options={'HIDDEN'})
    path = StringProperty(options={'HIDDEN'})

    def execute(self, context):
        attr = self.attr
        obj = context.maxwell_ui.path_resolve(self.path)
        setattr(obj, attr, not getattr(obj, attr))
        return {'FINISHED'}


class ImportMXS(Operator, ImportHelper):
    bl_idname = "maxwell.import_mxs"
    bl_label = "Import (.mxs)"
    bl_description = "Import Maxwell Render scene from MXS file"

    filename_ext = ".mxm"
    filter_glob = StringProperty(default="*.mxs", options={'HIDDEN'})

    def execute(self, context):
        from .. import maxwell

        data = context.blend_data
        scene = context.scene
        objs = {}

        mxs = maxwell.Scene()
        mxs.ReadMXS(self.filepath)

        it = maxwell.ObjectIterator()
        mxobj = it.First(mxs)
        while not mxobj.isNull:
            mesh = None
            name = mxobj.Name
            if mxobj.isMesh:
                nv = mxobj.NumVertexes
                if nv:
                    mesh = data.meshes.new(name)
                    mesh.vertices.add(nv)
                    for i in range(nv):
                        v = mxobj.GetVertex(i)
                        mesh.vertices[i].co = (v[0], -v[2], v[1])

                    vn = {}
                    nt = mxobj.NumTriangles
                    mesh.tessfaces.add(nt)
                    for i in range(nt):
                        v1, v2, v3, n1, n2, n3 = mxobj.GetTriangle(i)
                        vn[v1] = n1
                        vn[v2] = n2
                        vn[v3] = n3
                        f = mesh.tessfaces[i]
                        f.vertices = (v1, v2, v3, 0)
                        f.use_smooth = n1 != n2 or n2 != n3 or n3 != n1

                    nn = mxobj.NumNormals
                    norms = [mxobj.GetNormal(i) for i in range(nn)]
                    for v, n in vn.items():
                        mesh.vertices[v].normal = norms[n]

                    nuv = mxobj.NumChannelsUVW
                    for i in range(nuv):
                        uv_map = mesh.tessface_uv_textures.new()
                        for j in range(nt):
                            u1, v1, w1, u2, v2, w2, u3, v3, w3 = mxobj.GetTriangleUVW(j, i)
                            uv_map.data[j].uv_raw = (u1, -v1, u2, -v2, u3, -v3, 0.0, 0.0)
                    mesh.update()

            obj = data.objects.new(name, mesh)
            scene.objects.link(obj)
            objs[name] = obj

            m = Matrix(mxobj.BaseAndPivot["base"])
            matrix = CONVERSION_MATRIX * m * CONVERSION_MATRIX_INVERSE
            mxp = mxobj.Parent
            if not mxp.isNull:
                obj.parent = objs[mxp.Name]
            obj.matrix_local = matrix

            hide = mxobj.Hide
            obj.hide_render = hide
            obj.hide = hide

            mxobj = it.Next()
        return {'FINISHED'}

    @classmethod
    def register(cls):
        def menu_func(self, context):
            self.layout.operator_context = 'INVOKE_DEFAULT'
            self.layout.operator(cls.bl_idname, text="Maxwell Render (.mxs)")

        bpy.types.INFO_MT_file_import.append(menu_func)
