# -*- coding: utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Ildar Nikolaev"
__email__ = "nildar@users.sourceforge.net"

import os
import sys
import subprocess

import bpy
from bpy.types import Operator
from bpy.props import StringProperty, BoolProperty, EnumProperty
from bl_operators.presets import AddPresetBase
from bpy_extras.io_utils import ExportHelper

from ..engine import abspath, get_maxwell_path

if sys.platform == "win32":
    _MXED = "mxed.exe"
elif sys.platform == "darwin":
    _MXED = "Mxed.app/Contents/MacOS/Mxed"
else:
    _MXED = "mxed"


def get_filepath(ctx, path):
    """ depending on the settings return relative or absolute path """
    if path and ctx.user_preferences.filepaths.use_relative_paths:
        path = bpy.path.relpath(path)
    return path


#
# MXM
#

class EditMXM(Operator):
    bl_idname = "maxwell.material_edit_mxm"
    bl_label = "Editor"
    bl_description = "Edit MXM file in Maxwell Material Editor"

    def execute(self, context):
        try:
            executable = get_maxwell_path(_MXED)
            subprocess.call([executable, "-mxm:" + abspath(context.material.maxwell.mxm)])
        except Exception as ex:
            self.report(type={'WARNING'}, message="Exception: {0}".format(ex))
        return {'FINISHED'}


class BrowseMXM(Operator):
    bl_idname = "maxwell.material_browse_mxm"
    bl_label = "Browser"
    bl_description = "Browse and select material from Maxwell Material Browser"

    def execute(self, context):
        try:
            executable = get_maxwell_path(_MXED)
            ret = subprocess.check_output([executable, "-brwclose"])
            ret = ret.decode().strip().split(":", 1)
            if len(ret) == 2 and ret[0] == "MXM":
                context.material.maxwell.mxm = get_filepath(context, ret[1])
        except Exception as ex:
            self.report(type={'WARNING'}, message="Exception: {0}".format(ex))
        return {'FINISHED'}


class SaveMXM(Operator, ExportHelper):
    bl_idname = "maxwell.material_save_mxm"
    bl_label = "Save as MXM file"
    bl_description = "Save material as maxwell MXM file"

    filename_ext = ".mxm"
    filter_glob = StringProperty(default="*.mxm", options={'HIDDEN'})
    make_linked = BoolProperty(name="Make Linked")
    embed = BoolProperty(name="Embedded")

    def execute(self, context):
        filepath = self.filepath
        mat = context.material
        mx = mat.maxwell
        try:
            from ..export.material import Materials

            mats = Materials()
            mxm = mats.export(mat)
            if mx.type not in ('file', 'custom'):
                # doesn't save without it
                mxm.AddLayer().AddBSDF()
            mxm.Write(filepath)

            if self.make_linked:
                mx.mxm = get_filepath(context, filepath)
                mx.embed = self.embed
                mx.type = 'file'
        except Exception as ex:
            self.report(type={'WARNING'}, message="Exception: {0}".format(ex))
        return {'FINISHED'}

    def draw(self, context):
        layout = self.layout
        row = layout.row()
        row.prop(self, "make_linked")
        row = layout.row()
        row.prop(self, "embed")
        row.enabled = self.make_linked


class ReadMXM(Operator):
    bl_idname = "maxwell.material_read_mxm"
    bl_label = "Make Custom"
    bl_description = "Read material from MXM file and make it Custom"

    def execute(self, context):
        from .. import maxwell

        mxs = maxwell.Scene()

        mat = getattr(context, "material", None)
        if mat is None:
            mat = context.object.active_material
        mxm_path = mat.maxwell.mxm
        embed = mat.maxwell.embed
        mat["maxwell"] = {}  # fast clear

        mx_mat = mat.maxwell
        mx_mat.mxm = mxm_path
        mx_mat.embed = embed

        mxm_path = abspath(mxm_path)

        mxm_dir = os.path.dirname(mxm_path)
        mxs.AddSearchingPath(mxm_dir)

        mat_db = os.environ.get("MAXWELL3_MATERIALS_DATABASE")
        if mat_db and mat_db != mxm_dir:
            mxs.AddSearchingPath(mat_db)
            mxs.AddSearchingPath(os.path.join(mat_db, "textures"))
            mxs.AddSearchingPath(os.path.join(mat_db, "ior files"))
        if context.scene:
            for p in context.scene.maxwell.rp.searching_paths:
                p = abspath(p.name)
                if p:
                    mxs.AddSearchingPath(p)

        def search_file(_f):
            if _f:
                if not os.path.isfile(_f):
                    found_path = mxs.SearchFile(_f)
                    if found_path:
                        _f = found_path
                _f = os.path.abspath(_f)
            return _f

        textures = {}
        for tex in bpy.data.textures:
            if tex.type == 'IMAGE' and tex.image:
                mx_tex = tex.maxwell
                d = {
                    "filepath": abspath(tex.image.filepath),
                    "override": mx_tex.override,
                    "channel": mx_tex.channel,
                    "tile_x": mx_tex.tile_x,
                    "tile_y": mx_tex.tile_y,
                    "units": mx_tex.units,
                    "repeat": mx_tex.repeat[:],
                    "mirror_x": mx_tex.mirror_x,
                    "mirror_y": mx_tex.mirror_y,
                    "offset": mx_tex.offset[:],
                    "rotation": mx_tex.rotation,
                    "flip_x": mx_tex.flip_x,
                    "flip_y": mx_tex.flip_y,
                    "wide": mx_tex.wide,
                    "invert": mx_tex.invert,
                    "alpha_only": mx_tex.alpha_only,
                    "interpolation": mx_tex.interpolation,
                    "brightness": mx_tex.brightness,
                    "saturation": mx_tex.saturation,
                    "contrast": mx_tex.contrast,
                    "hue": mx_tex.hue,
                    "clamp": mx_tex.clamp[:]
                }
                s = str(sorted(d.items()))
                textures[s] = tex

        for i, t in enumerate(mat.texture_slots):
            if t:
                mat.texture_slots.clear(i)

        def add_texture_slot(_tex):
            slot = mat.texture_slots.add()
            slot.texture = _tex

        def get_texture(mxt):
            if mxt.isEmpty:
                return ""
            mxt_params = {
                "filepath": search_file(mxt.path),
                "override": mxt.globalMap,
                "channel": mxt.channel,
                "tile_x": mxt.vIsTiled,
                "tile_y": mxt.uIsTiled,
                "units": '1' if mxt.absolute else '0',
                "repeat": mxt.scale[:],
                "mirror_x": mxt.vIsMirrored,
                "mirror_y": mxt.uIsMirrored,
                "offset": mxt.offset[:],
                "rotation": mxt.rotation,
                "flip_x": mxt.flipRed,
                "flip_y": mxt.flipGreen,
                "wide": mxt.fullRange,
                "invert": mxt.invert,
                "alpha_only": mxt.alpha,
                "interpolation": mxt.interpolation,
                "brightness": mxt.brightness * 100.0,
                "saturation": mxt.saturation * 100.0,
                "contrast": mxt.contrast * 100.0,
                "hue": mxt.hue * 100.0,
                "clamp": (int(mxt.clampMin * 255), int(mxt.clampMax * 255))
            }
            s_key = str(sorted(mxt_params.items()))
            bl_tex = textures.get(s_key)
            if bl_tex:
                _name = bl_tex.name
                idx = mat.texture_slots.find(_name)
                if idx < 0:
                    add_texture_slot(bl_tex)
            else:
                filepath = mxt_params.pop("filepath")
                filename = os.path.basename(filepath)
                tex_name, ext = os.path.splitext(filename)
                bl_tex = bpy.data.textures.new(tex_name, 'IMAGE')
                filepath = get_filepath(context, filepath)
                try:
                    bl_tex.image = bpy.data.images.load(filepath)
                except Exception as ex:
                    self.report(type={'WARNING'}, message="Exception: {0}:".format(ex))
                bl_tex["maxwell"] = mxt_params
                textures[s_key] = bl_tex
                add_texture_slot(bl_tex)
                _name = bl_tex.name
            return _name

        def get_attribute(mxm_attr, attr):
            _name = get_texture(mxm_attr.texture)
            return _name, getattr(mxm_attr, attr)

        mxm = mxs.ReadMaterial(mxm_path)
        if not mxm.isNull:
            m = mxm.GlobalMap
            mx_map = mx_mat.override_map
            mx_map.channel = m.channel
            mx_map.tile_x = m.vIsTiled
            mx_map.tile_y = m.uIsTiled
            mx_map.units = '1' if m.absolute else '0'
            mx_map.mirror_x = m.vIsMirrored
            mx_map.mirror_y = m.uIsMirrored
            mx_map.repeat = m.scale
            mx_map.offset = m.offset
            mx_map.rotation = m.rotation

            mx_mat.bump_tex, mx_mat.bump = get_attribute(mxm.Bump, "value")
            mx_mat.bump_normal = mxm.NormalMapState

            mx_mat.matte = mxm.Matte
            mx_mat.shadow = mxm.MatteShadow
            mx_mat.dispersion = mxm.Dispersion
            mx_mat.color_id = mxm.ColorID

            # Displacement
            dp = mx_mat.displacement
            mxm_dp = mxm.Displacement
            dp.enabled = mxm_dp.Enabled
            dp.map = get_texture(mxm_dp.Map)
            # parameters
            p = mxm_dp.Parameters
            dp.type = str(p["type"])
            dp.level = p["level"]
            dp.smooth = p["smooth"]
            dp.offset = p["offset"]
            dp.method = str(p["method"])
            dp.uvi = str(p["uvi"])
            #  height map
            m = mxm_dp.HeightMap
            dp.absolute = m["absolute"]
            dp.height = m["height"]
            if dp.absolute:
                dp.height *= 100
            dp.adaptive = m["adaptive"]
            #  vector
            v = mxm_dp.Vector
            dp.scale = v["scale"]
            dp.transform = str(v["transform"])
            dp.mapping = str(v["mapping"])
            dp.preset = str(v["preset"])

            layers = mx_mat.layers
            for n_layer in range(mxm.NumLayers):
                mxm_layer = mxm.GetLayer(n_layer)
                layer = layers.add()
                layer.name = mxm_layer.Name
                layer.enabled = mxm_layer.Enabled
                layer.blend = str(mxm_layer.Blending)
                layer.opacity_tex, layer.opacity = get_attribute(mxm_layer.Opacity, "value")

                # Emitter
                mxm_emitter = mxm_layer.GetEmitter()
                if not mxm_emitter.isNull:
                    emitter = layer.emitter
                    emitter.enabled = mxm_emitter.Enabled
                    emitter.type = str(mxm_emitter.Type)
                    #  lobe
                    lobe = mxm_emitter.Lobe
                    emitter.lobe = str(lobe.Type)
                    ies = lobe.IES
                    if ies:
                        ies = get_filepath(context, search_file(lobe.IES))
                    emitter.ies = ies
                    emitter.intensity = lobe.Intensity
                    emitter.image = get_texture(lobe.Image)
                    #  temperature
                    emitter.temperature = mxm_emitter.Temperature
                    #  hdri
                    emitter.hdri_tex, emitter.hdri_intensity = get_attribute(mxm_emitter.MXI, "value")
                    #  pair
                    p = mxm_emitter.Pair
                    t, luminance = mxm_emitter.ActivePair
                    emitter.color_as_temp = (t == 2)
                    emitter.luminance = str(luminance)
                    if luminance == 0:
                        emitter.power = p["watts"]
                        emitter.efficacy = p["efficacy"]
                    else:
                        emitter.output = p["luminous"]
                    emitter.color = p["rgb"]
                    emitter.color_temp = p["temperature"]

                bsdfs = layer.bsdfs
                for n_bsdf in range(mxm_layer.NumBSDFs):
                    mxm_bsdf = mxm_layer.GetBSDF(n_bsdf)
                    bsdf = bsdfs.add()

                    bsdf.name = mxm_bsdf.Name
                    bsdf.enabled = mxm_bsdf.Enabled
                    bsdf.weight_tex, bsdf.weight = get_attribute(mxm_bsdf.Weight, "value")

                    refl = mxm_bsdf.Reflectance
                    attrs = refl.Attributes

                    # BSDF Properties
                    ior = refl.ComplexIor
                    if ior:
                        ior = get_filepath(context, search_file(ior))
                    bsdf.ior_file = ior
                    bsdf.ior_type = '1' if refl.ActiveIorMode else '0'
                    bsdf.refl_0_tex, bsdf.refl_0 = get_attribute(attrs["color"], "rgb")
                    bsdf.refl_90_tex, bsdf.refl_90 = get_attribute(attrs["color.tangential"], "rgb")
                    bsdf.trans_tex, bsdf.trans = get_attribute(attrs["transmittance.color"], "rgb")
                    d = refl.AbsorptionDistance
                    bsdf.atten_units = str(d[0])
                    bsdf.attenuation = d[1]
                    bsdf.nd, bsdf.abbe = refl.IOR
                    bsdf.k = refl.Conductor
                    bsdf.force_fresnel = refl.ForceFresnel
                    bsdf.cf_angle, bsdf.cf_roughness, bsdf.cf_enabled = refl.FresnelCustom

                    # Surface Properties
                    bsdf.roughness_tex, bsdf.roughness = get_attribute(mxm_bsdf.Roughness, "value")
                    bsdf.anisotropy_tex, bsdf.anisotropy = get_attribute(mxm_bsdf.Anisotropy, "value")
                    bsdf.angle_tex, bsdf.angle = get_attribute(mxm_bsdf.Angle, "value")
                    bsdf.bump_tex, bsdf.bump = get_attribute(mxm_bsdf.Bump, "value")
                    bsdf.bump_normal = mxm_bsdf.NormalMapState

                    # Subsurface Properties
                    bsdf.scattering_tex, bsdf.scattering = get_attribute(attrs["scattering"], "rgb")
                    bsdf.coef, bsdf.asymmetry, bsdf.single_sided = refl.ScatteringParameters
                    tex, val = get_attribute(refl.ScatteringThickness, "value")
                    bsdf.thickness_tex = tex
                    bsdf.thickness = val * 1000.0
                    r = refl.ScatteringThicknessRange
                    bsdf.thickness_min = r[0] * 1000.0
                    bsdf.thickness_max = r[1] * 1000.0

                    # Coating
                    mxm_coating = mxm_bsdf.GetCoating(0)
                    if not mxm_coating.isNull:
                        coating = bsdf.coating
                        coating.enabled = mxm_coating.Enabled

                        tex, val = get_attribute(mxm_coating.Thickness, "value")
                        coating.thickness_tex = tex
                        coating.thickness = val * 1000000000.0

                        thickness_min, thickness_max = mxm_coating.ThicknessRange
                        coating.thickness_min = thickness_min * 1000000000.0
                        coating.thickness_max = thickness_max * 1000000000.0

                        refl = mxm_coating.Reflectance
                        ior = refl.ComplexIor
                        if ior:
                            ior = get_filepath(context, search_file(ior))
                        coating.ior_file = ior
                        coating.ior_type = '1' if refl.ActiveIorMode else '0'
                        coating.nd = refl.IOR[0]
                        coating.k = refl.Conductor
                        coating.force_fresnel = refl.ForceFresnel
                        f = refl.FresnelCustom
                        coating.cf_angle = f[0]
                        coating.cf_enabled = f[2]

                        attrs = refl.Attributes
                        coating.refl_0_tex, coating.refl_0 = get_attribute(attrs["color"], "rgb")
                        coating.refl_90_tex, coating.refl_90 = get_attribute(attrs["color.tangential"], "rgb")
                if bsdfs:
                    layer.active_bsdf_index = 0
            if layers:
                mx_mat.active_layer_index = 0
        mx_mat.type = 'custom'
        return {'FINISHED'}


#
# Material Layer
#

class LayerAdd(Operator):
    bl_idname = "maxwell.material_add_layer"
    bl_label = "Add Layer"
    bl_description = "Add Maxwell Material Layer"

    def execute(self, context):
        mx = context.material.maxwell
        layers = mx.layers
        n = 1
        layer_name = "Layer"
        while layers.find(layer_name) >= 0:
            layer_name = "Layer.%03d" % n
            n += 1
        idx = len(layers)
        layer = layers.add()
        layer["name"] = layer_name
        mx.active_layer_index = idx
        bsdf = layer.bsdfs.add()
        bsdf.name = "BSDF"
        layer.active_bsdf_index = 0
        mx.update_color(context)
        return {'FINISHED'}


class LayerRemove(Operator):
    bl_idname = "maxwell.material_remove_layer"
    bl_label = "Remove Layer"
    bl_description = "Remove Maxwell Material Layer"

    def execute(self, context):
        mx = context.material.maxwell
        if mx.active_layer_index >= 0:
            mx.layers.remove(mx.active_layer_index)
            mx.active_layer_index -= 1
            mx.update_color(context)
        return {'FINISHED'}


class LayerUp(Operator):
    bl_idname = "maxwell.material_up_layer"
    bl_label = "Move Up Layer"
    bl_description = "Move Up Maxwell Material Layer"

    def execute(self, context):
        mx = context.material.maxwell
        idx = mx.active_layer_index
        if idx > 0:
            mx.layers.move(idx - 1, idx)
            mx.active_layer_index -= 1
            mx.update_color(context)
        return {'FINISHED'}


class LayerDown(Operator):
    bl_idname = "maxwell.material_down_layer"
    bl_label = "Move Down Layer"
    bl_description = "Move Down Maxwell Material Layer"

    def execute(self, context):
        mx = context.material.maxwell
        layers = mx.layers
        max_idx = len(layers) - 1
        idx = mx.active_layer_index
        if max_idx and max_idx > idx:
            layers.move(idx + 1, idx)
            mx.active_layer_index += 1
            mx.update_color(context)
        return {'FINISHED'}


class Layer(Operator):
    bl_idname = "maxwell.material_layer"
    bl_label = "Maxwell Material Layer Operator"

    action = EnumProperty(
        items=[
            ('COPY', "Copy Layer", "Copy Material Layer"),
            ('PASTE', "Paste Layer", "Paste Material Layer"),
            ('DUPLICATE', "Duplicate Layer", "Duplicate Material Layer")
        ],
        default='COPY'
    )
    tmp = {}

    def execute(self, context):
        mx = context.material.maxwell
        layer = mx["layers"][mx.active_layer_index]
        if self.action == 'PASTE':
            layer.update(self.tmp)
        else:
            tmp = layer.to_dict()
            del tmp["name"]
            if self.action == 'COPY':
                Layer.tmp = tmp
            elif self.action == 'DUPLICATE':
                bpy.ops.maxwell.material_add_layer()
                mx["layers"][mx.active_layer_index].update(tmp)
        return {'FINISHED'}


#
# Material BSDF
#

class BsdfAdd(Operator):
    bl_idname = "maxwell.material_add_bsdf"
    bl_label = "Add BSDF"
    bl_description = "Add Maxwell Material BSDF"

    def execute(self, context):
        mx = context.material.maxwell
        layer = mx.layers[mx.active_layer_index]
        bsdfs = layer.bsdfs
        n = 1
        bsdf_name = "BSDF"
        while bsdfs.find(bsdf_name) >= 0:
            bsdf_name = "BSDF%d" % n
            n += 1
        idx = len(bsdfs)
        bsdf = bsdfs.add()
        bsdf.name = bsdf_name
        layer.active_bsdf_index = idx
        mx.update_color(context)
        return {'FINISHED'}


class BsdfRemove(Operator):
    bl_idname = "maxwell.material_remove_bsdf"
    bl_label = "Remove BSDF"
    bl_description = "Remove Maxwell Material BSDF"

    def execute(self, context):
        mx = context.material.maxwell
        layer = mx.layers[mx.active_layer_index]
        if layer.active_bsdf_index >= 0:
            layer.bsdfs.remove(layer.active_bsdf_index)
            layer.active_bsdf_index -= 1
            mx.update_color(context)
        return {'FINISHED'}


class BsdfUp(Operator):
    bl_idname = "maxwell.material_up_bsdf"
    bl_label = "Move Up BSDF"
    bl_description = "Move Up Maxwell Material BSDF"

    def execute(self, context):
        mx = context.material.maxwell
        layer = mx.layers[mx.active_layer_index]
        if layer.active_bsdf_index > 0:
            idx = layer.active_bsdf_index
            layer.bsdfs.move(idx - 1, idx)
            layer.active_bsdf_index -= 1
        return {'FINISHED'}


class BsdfDown(Operator):
    bl_idname = "maxwell.material_down_bsdf"
    bl_label = "Move Down BSDF"
    bl_description = "Move Down Maxwell Material BSDF"

    def execute(self, context):
        mx = context.material.maxwell
        layer = mx.layers[mx.active_layer_index]
        max_idx = len(layer.bsdfs) - 1
        idx = layer.active_bsdf_index
        if max_idx and max_idx > idx:
            layer.bsdfs.move(idx + 1, idx)
            layer.active_bsdf_index += 1
        return {'FINISHED'}


class Bsdf(Operator):
    bl_idname = "maxwell.material_bsdf"
    bl_label = "Maxwell Material BSDF Operator"

    action = EnumProperty(
        items=[
            ('COPY', "Copy BSDF", "Copy Material BSDF"),
            ('PASTE', "Paste BSDF", "Paste Material BSDF"),
            ('DUPLICATE', "Duplicate BSDF", "Duplicate Material BSDF")
        ],
        default='COPY'
    )
    tmp = {}

    def execute(self, context):
        mx = context.material.maxwell
        layer = mx.layers[mx.active_layer_index]
        bsdf = layer["bsdfs"][layer.active_bsdf_index]
        if self.action == 'PASTE':
            bsdf.update(self.tmp)
        else:
            tmp = bsdf.to_dict()
            del tmp["name"]
            if self.action == 'COPY':
                Bsdf.tmp = tmp
            elif self.action == 'DUPLICATE':
                bpy.ops.maxwell.material_add_bsdf()
                layer["bsdfs"][layer.active_bsdf_index].update(tmp)
        return {'FINISHED'}


#
# Presets
#

class OpaquePresets(AddPresetBase, Operator):
    bl_idname = "maxwell.material_opaque_preset_add"
    bl_label = "Add Opaque Material Preset"

    preset_menu = "MAXWELL_MT_material_opaque_presets"
    preset_subdir = "bmaxwell/material/opaque"

    preset_defines = [
        "opaque = bpy.context.object.active_material.maxwell.opaque",
    ]

    preset_values = [
        "opaque.color",
        "opaque.color_map_enabled",
        "opaque.shininess",
        "opaque.shininess_map_enabled",
        "opaque.roughness"
        "opaque.roughness_map_enabled",
        "opaque.clearcoat"
    ]


class TransparentPresets(AddPresetBase, Operator):
    bl_idname = "maxwell.material_transparent_preset_add"
    bl_label = "Add Transparent Material Preset"

    preset_menu = "MAXWELL_MT_material_transparent_presets"
    preset_subdir = "bmaxwell/material/transparent"

    preset_defines = [
        "transparent = bpy.context.object.active_material.maxwell.transparent",
    ]

    preset_values = [
        "transparent.color",
        "transparent.color_map_enabled",
        "transparent.ior",
        "transparent.transparency",
        "transparent.roughness"
        "transparent.roughness_map_enabled",
        "transparent.specular_tint",
        "transparent.dispersion",
        "transparent.clearcoat"
    ]


class MetalPresets(AddPresetBase, Operator):
    bl_idname = "maxwell.material_metal_preset_add"
    bl_label = "Add Metal Material Preset"

    preset_menu = "MAXWELL_MT_material_metal_presets"
    preset_subdir = "bmaxwell/material/metal"

    preset_defines = [
        "metal = bpy.context.object.active_material.maxwell.metal",
    ]

    preset_values = [
        "metal.ior",
        "metal.tint",
        "metal.color",
        "metal.color_map_enabled",
        "metal.roughness"
        "metal.roughness_map_enabled",
        "metal.anisotropy",
        "metal.anisotropy_map_enabled",
        "metal.angle",
        "metal.angle_map_enabled",
        "metal.dust",
        "metal.dust_map_enabled"
    ]


class TranslucentPresets(AddPresetBase, Operator):
    bl_idname = "maxwell.material_translucent_preset_add"
    bl_label = "Add Translucent Material Preset"

    preset_menu = "MAXWELL_MT_material_translucent_presets"
    preset_subdir = "bmaxwell/material/translucent"

    preset_defines = [
        "translucent = bpy.context.object.active_material.maxwell.translucent",
    ]

    preset_values = [
        "translucent.scale",
        "translucent.ior",
        "translucent.color",
        "translucent.color_map_enabled",
        "translucent.hue_shift",
        "translucent.invert_hue",
        "translucent.vibrance",
        "translucent.density",
        "translucent.opacity",
        "translucent.roughness",
        "translucent.roughness_map_enabled",
        "translucent.specular_tint",
        "translucent.clearcoat",
        "translucent.clearcoat_ior"
    ]


class CarPaintPresets(AddPresetBase, Operator):
    bl_idname = "maxwell.material_carpaint_preset_add"
    bl_label = "Add Car Paint Material Preset"

    preset_menu = "MAXWELL_MT_material_carpaint_presets"
    preset_subdir = "bmaxwell/material/carpaint"

    preset_defines = [
        "carpaint = bpy.context.object.active_material.maxwell.carpaint",
    ]

    preset_values = [
        "carpaint.color",
        "carpaint.metallic",
        "carpaint.topcoat"
    ]


class HairPresets(AddPresetBase, Operator):
    bl_idname = "maxwell.material_hair_preset_add"
    bl_label = "Add Hair Material Preset"

    preset_menu = "MAXWELL_MT_material_hair_presets"
    preset_subdir = "bmaxwell/material/hair"

    preset_defines = [
        "mx = bpy.context.object.active_material.maxwell",
        "mx['hair'].clear()",
        "hair = mx.hair",
    ]

    preset_values = [
        "hair.color",
        "hair.color_map",
        "hair.color_map_enabled",
        "hair.primary_strength",
        "hair.primary_spread",
        "hair.primary_tint",
        "hair.secondary_strength",
        "hair.secondary_spread",
        "hair.secondary_tint"
    ]
