# -*- coding: utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Ildar Nikolaev"
__email__ = "nildar@users.sourceforge.net"

import io
import time
import struct
import threading

import bpy
import bmesh
from bpy.types import Operator
from bl_operators.presets import AddPresetBase

from .. import maxwell as mx
from ..engine import abspath, Logger
from ..export.material import Materials


class UVProjectionAdd(Operator):
    bl_idname = "maxwell.uv_projection_add"
    bl_label = "Add UV projection"
    bl_description = "Add UV projection"

    def execute(self, context):
        mx = context.object.data.maxwell
        p = mx.projections
        i = len(p)
        p.add()
        mx.active_projection = i
        return {'FINISHED'}


class UVProjectionRemove(Operator):
    bl_idname = "maxwell.uv_projection_remove"
    bl_label = "Remove UV projection"
    bl_description = "Remove UV projection"

    def execute(self, context):
        mx = context.object.data.maxwell
        p = mx.projections
        i = mx.active_projection
        if i >= 0:
            p.remove(i)
            t = len(p)
            if i >= t:
                mx.active_projection = t - 1
        return {'FINISHED'}


class ImportGrass(Operator):
    bl_idname = "maxwell.import_grass"
    bl_label = "Import Grass"

    def execute(self, context):
        scene = context.scene
        ob = context.object

        ps = ob.particle_systems.active
        pss = ps.settings

        us = scene.unit_settings
        scale = us.scale_length if us.system != 'NONE' else 1.0

        mesh = ob.to_mesh(scene, True, 'RENDER')
        try:
            s = mx.Scene()
            s.AxisConversion = "ZXY"
            s.SetSinglePrecisionOfGeometry()

            m = s.ImportMesh(mesh.name, mesh, [])
            m.BaseAndPivot = {"base": ob.matrix_world * scale}

            maps = {}
            mats = Materials(s)
            g = pss.maxwell.grass
            for _map, _name in [(g.density_map, "density"),
                                (g.length_map, "length"),
                                (g.angle_map, "angle"),
                                (g.start_bend_map, "start_bend"),
                                (g.bend_radius_map, "bend_radius"),
                                (g.bend_angle_map, "bend_angle"),
                                (g.cut_off_map, "cut_off")]:
                if _map:
                    _tex = mats.get_texture(pss, _map)
                    if _tex:
                        maps[_name] = _tex
            grass = m.ApplyGrass(g, None, maps)

            # reduce recalculation time when edited_clear() called
            pss.count = 0
            pss.emit_from = 'FACE'
            pss.hair_length = g.length / (scale * 100.0)

            m.ExportGrass(grass, ps, 1.0 / scale)
            ob.update_tag({'OBJECT'})

            bpy.ops.particle.connect_hair()
        finally:
            bpy.data.meshes.remove(mesh)
        return {'FINISHED'}


class GrassPresets(AddPresetBase, Operator):
    bl_idname = "maxwell.particles_grass_preset_add"
    bl_label = "Add Grass Preset"

    preset_menu = "MAXWELL_MT_particles_grass_presets"
    preset_subdir = "bmaxwell/particles/grass"

    preset_defines = [
        "mx = bpy.context.object.particle_systems.active.settings.maxwell",
        "grass = mx.grass",
    ]

    preset_values = [
        "grass.type",
        "grass.blade_points",
        "grass.density",
        "grass.length",
        "grass.length_var",
        "grass.root_width",
        "grass.tip_width",
        "grass.direction",
        "grass.angle",
        "grass.angle_var",
        "grass.start_bend",
        "grass.start_bend_var",
        "grass.bend_radius",
        "grass.bend_radius_var",
        "grass.bend_angle",
        "grass.bend_angle_var",
        "grass.cut_off",
        "grass.cut_off_var"
    ]


class ImportSea(Operator):
    bl_idname = "maxwell.import_sea"
    bl_label = "Import  Sea"
    bl_options = {'INTERNAL'}

    action = bpy.props.EnumProperty(
        items=[
            ('SINGLE_FRAME', "Single frame", "Import single frame"),
            ('FRAMES_RANGE', "Frames range", "Import frames range")
        ],
        default='SINGLE_FRAME'
    )

    def __init__(self):
        self._log = Logger(report=self.report)
        self._break = False
        self._thread = None
        self._event = None
        self._timer = None
        self._verts = None
        self._trigs = None
        self._frame = None
        self.frame_current = 0

    def _bake(self, sea, start, end, quality, mc_file):
        verts = None
        trigs = None
        mcf = None

        n = 2 ** (quality + 2)
        nverts = n * n + n // 2
        nframes = abs(end - start) + 1

        self._log("Sea: verts=%d frames=%d" % (nverts, nframes))
        pcb = time.perf_counter()

        try:
            if mc_file:
                mc_file = abspath(mc_file)
                self._log("Sea: %s" % mc_file)
                mcf = open(mc_file, "wb")
                mcf.write(b'POINTCACHE2\x00')
                mcf.write(struct.pack('iiffi', 1, nverts, start, 1.0, nframes))

            mxs = mx.Scene()
            mxs.AxisConversion = "ZXY"
            mxs.SetSinglePrecisionOfGeometry()
            mxobj = mxs.CreateSea("Sea", sea, quality)
            if self._break:
                return
            i = 1 if start <= end else -1
            for f in range(start, end + i, i):
                self._frame = f

                self._event.wait()
                if self._break:
                    break
                self._event.clear()

                pc = time.perf_counter()

                mxobj.SetGeometryLoaderParams({"Reference Time": sea.time})
                self._log("Sea: t=%f rt=%f f=%d" % (time.perf_counter() - pc, sea.time, f))
                if self._break:
                    break

                buf = mxobj.GetVerticesBuffer()
                self._log("Sea: t=%f buf" % (time.perf_counter() - pc))
                if f == start:
                    verts = memoryview(buf).cast('f').tolist()
                    self._log("Sea: t=%f verts " % (time.perf_counter() - pc))
                    trigs = mxobj.GetTriangles()
                    self._log("Sea: t=%f trigs" % (time.perf_counter() - pc))
                if self._break:
                    break

                if mcf is not None:
                    mcf.write(buf)
                    self._log("Sea: t=%f file" % (time.perf_counter() - pc))

            if not self._break:
                self._verts = verts
                self._trigs = trigs
        finally:
            if mcf is not None:
                mcf.close()
        self._log("Sea: thread t=%f" % (time.perf_counter() - pcb))

    def cancel(self, context):  # called by blender, i.e. on closing
        self._log("Sea: cancel")

        wm = context.window_manager
        wm.event_timer_remove(self._timer)

        self._break = True
        self._event.set()
        self._thread.join()

        context.scene.frame_set(self.frame_current)
        wm.progress_end()

    def modal(self, context, event):
        if event.type == 'ESC':
            self.cancel(context)
            return {'CANCELLED'}

        if self._thread.is_alive():
            if event.type == 'TIMER':
                scene = context.scene
                if self._frame != scene.frame_current:
                    scene.frame_set(self._frame)
                    self._event.set()
                context.window_manager.progress_update(self._frame)
            return {'RUNNING_MODAL'}

        pc = time.perf_counter()

        wm = context.window_manager
        wm.event_timer_remove(self._timer)
        context.scene.frame_set(self.frame_current)

        if self._verts is None or self._trigs is None:
            return {'CANCELLED'}

        mesh = context.mesh
        if mesh is None:
            if context.object.type != 'MESH':
                self._log("Not a mesh", {'WARNING'})
                return {'CANCELLED'}
            mesh = context.object.data

        # fast clear the existing mesh
        bm = bmesh.new()
        bm.verts.new()
        bm.to_mesh(mesh)

        # load new geometry data
        mesh.vertices.add(len(self._verts) // 3 - 1)
        mesh.tessfaces.add(len(self._trigs) // 3)
        mesh.vertices.foreach_set("co", self._verts)
        mesh.tessfaces.foreach_set("vertices", self._trigs)
        mesh.update()

        self._log("Sea: mesh t=%f\n" % (time.perf_counter() - pc))

        if self.action == 'FRAMES_RANGE':
            sea = mesh.maxwell.sea
            mods = context.object.modifiers
            mod = mods.get(sea.mc_modifier)
            if mod is None or mod.type != 'MESH_CACHE':
                mod = mods.new(sea.mc_modifier, 'MESH_CACHE')
                sea.mc_modifier = mod.name
                mod.filepath = sea.mc_file
                mod.cache_format = 'PC2'

        wm.progress_end()
        return {'FINISHED'}

    def execute(self, context):
        wm = context.window_manager
        scene = context.scene
        mesh = context.mesh
        if mesh is None:
            if context.object.type != 'MESH':
                self._log("Not a mesh", {'WARNING'})
                return {'CANCELLED'}
            mesh = context.object.data
        sea = mesh.maxwell.sea

        frame_current = scene.frame_current
        self.frame_current = frame_current
        if self.action == 'SINGLE_FRAME':
            sf = ef = frame_current
            quality = sea.get("quality", 6)
            mc_file = False
        elif self.action == 'FRAMES_RANGE':
            sf, ef = sea.mc_frames
            quality = sea.get("mc_quality", 4)
            mc_file = sea.mc_file
            # save defaults values to rna properties
            sea.mc_quality = str(quality)
            sea.mc_file = mc_file
        else:
            self._log("Invalid action", {'ERROR'})
            return {'CANCELLED'}
        self._frame = sf

        wm.progress_begin(sf, ef + 1)

        self._event = threading.Event()
        self._thread = threading.Thread(
            target=self._bake,
            args=(sea, sf, ef, quality, mc_file)
        )
        if frame_current != sf:
            scene.frame_set(sf)
        self._break = False
        self._thread.start()
        self._event.set()

        self._timer = wm.event_timer_add(0.1, context.window)

        wm.modal_handler_add(self)
        return {'RUNNING_MODAL'}
