# -*- coding: utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Ildar Nikolaev"
__email__ = "nildar@users.sourceforge.net"

import bpy
from bpy.types import Operator


class BakeTextureAdd(Operator):
    bl_idname = "maxwell.bake_add_texture"
    bl_label = "Add texture"
    bl_description = "Add texture for baking"

    def execute(self, context):
        mx = context.texture.maxwell
        idx = len(mx.bake_textures)
        mx.bake_textures.add()
        mx.active_bt_index = idx
        return {'FINISHED'}


class BakeTextureRemove(Operator):
    bl_idname = "maxwell.bake_remove_texture"
    bl_label = "Remove texture"
    bl_description = "Remove texture from baking"

    def execute(self, context):
        mx = context.texture.maxwell
        mx.bake_textures.remove(mx.active_bt_index)
        if mx.active_bt_index >= 0:
            mx.active_bt_index -= 1
        return {'FINISHED'}


class BakeTextures(Operator):
    bl_idname = "maxwell.bake_textures"
    bl_label = "Bake textures"
    bl_description = "Bake textures"

    def execute(self, context):
        bake_mesh = None
        bake_obj = None
        bake_mat = None
        mat_slot = None

        data = context.blend_data
        scenes = data.scenes

        obj = context.object
        tex = context.texture
        mx = tex.maxwell

        BAKE_NAME = ".temp_maxwell_bake_texture"

        bake_scene = scenes.get(BAKE_NAME)
        if not bake_scene:
            bake_scene = scenes.new(BAKE_NAME)
        bake_scene.render.bake_type = 'TEXTURE'
        bake_scene.render.use_bake_clear = mx.bake_clear

        try:
            bake_mesh = obj.to_mesh(context.scene, True, 'RENDER')
            bake_mesh.name = BAKE_NAME

            bake_obj = data.objects.new(BAKE_NAME, bake_mesh)
            bake_scene.objects.link(bake_obj)
            bake_obj.select = True
            bake_scene.objects.active = bake_obj

            bake_mat = context.material.copy()
            bake_mat.name = BAKE_NAME
            bake_mat.diffuse_color = mx.bake_color

            mi = obj.active_material_index
            for i, slot in enumerate(bake_obj.material_slots):
                if i == mi:
                    slot.material = bake_mat
                    mat_slot = slot
                else:
                    slot.material = None

            for i, slot in enumerate(bake_mat.texture_slots):
                if slot and slot.use:
                    bt = mx.bake_textures.get(slot.name)
                    if bt and bt.enabled:
                        bake_mat.use_textures[i] = True
                        slot.use_map_color_diffuse = True
                    else:
                        bake_mat.use_textures[i] = False
                        slot.use = False

            img = tex.image
            uv_map = bake_mesh.uv_textures[mx.channel]
            for uv in uv_map.data:
                uv.image = img
            uv_map.active = True
            uv_map.active_render = True

            bake_mesh.update()

            ctx = {
                "blend_data": data,
                "window": context.window,
                "region": context.region,
                "screen": context.screen,
                "area": context.area,
                "scene": bake_scene,
                "object": bake_obj,
                "edit_object": []
            }
            bpy.ops.object.bake_image(ctx)

            if mx.bake_save:
                img.save()
        finally:
            if mat_slot:
                mat_slot.material = None
            if bake_mat:
                data.materials.remove(bake_mat)
            if bake_obj:
                bake_scene.objects.unlink(bake_obj)
                data.objects.remove(bake_obj)
            if bake_mesh:
                data.meshes.remove(bake_mesh)
            scenes.remove(bake_scene)
        return {'FINISHED'}
