# -*- coding: utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Ildar Nikolaev"
__email__ = "nildar@users.sourceforge.net"

import os
import sys
import subprocess

import bpy
from bpy.types import Operator

from ..engine import abspath
from .. import BMaxwellPreferences


def export_uvmap(mesh):
    uvs = {}
    uv_data = mesh.uv_layers.active.data
    uv_file = abspath(mesh.headus_uvlayout.uv_file)
    with open(uv_file, "w") as file:
        for v in mesh.vertices:
            file.write("v %f %f %f\n" % (v.co[0], -v.co[2], v.co[1]))
        for d in uv_data:
            t = d.uv[:]
            if not t in uvs:
                file.write("vt %f %f\n" % t)
                uvs[t] = len(uvs)
        for p in mesh.polygons:
            line = ""
            for i in p.loop_indices:
                loop = mesh.loops[i]
                v1 = loop.vertex_index + 1
                v2 = uvs[uv_data[loop.index].uv[:]] + 1
                line += " %d/%d" % (v1, v2)
            if line:
                file.write("f%s\n" % line)
    return uv_file


def run_uvlayout(mode, uv_file):
    uvlayout_path = abspath(BMaxwellPreferences.get_preferences().uvlayout_path)

    args = "-plugin,Poly"
    if mode == 'EDIT':
        args += ",Edit"
    elif mode == 'NEW':
        args += ",New"

    if sys.platform == "darwin":
        # TODO: Inspect
        executable = "Contents/MacOS/uvlayout-plugin"
        t = os.path.join(uvlayout_path, "Contents/MacOS/uvlayout-plugin")
        if not os.path.isfile(t):
            executable = t
        else:
            t = os.path.join(uvlayout_path, "headus-UVLayout-Pro.app", executable)
            if not os.path.isfile(t):
                return {'CANCELLED'}
            executable = t
        subprocess.Popen([executable, uv_file])
    else:
        if sys.platform == "win32":
            executable = "uvlayout.exe"
        else:
            executable = "uvlayout"
        executable = os.path.join(uvlayout_path, executable)
        subprocess.Popen([executable, args, uv_file])


class UVExport(Operator):
    bl_idname = "headus_uvlayout.export"
    bl_label = "Export"
    bl_description = "Export selected UV Map to the file"
    bl_options = {'INTERNAL'}

    @classmethod
    def poll(cls, context):
        mesh = context.mesh
        return mesh and mesh.uv_layers.active

    def execute(self, context):
        if export_uvmap(context.mesh):
            self.report({'INFO'}, "UV Map exported to the file.")
        return {'FINISHED'}


class UVNew(Operator):
    bl_idname = "headus_uvlayout.new"
    bl_label = "New"
    bl_description = "Create new UV map in UVLayout"
    bl_options = {'INTERNAL'}
    
    @classmethod
    def poll(cls, context):
        mesh = context.mesh
        return mesh and mesh.uv_layers.active

    def execute(self, context):
        try:
            uv_file = export_uvmap(context.mesh)
            if uv_file:
                run_uvlayout('NEW', uv_file)
        except Exception as ex:
            self.report(type={'WARNING'}, message="Exception: {0}".format(ex))
        return {'FINISHED'}


class UVEdit(Operator):
    bl_idname = "headus_uvlayout.edit"
    bl_label = "Edit"
    bl_description = "Export selected UV Map to the file and open it in UVLayout"
    bl_options = {'INTERNAL'}
    
    @classmethod
    def poll(cls, context):
        mesh = context.mesh
        return mesh and mesh.uv_layers.active

    def execute(self, context):
        try:
            uv_file = export_uvmap(context.mesh)
            if uv_file:
                run_uvlayout('EDIT', uv_file)
        except Exception as ex:
            self.report(type={'WARNING'}, message="Exception: {0}".format(ex))
        return {'FINISHED'}


class UVUpdate(bpy.types.Operator):
    bl_idname = "headus_uvlayout.update"
    bl_label = "Update"
    bl_description = "Update selected UV Map from the existing file"

    @classmethod
    def poll(cls, context):
        mesh = context.mesh
        return mesh and mesh.uv_layers.active

    def invoke(self, context, event):
        mesh = context.mesh
        uv_file = abspath(mesh.headus_uvlayout.uv_file)
        uv_file = "{0[0]}.out".format(os.path.splitext(uv_file))

        verts = []
        faces = []
        with open(uv_file, 'rb') as file:
            for line in file:
                data = line.split()
                if data:
                    s = data[0]
                    if s == b'vt':
                        v = float(data[1]), float(data[2])
                        verts.append(v)
                    elif s == b'f':
                        f = [t.split(b'/') for t in data[1:]]
                        faces.append(f)

        # TODO: Add validation for number of vertices and faces in uv
        uv_map = mesh.uv_layers.active.data
        for i, polygon in enumerate(mesh.polygons):
            for j, loop_index in enumerate(polygon.loop_indices):
                k = int(faces[i][j][1])
                loop = mesh.loops[loop_index]
                uv_map[loop.index].uv = verts[k - 1]
        return {'FINISHED'}
