# -*- coding: utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
# This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Ildar Nikolaev"
__email__ = "nildar@users.sourceforge.net"

import bpy
from bpy.types import PropertyGroup, Material
from bpy.props import (
    BoolProperty,
    FloatProperty,
    StringProperty,
    EnumProperty,
    FloatVectorProperty,
    PointerProperty,
    CollectionProperty,
    IntProperty
)
from mathutils import Color
from math import log, pow

from .texture import Projection


_clamp = lambda v: 0.0 if v < 0.0 else 1.0 if v > 1.0 else v


def k2c(t):
    """Convert color temperature to RGB
    #1 method: http://wiki.blender.org/index.php/User:Juicyfruit/Color_Temperature
    #2 method: http://www.tannerhelland.com/4435/convert-temperature-rgb-algorithm-code/
    http://www.dpreview.com/forums/thread/1219161#forumsPostDetail13310284
    """
    if 1:
        if t <= 6500.0:
            r = 1.0
            g = -2902.1955373783176 / (t + 1669.5803561666639) + 1.3302673723350029
            b = -8257.7997278925690 / (t + 2575.2827530017594) + 1.8993753891711275
        else:
            r = 1745.0425298314172 / (t + -2666.3474220535695) + 0.55995389139931482
            g = 1216.6168361476490 / (t + -2173.1012343082230) + 0.70381203140554553
            b = -8257.7997278925690 / (t + 2575.2827530017594) + 1.8993753891711275
    else:
        k = t / 100
        if k <= 66:
            r = 1.0
            g = (99.4708025861 * log(k) - 161.1195681661) / 255
            if k <= 19:
                b = 0.0
            else:
                b = (138.5177312231 * log(k - 10) - 305.0447927307) / 255
        else:
            k -= 60
            r = (329.698727446 * pow(k, -0.1332047592)) / 255
            g = (288.1221695283 * pow(k, -0.0755148492)) / 255
            b = 1.0
    return Color([_clamp(r), _clamp(g), _clamp(b)])


def _color_sync(self, context):
    mat = self.id_data
    mx = mat.maxwell
    if mx.auto_color:
        clr = getattr(self, "color", None)
        if clr is None:
            t = getattr(self, "type", 'custom')
            if t == 'file':
                return
            elif t == 'custom':
                clr = Color()
                x = 1.0
                for layer in mx.layers:
                    if layer.enabled:
                        t = []
                        weight = 0
                        for bsdf in layer.bsdfs:
                            if bsdf.enabled:
                                weight += bsdf.weight
                                t.append((bsdf.weight, bsdf.refl_0))
                        if weight > 0:
                            lc = Color()
                            for w, c in t:
                                lc += c * w / weight
                            y = x * layer.opacity / 100
                            clr += lc * y
                            x -= y
                            if x <= 0.0:
                                break
                if x == 1.0:
                    return
            else:
                clr = getattr(self, t).color
        mat.diffuse_color = clr


class MaxwellCoatingProperties(PropertyGroup):
    expanded = BoolProperty(
        options={'HIDDEN'}
    )
    enabled = BoolProperty(
        name="Enabled Coating"
    )
    thickness = FloatProperty(
        name="Thickness",
        description="Thickness (nanometers)",
        default=500.0,
        min=1.0, max=1000000.0,
        soft_min=1.0, soft_max=1000000.0,
        precision=3,
        step=1000
    )
    thickness_tex = StringProperty(
        name="Thickness Texture"
    )
    thickness_min = FloatProperty(
        name="Min",
        description="Define the minimum virtual thickness (nanometers)",
        default=100.0,
        min=1.0, max=1000000.0,
        soft_min=1.0, soft_max=1000000.0,
        precision=3,
        step=1000
    )
    thickness_max = FloatProperty(
        name="Max",
        description="Define the maximum virtual thickness (nanometers)",
        default=1000.0,
        min=1.0, max=1000000.0,
        soft_min=1.0, soft_max=1000000.0,
        precision=3,
        step=1000
    )
    ior_type = EnumProperty(
        name="IOR",
        items=[
            ('0', "Custom", "Custom", 0),
            ('1', "Measured", "Measured Data", 1)
        ],
        default='0'
    )
    ior_file = StringProperty(
        name="File",
        subtype='FILE_PATH'
    )
    refl_0 = FloatVectorProperty(
        name="Reflectance 0",
        default=(0.6, 0.6, 0.6),
        min=0.0, max=1.0,
        soft_min=0.0, soft_max=1.0,
        subtype='COLOR'
    )
    refl_0_tex = StringProperty(
        name="Reflectance 0"
    )
    refl_90 = FloatVectorProperty(
        name="Reflectance 90",
        default=(1.0, 1.0, 1.0),
        min=0.0, max=1.0,
        soft_min=0.0, soft_max=1.0,
        subtype='COLOR'
    )
    refl_90_tex = StringProperty(
        name="Reflectance 90"
    )
    nd = FloatProperty(
        name="Nd",
        description="The ND determines the index of refraction of the transparent material",
        default=3.0,
        min=0.01, max=1000.0,
        soft_min=0.01, soft_max=1000.0,
        precision=3,
        step=100
    )
    force_fresnel = BoolProperty(
        name="Force Fresnel",
        default=False
    )
    k = FloatProperty(
        name="K",
        default=0.0,
        min=0.0, soft_min=0.0,
        max=1000.0, soft_max=1000.0,
        precision=3,
        step=100
    )
    cf_enabled = BoolProperty(
        name="Custom Fresnel",
        default=False
    )
    cf_angle = FloatProperty(
        name="Angle",
        default=75.0,
        min=0.0, max=90.0,
        soft_min=0.0, soft_max=90
    )


class MaxwellBSDFProperties(PropertyGroup):
    # if we need unique names for bsdfs in layer
    #def update_name(self, context):
    #	mx = context.material.maxwell
    #	layer = mx.layers[mx.active_layer_index]
    #	bsdfs = layer.bsdfs
    #	if bsdfs.keys().count(self.name) > 1:
    #		n = 1
    #		bsdf_name = "BSDF"
    #		while bsdfs.find(bsdf_name) >= 0:
    #			bsdf_name = "BSDF.%03d" % n
    #			n += 1
    #		self["name"] = bsdf_name

    enabled = BoolProperty(
        name="Enabled",
        default=True,
        update=_color_sync
    )
    weight = FloatProperty(
        name="Weight",
        default=100.0,
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0,
        precision=1,
        update=_color_sync
    )
    weight_tex = StringProperty(
        name="Weight"
    )
    # BSDF Properties
    bsdf_expanded = BoolProperty(
        default=True,
        options={'HIDDEN'}
    )
    ior_type = EnumProperty(
        name="IOR",
        items=[
            ('0', "Custom", "Custom", 0),
            ('1', "Measured", "Measured Data", 1)
        ],
        default='0'
    )
    ior_file = StringProperty(
        name="File",
        subtype='FILE_PATH'
    )
    refl_0 = FloatVectorProperty(
        name="Reflectance 0",
        default=(0.6, 0.6, 0.6),
        min=0.0, max=1.0,
        soft_min=0.0, soft_max=1.0,
        subtype='COLOR',
        update=_color_sync
    )
    refl_0_tex = StringProperty(
        name="Reflectance 0"
    )
    refl_90 = FloatVectorProperty(
        name="Reflectance 90",
        default=(1.0, 1.0, 1.0),
        min=0.0, max=1.0,
        soft_min=0.0, soft_max=1.0,
        subtype='COLOR'
    )
    refl_90_tex = StringProperty(
        name="Reflectance 90"
    )
    trans = FloatVectorProperty(
        name="Transmittance",
        min=0.0, max=1.0,
        soft_min=0.0, soft_max=1.0,
        subtype='COLOR'
    )
    trans_tex = StringProperty(
        name="Transmittance"
    )
    attenuation = FloatProperty(
        name="Attenuation",
        default=1.0,
        min=0.0, max=999.0,
        soft_min=0.0, soft_max=999.0
    )
    atten_units = EnumProperty(
        name="Units",
        items=[
            ('0', "nm", "Nanometers", 0),
            ('1', "um", "Microns", 1),
            ('2', "mm", "Millimeters", 2),
            ('3', "m", "Meters", 3),
            ('4', "cm", "Centimeters", 4),
            ('5', "dm", "Decimeters", 5)
        ],
        default='0'
    )
    nd = FloatProperty(
        name="Nd",
        description="The ND determines the index of refraction of the transparent material",
        default=3.0,
        min=0.01, max=1000.0,
        soft_min=0.01, soft_max=1000.0,
        precision=3,
        step=100
    )
    force_fresnel = BoolProperty(
        name="Force Fresnel"
    )
    k = FloatProperty(
        name="K",
        default=0.0,
        min=0.0, max=1000.0,
        soft_min=0.0, soft_max=1000.0,
        precision=3,
        step=100
    )
    abbe = FloatProperty(
        name="Abbe",
        description="Abbe (dispersion). Dispersion must be enabled in the global 'Maxwell Material' panel",
        default=50.0
    )
    cf_enabled = BoolProperty(
        name="Custom Fresnel"
    )
    cf_angle = FloatProperty(
        name="Angle",
        default=75.0,
        min=0.0, max=90.0,
        soft_min=0.0, soft_max=90.0
    )
    cf_roughness = FloatProperty(
        name="Roughness",
        default=0.0,
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0
    )
    # Surface Properties
    surf_expanded = BoolProperty(
        options={'HIDDEN'}
    )
    roughness = FloatProperty(
        name="Roughness",
        default=100.0,
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0
    )
    roughness_tex = StringProperty(
        name="Roughness"
    )
    bump_tex = StringProperty(
        name="Bump"
    )
    bump = FloatProperty(
        name="Bump",
        default=30.0,
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0
    )
    bump_normal = BoolProperty(
        name="Bump"
    )
    anisotropy = FloatProperty(
        name="Anisotropy",
        default=0.0,
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0
    )
    anisotropy_tex = StringProperty(
        name="Anisotropy"
    )
    angle = FloatProperty(
        name="Angle",
        default=0.0,
        min=0.0, max=360.0,
        soft_min=0.0, soft_max=360.0
    )
    angle_tex = StringProperty(
        name="Angle"
    )
    # Subsurface Properties
    subs_expanded = BoolProperty(
        options={'HIDDEN'}
    )
    scattering = FloatVectorProperty(
        name="Scattering",
        default=(0.5, 0.5, 0.5),
        min=0.0, max=1.0,
        soft_min=0.0, soft_max=1.0,
        subtype='COLOR'
    )
    scattering_tex = StringProperty(
        name="Scattering Texture"
    )
    coef = FloatProperty(
        name="Coef",
        default=0.0,
        min=0.0, max=99999.0,
        soft_min=0.0, soft_max=99999.0
    )
    asymmetry = FloatProperty(
        name="Asymmetry",
        default=0.0,
        min=-1.0, max=1.0,
        soft_min=-1.0, soft_max=1.0,
        precision=3
    )
    single_sided = BoolProperty(
        name="Single Sided"
    )
    thickness = FloatProperty(
        name="Thickness",
        description="Thickness (mm)",
        default=1.0,
        min=0.001, max=10.0,
        soft_min=0.001, soft_max=10.0,
        precision=3
    )
    thickness_tex = StringProperty(
        name="Thickness Texture"
    )
    thickness_min = FloatProperty(
        name="Min",
        description="Define the minimum virtual thickness (mm)",
        default=0.001,
        min=0.001, max=10.0,
        soft_min=0.001, soft_max=10.0,
        precision=3
    )
    thickness_max = FloatProperty(
        name="Max",
        description="Define the maximum virtual thickness (mm)",
        default=10.0,
        min=0.001, max=10.0,
        soft_min=0.001, soft_max=10.0,
        precision=3
    )
    coating = PointerProperty(
        name="Coating",
        type=MaxwellCoatingProperties
    )


class MaxwellEmitterProperties(PropertyGroup):
    expanded = BoolProperty(
        description="Show/Hide Emitter properties",
        options={'HIDDEN'}
    )
    enabled = BoolProperty(
        description="Enable Emitter",
        default=False
    )
    type = EnumProperty(
        name="Emission",
		description="Emission",
        items=[
            ('0', "Color", "", 0),
            ('1', "Temperature", "", 1),
            ('2', "HDR Image", "", 2)
        ],
        default='0'
    )

    # Color
    def _get_color(self):
        return k2c(self.color_temp)

    color = FloatVectorProperty(
        name="Color",
        default=(1.0, 1.0, 1.0),
        min=0.0, max=1.0,
        soft_min=0.0, soft_max=1.0,
        subtype='COLOR'
    )
    color_temp = FloatProperty(
        name="Temperature",
        description="Correlated temperature color",
        default=6500.0,
        min=273.0, max=100000.0,
        soft_min=273.0, soft_max=100000.0,
        precision=1,
        step=1000
    )
    color_as_temp = BoolProperty(
        name="K"
    )
    color_tmp = FloatVectorProperty(
        name="Color",
        min=0.0, max=1.0,
        soft_min=0.0, soft_max=1.0,
        subtype='COLOR',
        options={'HIDDEN'},
        get=_get_color
    )
    # Lobe
    lobe = EnumProperty(
        name="Type",
		description="Lobe type",
        items=[
            ('0', "Area", "", 0),
            ('1', "IES", "", 1),
            ('2', "Spot", "", 2)
        ],
        default='0'
    )
    # Lobe.Luminance
    luminance = EnumProperty(
        name="Luminance",
        items=[
            ('0', "Power & Efficacy", "Power & Efficacy", 0),
            ('1', "Lumen", "Lumen", 1),
            ('2', "Lux", "Lux", 2),
            ('3', "Candela", "Candela", 3),
            ('4', "Luminance", "Luminance", 4)
        ],
        default='0'
    )
    # Power & Efficacy
    power = FloatProperty(
        name="Power",
        description="Power, W",
        default=40.0,
        min=0.0, max=1000000000.0,
        soft_min=0.0, soft_max=1000.0,
        precision=1,
        step=100
    )
    efficacy = FloatProperty(
        name="Efficacy",
        description="Efficacy, lm/W",
        default=17.6,
        min=0.0, max=683.0,
        soft_min=0.0, soft_max=683.0,
        precision=1,
        step=100
    )
    # Lumen, Lux, Candela, Luminance
    output = FloatProperty(
        name="Output",
        default=100.0
    )
    # Lobe.IES
    ies = StringProperty(
        name="IES/EULUMDAT",
        subtype='FILE_PATH'
    )
    # Lobe.Intensity
    intensity = FloatProperty(
        name="Intensity",
        default=1.0,
        min=0.0, max=100000.0,
        soft_min=0.0, soft_max=100000.0,
        precision=1,
        step=100
    )
    # Lobe.Image
    image = StringProperty(
        name="Map"
    )
    # Temperature
    temperature = FloatProperty(
        name="Temperature",
        default=6500.0,
        min=273.0, max=10000.0,
        soft_min=273.0, soft_max=10000.0,
        precision=3,
        step=1000
    )
    # HDR Image
    hdri_tex = StringProperty(
        name="Image"
    )
    hdri_intensity = FloatProperty(
        name="Intensity",
        description="Intensity",
        default=1.0,
        min=0.0, max=1000000.0,
        soft_min=0.0, soft_max=1000000.0,
        precision=1,
        step=100
    )


class MaxwellLayerProperties(PropertyGroup):
    def update_name(self, context):
        mat = getattr(context, "material", None)
        if mat is None:
            mat = context.object.active_material
        mx = mat.maxwell
        layers = mx.layers
        if layers.keys().count(self.name) > 1:
            n = 1
            layer_name = "Layer"
            while layers.find(layer_name) >= 0:
                layer_name = "Layer.%03d" % n
                n += 1
            self["name"] = layer_name

    enabled = BoolProperty(
        name="Enabled",
        default=True,
        update=_color_sync
    )
    name = StringProperty(
        name="Name",
        default="Layer",
        update=update_name
    )
    opacity = FloatProperty(
        name="Opacity",
        description="Like in Photoshop, you can set the Opacity value of a layer to define its visibility, "
                    "and thus, the visibility of the layers underneath",
        default=100.0,
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0,
        precision=1,
        step=100,
        update=_color_sync
    )
    opacity_tex = StringProperty(
        name="Mask",
        description="In addition to Opacity values, you can also map the opacity of a layer using a mask map, "
                    "a greyscale image that indicates where the layer is visible"
    )
    blend = EnumProperty(
        name="Layer Blending",
        items=[
            ('0', "Normal", "", 0),
            ('1', "Additive", "", 1)
        ],
        default='0'
    )
    emitter = PointerProperty(
        name="Emitter",
        type=MaxwellEmitterProperties
    )
    bsdfs = CollectionProperty(
        name="Material BSDFs",
        type=MaxwellBSDFProperties
    )
    active_bsdf_index = IntProperty(
        name="Active BSDF",
        default=-1
    )


class MaxwellDisplacementProperties(PropertyGroup):
    expanded = BoolProperty(
        options={'HIDDEN'}
    )
    enabled = BoolProperty(
        name="Enable Displacement",
        description="Enable Displacement"
    )
    # Global Properties
    map = StringProperty(
        name="Map"
    )
    type = EnumProperty(
        name="Type",
        items=[
            ('0', "On-The-Fly", "", 0),
            ('1', "Pretessellated", "", 1),
            ('2', "Vector", "", 2)
        ],
        default='1'
    )
    level = IntProperty(
        name="Subdivision",
        description="Subdivision Level",
        default=6
    )
    adaptive = BoolProperty(
        name="Adaptive",
        description="Enable auto adaptive subdivision"
    )
    smooth = BoolProperty(
        name="Smoothing",
        default=True
    )
    offset = FloatProperty(
        name="Offset",
        min=-1000.0, max=1000.0,
        soft_min=-1000.0, soft_max=1000.0,
        default=0.5,
        precision=1
    )
    method = EnumProperty(
        name="Method",
        description="Subdivision Method",
        items=[
            ('0', "Flat", "", 0),
            ('1', "Catmull/Loop", "", 1)
        ],
        default='0'
    )
    uvi = EnumProperty(
        name="UV Interpolation",
        items=[
            ('0', "None", "", 0),
            ('1', "Edges", "", 1),
            ('2', "Edges and Corners", "", 2),
            ('3', "Sharp", "", 3)
        ],
        default='2'
    )
    # HeightMap Properties
    height = FloatProperty(
        name="Height",
        min=-1000.0, max=1000.0,
        soft_min=-1000.0, soft_max=1000.0,
        default=2.0,
        precision=1,
        step=100
    )
    absolute = BoolProperty(
        name="Absolute",
        description="Use real units for height"
    )
    # Vector 3D Properties
    v3d_expanded = BoolProperty(
        options={'HIDDEN'},
        default=True
    )
    preset = EnumProperty(
        name="Preset",
        items=[
            ('0', "Custom", "", 0),
            ('1', "Zbrush Tangent", "", 1),
            ('2', "Zbrush World", "", 2),
            ('3', "Mudbox Absolute Tangent", "", 3),
            ('4', "Mudbox Object", "", 4),
            ('5', "Mudbox World", "", 5),
            ('6', "Realflow", "", 6),
            ('7', "Modo", "", 7)
        ],
        default='0'
    )
    transform = EnumProperty(
        name="Transform",
        items=[
            ('0', "Tangent", "", 0),
            ('1', "Object", "", 1),
            ('2', "World", "", 2),
            ('3', "Realflow", "", 3)
        ],
        default='0'
    )
    mapping = EnumProperty(
        name="RGB Mapping",
        items=[
            ('0', "XYZ", "", 0),
            ('1', "XZY", "", 1),
            ('2', "YZX", "", 2),
            ('3', "YXZ", "", 3),
            ('4', "ZXY", "", 4),
            ('5', "ZYX", "", 5)
        ],
        default='0'
    )
    scale = FloatVectorProperty(
        name="Scale",
        description="Scale multiplier in each axis",
        default=(1.0, 1.0, 1.0),
        min=-100000.0, max=100000.0,
        soft_min=-100000.0, soft_max=100000.0,
        subtype='XYZ'
    )


class MaxwellAGSProperties(PropertyGroup):
    ui_collapsed = BoolProperty(
        options={'HIDDEN'}
    )
    color = FloatVectorProperty(
        name="Color",
        default=(1.0, 1.0, 1.0),
        min=0.0, max=1.0,
        soft_min=0.0, soft_max=1.0,
        subtype='COLOR',
        update=_color_sync
    )
    reflection = FloatProperty(
        name="Reflection",
        default=12.0,
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0,
        subtype='PERCENTAGE',
        precision=1,
    )
    type = EnumProperty(
        name="Type",
        items=[
            ('0', "Normal", "", 0),
            ('1', "Clear", "", 1)
        ],
        default='0'
    )


class MaxwellOpaqueProperties(PropertyGroup):
    ui_collapsed = BoolProperty(
        options={'HIDDEN'}
    )
    color = FloatVectorProperty(
        name="Color",
        default=(1.0, 1.0, 1.0),
        min=0.0, max=1.0,
        soft_min=0.0, soft_max=1.0,
        subtype='COLOR',
        update=_color_sync
    )
    color_map = StringProperty(
        name="Color Map"
    )
    color_map_enabled = BoolProperty(
        name="Color Map Enabled"
    )
    shininess = FloatProperty(
        name="Shininess",
        default=35.0,
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0,
        subtype='PERCENTAGE',
        precision=1,
    )
    shininess_map = StringProperty(
        name="Shininess Map"
    )
    shininess_map_enabled = BoolProperty(
        name="Shininess Map Enabled"
    )
    roughness = FloatProperty(
        name="Roughness",
        default=5.0,
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0,
        subtype='PERCENTAGE',
        precision=1,
    )
    roughness_map = StringProperty(
        name="Roughness Map"
    )
    roughness_map_enabled = BoolProperty(
        name="Roughness Map Enabled"
    )
    clearcoat = BoolProperty(
        name="Clearcoat"
    )


class MaxwellTransparentProperties(PropertyGroup):
    ui_collapsed = BoolProperty(
        options={'HIDDEN'}
    )
    color = FloatVectorProperty(
        name="Color",
        default=(1.0, 1.0, 1.0),
        min=0.0, max=1.0,
        soft_min=0.0, soft_max=1.0,
        subtype='COLOR',
        update=_color_sync
    )
    color_map = StringProperty(
        name="Color Map"
    )
    color_map_enabled = BoolProperty(
        name="Color Map Enabled"
    )
    ior = FloatProperty(
        name="Ref. Index",
        default=1.51,
        min=1.001, max=2.5,
        precision=3
    )
    transparency = FloatProperty(
        name="Transparency",
        default=10.0,
        min=0.1, max=999.0,
        precision=1
    )
    roughness = FloatProperty(
        name="Roughness",
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0,
        subtype='PERCENTAGE',
        precision=1,
    )
    roughness_map = StringProperty(
        name="Roughness Map"
    )
    roughness_map_enabled = BoolProperty(
        name="Roughness Map Enabled"
    )
    specular_tint = FloatProperty(
        name="Specular Tint",
        min=0.0, max=100.0,
        subtype='PERCENTAGE',
        precision=1
    )
    dispersion = FloatProperty(
        name="Dispersion",
        min=0.0, max=100.0,
        subtype='PERCENTAGE',
        precision=1
    )
    clearcoat = BoolProperty(
        name="Clearcoat"
    )


class MaxwellMetallProperties(PropertyGroup):
    ui_collapsed = BoolProperty(
        options={'HIDDEN'}
    )
    ior = EnumProperty(
        name="Metal IOR",
        items=[
            ('0', "Aluminium", "Aluminium", 0),
            ('1', "Chromium", "Chromium", 1),
            ('2', "Cobalt", "Cobalt", 2),
            ('3', "Copper", "Copper", 3),
            ('4', "Germanium", "Germanium", 4),
            ('5', "Gold", "Gold", 5),
            ('6', "Iron", "Iron", 6),
            ('7', "Nickel", "Nickel", 7),
            ('8', "Silver", "Silver", 8),
            ('9', "Titanium", "Titanium", 9),
            ('10', "Vanadium", "Vanadium", 10)
        ],
        default='0'
    )
    tint = FloatProperty(
        name="Tint",
        default=0.0,
        min=0.0, max=100.0,
        precision=1
    )
    color = FloatVectorProperty(
        name="Color",
        default=(1.0, 1.0, 1.0),
        min=0.0, max=1.0,
        subtype='COLOR',
        update=_color_sync
    )
    color_map = StringProperty(
        name="Color Map"
    )
    color_map_enabled = BoolProperty(
        name="Color Map Enabled"
    )
    roughness = FloatProperty(
        name="Roughness",
        min=0.0, max=100.0,
        subtype='PERCENTAGE',
        precision=1,
    )
    roughness_map = StringProperty(
        name="Roughness Map"
    )
    roughness_map_enabled = BoolProperty(
        name="Roughness Map Enabled"
    )
    anisotropy = FloatProperty(
        name="Anisotropy",
        min=0.0, max=100.0,
        subtype='PERCENTAGE',
        precision=1,
    )
    anisotropy_map = StringProperty(
        name="Anisotropy Map"
    )
    anisotropy_map_enabled = BoolProperty(
        name="Anisotropy Map Enabled"
    )
    angle = FloatProperty(
        name="Angle",
        min=0.0, max=360.0,
        precision=1,
    )
    angle_map = StringProperty(
        name="Angle Map"
    )
    angle_map_enabled = BoolProperty(
        name="Angle Map Enabled"
    )
    dust = FloatProperty(
        name="Dust & Dirt",
        min=0.0, max=100.0,
        subtype='PERCENTAGE',
        precision=1,
    )
    dust_map = StringProperty(
        name="Dust & Dirt Map"
    )
    dust_map_enabled = BoolProperty(
        name="Dust & Dirt Map Enabled"
    )
    perforation_map = StringProperty(
        name="Perforation Map"
    )
    perforation_map_enabled = BoolProperty(
        name="Perforation Map Enabled"
    )


class MaxwellTranslucentProperties(PropertyGroup):
    ui_collapsed = BoolProperty(
        options={'HIDDEN'}
    )
    scale = FloatProperty(
        name="Scale",
        default=10.0,
        min=0.00001, max=1000000.0,
        precision=2
    )
    ior = FloatProperty(
        name="Ref. Index",
        default=1.51,
        min=1.001, max=2.5,
        precision=3
    )
    color = FloatVectorProperty(
        name="Color",
        default=(1.0, 1.0, 1.0),
        min=0.0, max=1.0,
        soft_min=0.0, soft_max=1.0,
        subtype='COLOR',
        update=_color_sync
    )
    color_map = StringProperty(
        name="Color Map"
    )
    color_map_enabled = BoolProperty(
        name="Color Map Enabled"
    )
    hue_shift = FloatProperty(
        name="Hue Shift",
        min=-120.0, max=120,
        precision=1
    )
    invert_hue = BoolProperty(
        name="Invert Hue"
    )
    vibrance = FloatProperty(
        name="Vibrance",
        min=0.0, max=100.0,
        precision=1
    )
    density = FloatProperty(
        name="Density",
        min=0.0, max=100.0,
        precision=1
    )
    opacity = FloatProperty(
        name="Opacity",
        min=0.0, max=100.0,
        precision=1
    )
    roughness = FloatProperty(
        name="Roughness",
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0,
        subtype='PERCENTAGE',
        precision=1,
    )
    roughness_map = StringProperty(
        name="Roughness Map"
    )
    roughness_map_enabled = BoolProperty(
        name="Roughness Map Enabled"
    )
    specular_tint = FloatProperty(
        name="Specular Tint",
        min=0.0, max=100.0,
        subtype='PERCENTAGE',
        precision=1
    )
    clearcoat = BoolProperty(
        name="Clearcoat"
    )
    clearcoat_ior = FloatProperty(
        name="Clearcoat IOR",
        default=1.51,
        min=1.001, max=2.5,
        precision=3
    )


class MaxwellCarPaintProperties(PropertyGroup):
    ui_collapsed = BoolProperty(
        options={'HIDDEN'}
    )
    color = FloatVectorProperty(
        name="Color",
        default=(0.392157, 0.0, 0.062745),
        min=0.0, max=1.0,
        soft_min=0.0, soft_max=1.0,
        subtype='COLOR',
        update=_color_sync
    )
    metallic = FloatProperty(
        name="Reflection",
        default=100.0,
        min=0.0, max=100.0,
        precision=1,
    )
    topcoat = FloatProperty(
        name="Topcoat",
        default=50.0,
        min=1.001, max=100.0,
        precision=3,
    )


class MaxwellHairProperties(PropertyGroup):
    ui_collapsed = BoolProperty(
        options={'HIDDEN'}
    )
    color = FloatVectorProperty(
        name="Color",
        default=(1.0, 1.0, 1.0),
        min=0.0, max=1.0,
        subtype='COLOR',
        update=_color_sync
    )
    color_map = StringProperty(
        name="Color Map"
    )
    color_map_enabled = BoolProperty(
        name="Color Map Enabled"
    )
    rt_map = StringProperty(
        name="Root-Tip Map"
    )
    rt_weight = FloatProperty(
        name="Root-Tip Weight",
        default=50.0,
        min=1.0, max=100.0,
        precision=1,
    )
    rt_weight_map = StringProperty(
        name="Root-Tip Weight Map"
    )
    rt_weight_map_enabled = BoolProperty(
        name="Root-Tip Weight Map Enabled"
    )
    primary_strength = FloatProperty(
        name="Strength",
        default=40.0,
        min=1.0, max=100.0,
        precision=1,
    )
    primary_spread = FloatProperty(
        name="Spread",
        default=36.0,
        min=1.0, max=100.0,
        precision=1,
    )
    primary_tint = FloatVectorProperty(
        name="Tint",
        default=(1.0, 1.0, 1.0),
        min=0.0, max=1.0,
        subtype='COLOR'
    )
    secondary_strength = FloatProperty(
        name="Strength",
        default=40.0,
        min=1.0, max=100.0,
        precision=1,
    )
    secondary_spread = FloatProperty(
        name="Spread",
        default=45.0,
        min=1.0, max=100.0,
        precision=1,
    )
    secondary_tint = FloatVectorProperty(
        name="Tint",
        default=(1.0, 1.0, 1.0),
        min=0.0, max=1.0,
        subtype='COLOR'
    )


class MaxwellOverrideMapProperties(PropertyGroup, Projection):
    pass
    
class MaxwellMaterialProperties(PropertyGroup):
    def update_color(self, context):
        _color_sync(self, context)

    auto_color = BoolProperty(
        name="Auto Color Sync",
        update=_color_sync
    )
    # Global Properties
    override_map = PointerProperty(
        name="Override Map",
        type=MaxwellOverrideMapProperties
    )
    bump = FloatProperty(
        name="Bump/Normal",
        default=30,
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0,
        precision=1,
        step=100
    )
    bump_tex = StringProperty(
        name="Global Bump"
    )
    bump_normal = BoolProperty(
        name="Normal"
    )
    dispersion = BoolProperty(
        name="Dispersion"
    )
    shadow = BoolProperty(
        name="Shadow"
    )
    matte = BoolProperty(
        name="Matte"
    )
    color_id = FloatVectorProperty(
        name="Material ID",
        default=(1.0, 1.0, 1.0),
        min=0.0, max=1.0,
        soft_min=0.0, soft_max=1.0,
        subtype='COLOR'
    )
    
    def _get_pi_color(self):
        # 0...32767, 0...7FFF
        # channel 0...31, 0...1F
        # color would be gamma corrected while render (g=2.2 bug?)
        ig = 5 / 11 # 1 / 2.2, gamma inverse
        idx = self.id_data.pass_index
        r = (((idx >> 10) & 0x1f) / 31) ** ig
        g = (((idx >> 5) & 0x1f) / 31) ** ig
        b = ((idx & 0x1f) / 31) ** ig
        return (r, g, b)

    pi_as_color = BoolProperty(
        name="Pass Index",
        description="Convert pass index to color id",
        default=True
    )
    pi_color = FloatVectorProperty(
        name="Material ID",
        subtype='COLOR',
        get=_get_pi_color
    )
    displacement = PointerProperty(
        name="Displacement",
        type=MaxwellDisplacementProperties
    )
    layers = CollectionProperty(
        name="Layers",
        type=MaxwellLayerProperties
    )
    active_layer_index = IntProperty(
        name="Active Layer",
        default=-1,
        min=-1
    )
    type = EnumProperty(
        name="Material Type",
        items=[
            ('file', "Referenced", ""),
            ('custom', "Custom", ""),
            ('ags', "AGS", ""),
            ('opaque', "Opaque", ""),
            ('transparent', "Transparent", ""),
            ('metal', "Metal", ""),
            ('translucent', "Translucent", ""),
            ('carpaint', "Car Paint", ""),
            ('hair', "Hair", "")
        ],
        default='file',
        update=_color_sync
    )
    # Referenced
    mxm = StringProperty(
        name="MXM File",
        subtype='FILE_PATH'
    )
    embed = BoolProperty(
        name="Embedded",
        description="Embed material in maxwell scene file while export"
    )
    # Presets
    ags = PointerProperty(type=MaxwellAGSProperties)
    opaque = PointerProperty(type=MaxwellOpaqueProperties)
    transparent = PointerProperty(type=MaxwellTransparentProperties)
    metal = PointerProperty(type=MaxwellMetallProperties)
    translucent = PointerProperty(type=MaxwellTranslucentProperties)
    carpaint = PointerProperty(type=MaxwellCarPaintProperties)
    hair = PointerProperty(type=MaxwellHairProperties)

    @classmethod
    def register(cls):
        Material.maxwell = PointerProperty(type=cls)

    @classmethod
    def unregiser(cls):
        del Material.maxwell
