# -*- coding: utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Ildar Nikolaev"
__email__ = "nildar@users.sourceforge.net"

import os

import bpy
from bpy.types import (
    PropertyGroup,
    Object,
    Mesh,
    Curve,
    TextCurve,
    MetaBall
)
from bpy.props import (
    PointerProperty,
    StringProperty,
    EnumProperty,
    FloatProperty,
    FloatVectorProperty,
    CollectionProperty,
    BoolProperty,
    IntProperty
)
from .particles import MaxwellVolumetricExtProperties


_PROJECTION_TYPES = [
    ('PLANAR', "Planar", "Planar"),
    ('SPHERICAL', "Spherical", "Spherical"),
    ('CYLINDRICAL', "Cylindrical", "Cylindrical"),
    ('CUBIC', "Cubic", "Cubic"),
    ('LOCKED', "Locked", "Locked"),
]


class MaxwellProjectionProperties(PropertyGroup):
    def _get_items(self, context):
        if hasattr(self.id_data, "uv_layers"):
            return _PROJECTION_TYPES
        return _PROJECTION_TYPES[:-1]

    @property
    def type_label(self):
        return _PROJECTION_TYPES[self.get("type", 0)][1]

    name = StringProperty(
        name="UV Map"
    )
    type = EnumProperty(
        name="Type",
        description="Projection type",
        items=_get_items
    )
    channel = IntProperty(
        name="Channel",
        min=0, max=254,
        options=set(),
        subtype='UNSIGNED'
    )
    object = StringProperty(
        name="Object",
        description="Object"
    )
    position = FloatVectorProperty(
        name="Position",
        description="Position",
        default=(0.0, 0.0, 0.0),
        precision=4,
        step=100,
        subtype='TRANSLATION'
    )
    rotation = FloatVectorProperty(
        name="Rotation",
        description="Rotation",
        default=(0.0, 0.0, 0.0),
        precision=4,
        step=100,
        subtype='EULER',
        unit='ROTATION'
    )
    scale = FloatVectorProperty(
        name="Scale",
        description="Scale",
        default=(1.0, 1.0, 1.0),
        precision=4,
        step=0.1,
        subtype='XYZ'
    )
    latitude = FloatVectorProperty(
        name="Latitude",
        size=2,
        default=(-90.0, 90.0),
        min=-90.0, max=90.0,
        soft_min=-90.0, soft_max=90.0
    )
    longitude = FloatVectorProperty(
        name="Longitude",
        size=2,
        default=(-180.0, 180.0),
        min=-180.0, max=180.0,
        soft_min=-180.0, soft_max=180.0
    )
    angle = FloatVectorProperty(
        name="Perimeter",
        size=2,
        default=(0.0, 0.0),
        min=0.0, max=1.0,
        soft_min=0.0, soft_max=1.0,
        subtype='EULER'
    )


class MaxwellSeaProperties(PropertyGroup):
    _quality_items=[
        ('0', "4x4", "4x4", 0),
        ('1', "8x8", "8x8", 1),
        ('2', "16x16", "16x16", 2),
        ('3', "32x32", "32x32", 3),
        ('4', "64x64", "64x64", 4),
        ('5', "128x128", "128x128", 5),
        ('6', "256x256", "256x256", 6),
        ('7', "512x512", "512x512", 7),
        ('8', "1024x1024", "1024x1024", 8),
        ('9', "2048x2048", "2048x2048", 9),
        ('10', "4096x4096", "4096x4096", 10),
        ('11', "8192x8192", "8192x8192", 11)
    ]

    enabled = BoolProperty(
        name="Enabled",
        description="Export mesh as Maxwell Sea"
    )
    # Geometry
    ui_geometry_collapsed = BoolProperty(
    )
    material = StringProperty(
        name = "Material"
    )
    quality = EnumProperty(
        name="Quality",
        items=_quality_items,
        default='6',
        options=set()
    )
    time = FloatProperty(
        name="Reference Time",
        min=0.0, max=100000.0,
        precision=4
    )
    vertical_scale = FloatProperty(
        name="Vertical Scale",
        min=0.0, max=100000.0,
        default=0.1,
        precision=5,
    )
    dimension = FloatProperty(
        name="Dimension",
        min=0.0, max=1000000.0,
        default=250.0,
        precision=2
    )
    depth = FloatProperty(
        name="Depth",
        min=0.0, max=100000.0,
        default=250.0,
        precision=2
    )
    seed = IntProperty(
        name="Seed",
        subtype='UNSIGNED',
        min=0, max=65535,
        default=4217
    )
    choppyness = BoolProperty(
        name="Choppyness"
    )
    choppy_factor = FloatProperty(
        name="Choppy factor",
        min=0.0, max=100000.0,
        precision=2
    )
    # Wind
    ui_wind_collapsed = BoolProperty(
    )
    wind_speed = FloatProperty(
        name="Wind Speed",
        min=0.0, max=100000.0,
        default=30.0,
        precision=3
    )
    wind_direction = FloatProperty(
        name="Wind Direction",
        min=0.0, max=360.0,
        default=45.0,
        precision=1
    )
    weight_against_wind = FloatProperty(
        name="Weight Against Wind",
        min=0.0, max=1.0,
        default=0.5,
        precision=4
    )
    wind_alignment = FloatProperty(
        name="Wind Alignment",
        min=0.0, max=100000.0,
        default=2.0,
        precision=4
    )
    min_wave_length = FloatProperty(
        name="Min Wave Length",
        min=0.0, max=100000,
        default=0.1,
        precision=4
    )
    # Bake to Mesh Cache file
    def _get_mc_file(self):
        f = self.get("_mc")
        if f is None:
            f = "//" + self.id_data.name + ".pc2"
        return f

    def _set_mc_file(self, value):
        self["_mc"] = value
    
    ui_mc_collapsed = BoolProperty(
    )
    mc_modifier = StringProperty(
        name="Modifier",
        description="Mesh Cache modifier name",
        default="Maxwell Sea"
    )
    mc_file = StringProperty(
        name="Cache File",
        description="Mesh Cache File (*.pc2)",
        subtype='FILE_PATH',
        get=_get_mc_file,
        set=_set_mc_file
    )
    mc_quality = EnumProperty(
        name="Quality",
        items=_quality_items,
        default='4',
        options=set()
    )
    mc_frames = bpy.props.IntVectorProperty(
        name="Frames",
        size=2,
        default=(1, 2),
        options=set()
    )


class MaxwellObjectDataProperties(PropertyGroup):
    use_projections = BoolProperty(
        description="Use Maxwell Render UV Sets"
    )
    projections = CollectionProperty(
        type=MaxwellProjectionProperties
    )
    active_projection = IntProperty(
        default=0
    )
    # extensions
    sea = PointerProperty(
        type=MaxwellSeaProperties
    )

    @classmethod
    def register(cls):
        Mesh.maxwell = PointerProperty(type=cls)
        Curve.maxwell = PointerProperty(type=cls)
        TextCurve.maxwell = PointerProperty(type=cls)
        MetaBall.maxwell = PointerProperty(type=cls)

    @classmethod
    def unregiser(cls):
        del Mesh.maxwell
        del Curve.maxwell
        del TextCurve.maxwell
        del MetaBall.maxwell


class MaxwellReferenceProperties(PropertyGroup):
    path = StringProperty(
        name="File (.mxs)",
        description="MXS Reference",
        subtype='FILE_PATH'
    )
    flags = EnumProperty(
        name="Override Flags",
        items=[
            ('HIDE', "Hidden", "Hide", 1),
            ('CAMERA', "Hidden from Camera", "Hidden from camera", 2),
            ('REFL_REFR', "Hidden from Refl & Refr", "Hidden from reflection & refraction", 4),
            ('HIDE_TO_GI', "Hidden from Global illumination", "Hidden from global illumination", 8)
        ],
        options={
            'ANIMATABLE',
            'ENUM_FLAG'
        }
    )
    flag_hidden = BoolProperty(
        name="Hidden",
        get=lambda s: 'HIDE' in s.flags,
        set=lambda s, v: setattr(s, "flags", s.flags | {'HIDE'} if v else s.flags - {'HIDE'})
    )
    flag_camera = BoolProperty(
        name="Camera",
        description="Hidden from Camera",
        get=lambda s: 'CAMERA' in s.flags,
        set=lambda s, v: setattr(s, "flags", s.flags | {'CAMERA'} if v else s.flags - {'CAMERA'})
    )
    flag_rr = BoolProperty(
        name="Reflection & Refraction",
        description="Hidden from Reflection & Refraction",
        get=lambda s: 'REFL_REFR' in s.flags,
        set=lambda s, v: setattr(s, "flags", s.flags | {'REFL_REFR'} if v else s.flags - {'REFL_REFR'})
    )
    flag_gi = BoolProperty(
        name="Global illumination",
        description="Hidden from Global illumination",
        get=lambda s: 'HIDE_TO_GI' in s.flags,
        set=lambda s, v: setattr(s, "flags", s.flags | {'HIDE_TO_GI'} if v else s.flags - {'HIDE_TO_GI'})
    )


class MaxwellObjectProperties(PropertyGroup):
    type = EnumProperty(
        name="Export As",
        items=[
            ('NONE', "None", "Doesn't export"),
            ('REFERENCE', "Reference", "Reference"),
            ('VOLUMETRIC', "Volumetric", "Volumetric Extension")
        ],
        default='NONE'
    )
    reference = PointerProperty(
        name="Reference",
        type=MaxwellReferenceProperties
    )
    volumetric = PointerProperty(
        name="Volumetric",
        type=MaxwellVolumetricExtProperties
    )
    hide_to_camera = BoolProperty(
        name="Camera",
        description="Hide to Camera"
    )
    hide_to_rr = BoolProperty(
        name="Reflections & Refractions",
        description="Hide to Reflections & Refractions"
    )
    hide_to_gi = BoolProperty(
        name="Global Illumination",
        description="Hide to Global Illumination"
    )
    cd_refl = BoolProperty(
        name="Reflection",
        description="Do Direct Coustics Reflection",
        default=True
    )
    cd_refr = BoolProperty(
        name="Refraction",
        description="Do Direct Coustics Refraction",
        default=True
    )
    ci_refl = BoolProperty(
        name="Reflection",
        description="Do Indirect Coustics Reflection",
        default=True
    )
    ci_refr = BoolProperty(
        name="Refraction",
        description="Do Indirect Coustics Refraction",
        default=True
    )
    ecp = BoolProperty(
        name="Excluded Of CutPlanes",
        description="Excluded Of CutPlanes"
    )
    opacity = FloatProperty(
        name="Opacity",
        default=100.0,
        min=0.0, max=100.0,
        subtype='PERCENTAGE'
    )
    backface_material = StringProperty(
        name="Backface Material"
    )

    def _get_pi_color(self):
        # 0...32767, 0...7FFF
        # channel 0...31, 0...1F
        # color would be gamma corrected while render (g=2.2 bug?)
        ig = 5 / 11 # 1 / 2.2, gamma inverse
        idx = self.id_data.pass_index
        r = (((idx >> 10) & 0x1f) / 31) ** ig
        g = (((idx >> 5) & 0x1f) / 31) ** ig
        b = ((idx & 0x1f) / 31) ** ig
        return (r, g, b)

    color_id = FloatVectorProperty(
        name="Color ID",
        default=(0.0, 0.0, 0.0),
        min=0.0, max=1.0,
        subtype='COLOR'
    )
    pi_as_color = BoolProperty(
        name="Pass Index",
        description="Pass Index as Color",
        default=True
    )
    pi_color = FloatVectorProperty(
        name="Color ID",
        subtype='COLOR',
        get=_get_pi_color
    )

    @classmethod
    def register(cls):
        Object.maxwell = PointerProperty(type=cls)

    @classmethod
    def unregiser(cls):
        del Object.maxwell


class HeadusUVLayoutProperties(PropertyGroup):
    def _get(self):
        id_data = self.id_data
        uv_layers = getattr(id_data, "uv_layers", None)
        if uv_layers is not None:
            uvl = uv_layers.active
            if uvl:
                p = self.get("_paths")
                if p:
                    return p.get(uvl.name, "")
                tmp = bpy.context.user_preferences.filepaths.temporary_directory
                return os.path.join(tmp, "%s-%s.obj" % (id_data.name, uvl.name))
        return ""

    def _set(self, value):
        id_data = self.id_data
        uv_layers = getattr(id_data, "uv_layers", None)
        if uv_layers is not None:
            # TODO: cleanup an unreferenced paths
            uvl = uv_layers.active
            if uvl:
                p = self.get("_paths")
                if p is None:
                    self["_paths"] = {}
                    p = self["_paths"]
                p[uvl.name] = value

    uv_file = StringProperty(
        name="UV File",
        subtype='FILE_PATH',
        get=_get,
        set=_set
    )
    ui_collapsed = BoolProperty(
        default=True
    )

    @classmethod
    def register(cls):
        Mesh.headus_uvlayout = PointerProperty(type=cls)

    @classmethod
    def unregiser(cls):
        del Mesh.headus_uvlayout
