# -*- coding: utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Ildar Nikolaev"
__email__ = "nildar@users.sourceforge.net"

from bpy.types import PropertyGroup, ParticleSettings
from bpy.props import (
    FloatProperty,
    FloatVectorProperty,
    StringProperty,
    IntProperty,
    BoolProperty,
    EnumProperty,
    PointerProperty
)


class MaxwellHairExtProperties(PropertyGroup):
    root = FloatProperty(
        name="Root Radius",
        default=0.005,
        min=0.0,
        soft_min=0.0,
        precision=3,
        step=0.1
    )
    tip = FloatProperty(
        name="Tip Radius",
        default=0.001,
        min=0.0,
        soft_min=0.0,
        precision=3,
        step=0.1
    )
    uv_map = StringProperty(
        name="UV Channel 1",
        description="UV Map for 'Channel 1'"
    )


class MaxwellGrassExtProperties(PropertyGroup):
    # Primitive
    # TODO: Implement?
    #backface = StringProperty(
    #    name="Backface Material"
    #)
    type = EnumProperty(
        name="Type",
        description="Primitive Type",
        items=[
            ('0', "Curve", "The Curve primitive renders a smooth curved blade of grass"),
            ('1', "Flat", "The Flat primitive renders a segmented flat strip"),
            ('2', "Cylinder", "The Cylinder primitive renders each guide as a rounded cylinder")],
        default='0'
    )
    blade_points = IntProperty(
        name="Points Per Blade",
        description="Sets the number of points to use for the primitive you have chosen",
        default=4,
        min=2, max=20,
        soft_min=2, soft_max=20,
        subtype='UNSIGNED'
    )
    # Grass Density
    density = IntProperty(
        name="Density",
        description="Controls the number of blades per square meter",
        default=100,
        min=0, max=100000000,
        soft_min=0, soft_max=100000,
        subtype='UNSIGNED'
    )
    density_map = StringProperty(
        name="Density Map",
        description="Specify a greyscale map to control where the grass will grow"
    )
    seed = IntProperty(
        name="Seed",
        description="This number is used to generate the random distribution of the grass blades",
        min=0, max=16300,
        soft_min=0, soft_max=16300,
        subtype='UNSIGNED'
    )
    # Blade Length
    length = FloatProperty(
        name="Length",
        description="The length of the blades in centimeters",
        default=10.0,
        min=0.0, max=100000.0,
        soft_min=0.0, soft_max=100000.0,
        precision=3
    )
    length_map = StringProperty(
        name="Length Map",
        description="Load a greyscale map to control the height of the grass over its growth area"
    )
    length_var = FloatProperty(
        name="Length Variation",
        description="Randomize the length of each blade, specifying the range of possible length values "
                    "as a percentage of the nominal Length parameter",
        default=20.0,
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0,
        subtype='PERCENTAGE',
        precision=1
    )
    # Width
    root_width = FloatProperty(
        name="Root Width",
        description="Specify the width (or radius, in the cylindrical grass) in millimeters of each blade at its root",
        default=5.0,
        min=0.1, max=100000.0,
        soft_min=0.1, soft_max=100000.0,
        precision=3
    )
    tip_width = FloatProperty(
        name="Tip Width",
        description="Specify the width (or radius, in the cylindrical grass) in millimeters of each blade at its tip",
        default=1.0,
        min=0.1, max=100000.0,
        soft_min=0.1, soft_max=100000.0,
        precision=3
    )
    # Angle
    direction = EnumProperty(
        name="Direction",
        items=[
            ('0', "Polygon Normal", "", 0),
            ('1', "World-Y", "", 1)
        ],
        default='0'
    )
    angle = FloatProperty(
        name="Initial Angle",
        description="The initial angle at which the blades will grow",
        default=80.0,
        min=0.0, max=90.0,
        soft_min=0.0, soft_max=90.0,
        precision=1
    )
    angle_map = StringProperty(
        name="Initial Angle Map",
        description="You can use an RGB map which will define the initial angle"
    )
    angle_var = FloatProperty(
        name="Initial Angle Variation",
        description="Randomize the growth angle by a percentage of the initial angle variation",
        default=25.0,
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0,
        subtype='PERCENTAGE',
        precision=1
    )
    # Bend
    start_bend = FloatProperty(
        name="Start Bend",
        description="The length of the blade (in percentage) at which the blade starts bending",
        default=40.0,
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0,
        subtype='PERCENTAGE',
        precision=1
    )
    start_bend_map = StringProperty(
        name="Start Bend Map",
        description="Load a greyscale map to control the Start Bend values"
    )
    start_bend_var = FloatProperty(
        name="Start Bend Variation",
        description="Randomize the Start Bend parameter of each blade",
        default=25.0,
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0,
        subtype='PERCENTAGE',
        precision=1
    )
    bend_radius = FloatProperty(
        name="Bend Radius",
        description="The radius in centimeters of the bending",
        default=10.0,
        min=0.0, max=10000.0,
        soft_min=0.0, soft_max=10000.0,
        precision=1
    )
    bend_radius_map = StringProperty(
        name="Bend Radius Map",
        description="Load a greyscale map to control the Bend Radius of the grass across its growth area"
    )
    bend_radius_var = FloatProperty(
        name="Bend Radius Variation",
        description="Randomize the bending radius of each blade",
        default=25.0,
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0,
        subtype='PERCENTAGE',
        precision=1
    )
    bend_angle = FloatProperty(
        name="Bend Angle",
        description="The bending angle in degrees",
        default=80.0,
        min=0.0, max=360.0,
        soft_min=0.0, soft_max=360.0,
        precision=1
    )
    bend_angle_map = StringProperty(
        name="Bend Angle Map",
        description="Load a greyscale map to control the Bend Angle at different areas of its growth surface"
    )
    bend_angle_var = FloatProperty(
        name="Bend Angle Variation",
        description="Randomize the Bend Angle parameter",
        default=25.0,
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0,
        subtype='PERCENTAGE',
        precision=1
    )
    # Cut off
    cut_off = FloatProperty(
        name="Cut Off",
        description="The length of the blade (in percentage) at which the blade will be cut",
        default=100.0,
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0,
        subtype='PERCENTAGE',
        precision=1
    )
    cut_off_map = StringProperty(
        name="Cut Off Map",
        description="Load a greyscale map to control different Cut Off values "
                    "at different areas of the growing surface"
    )
    cut_off_var = FloatProperty(
        name="Cut Off Variation",
        description="Randomize the cut length parameter of each blade",
        default=0.0,
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0,
        subtype='PERCENTAGE',
        precision=1
    )
    # Level of Detail
    lod_enabled = BoolProperty(
        name="Enable LOD",
        description="This option allows you to optimize your scene by rendering more blades near the camera "
                    "and gradually less blades as the distance from the camera increases"
    )
    lod_min = FloatProperty(
        name="Min Distance",
        description="The distance from the camera where the LOD won't have any effect",
        default=10.0,
        min=0.0, max=100000.0,
        soft_min=0.0, soft_max=100000.0
    )
    lod_max = FloatProperty(
        name="Max Distance",
        description="The distance at which the blade density will have reached the density "
                    "specified in the Max Distance Density parameter",
        default=50.0,
        min=0.0, max=100000.0,
        soft_min=0.0, soft_max=100000.0
    )
    lod_max_density = FloatProperty(
        name="Max Distance Density",
        description="The number of blades per meter you wish to have when the Max Distance is reached",
        default=10.0,
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0,
        subtype='PERCENTAGE'
    )
    # Display
    display_percent = IntProperty(
        name="Display Percent",
        description="The percentage of blades to show in the Maxwell Studio viewport",
        default=10,
        min=0, max=100,
        soft_min=0, soft_max=100,
        subtype='PERCENTAGE'
    )
    display_blades = IntProperty(
        name="Display Max. Blades",
        description="The maximum number of blades to display in the Maxwell Studio viewport",
        default=1000,
        min=0, max=100000,
        soft_min=0, soft_max=100000
    )


class MaxwellParticlesExtProperties(PropertyGroup):
    radius = FloatProperty(
        name="Radius Multiplier",
        default=1.0,
        min=0.000001, max=1000000.0,
        soft_min=0.000001, soft_max=100.0,
        precision=6,
        step=0.001
    )
    motion_blur = FloatProperty(
        name="Motion Blur Multiplier",
        default=1.0,
        min=0.0, max=1000000.0,
        soft_min=0.0, soft_max=100.0,
        precision=6,
        step=0.001
    )
    shutter_speed = FloatProperty(
        name="Shutter Speed",
        default=125.0,
        min=0.000001, max=1000000.0,
        soft_min=0.000001, soft_max=100.0,
        precision=2,
        step=100.0
    )
    extra_particles = IntProperty(
        name="Extra Particles",
        description="Extra particles per particle",
        default=0,
        min=0, max=100000000,
        soft_min=0, soft_max=10000,
        step=1
    )
    extra_dispersion = FloatProperty(
        name="Dispersion",
        description="Extra particles dispersion",
        default=0.0,
        min=0.0, max=1000000.0,
        soft_min=0.0, soft_max=100.0,
        precision=2,
        step=10.0,
    )
    extra_deformation = FloatProperty(
        name="Deformation",
        description="Extra particles deformation",
        default=0.0,
        min=0.0, max=1000000.0,
        soft_min=0.0, soft_max=100.0,
        precision=2,
        step=10.0
    )


class MaxwellMesherExtProperties(PropertyGroup):
    # Globals
    mesher = EnumProperty(
        name="Mesh Generator",
        items=[
            ('0', 'RFRKMesh', "", 0),
            ('1', 'HybridoMesh', "", 1)
        ],
        default='0'
    )
    # Sequence
    radius = FloatProperty(
        name="Particle Radius",
        default=0.08,
        min=0.0, max=100000.0,
        soft_min=0.0, soft_max=100000.0,
        precision=2,
        step=1.0
    )
    core = FloatProperty(
        name="Core filter",
        default=0.0,
        min=0.0, max=1.0,
        soft_min=0.0, soft_max=1.0,
        precision=2,
        step=1.0
    )
    splash = FloatProperty(
        name="Splash filter",
        default=0.0,
        min=0.0, max=1.0,
        soft_min=0.0, soft_max=1.0,
        precision=2,
        step=1.0
    )
    velocity = FloatProperty(
        name="Max. Velocity",
        default=1000.0,
        min=0.0, max=100000.0,
        soft_min=0.0, soft_max=100000.0,
        precision=2,
        step=1000.0
    )
    # Mesh parameters
    poly = FloatProperty(
        name="Polygon Size",
        default=0.05,
        min=0.0, max=1000.0,
        soft_min=0.0, soft_max=1000.0,
        precision=2,
        step=1.0
    )
    wn = BoolProperty(
        name="Weight Normalization"
    )
    smooth = FloatProperty(
        name="Smooth",
        default=50.0,
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0,
        precision=2,
        step=100.0
    )
    iso_level = FloatProperty(
        name="Iso Level",
        default=0.0,
        min=-10.0, max=10.0,
        soft_min=10.0, soft_max=10.0,
        precision=2,
        step=100.0
    )
    fix_bbox = BoolProperty(
        name="Fix Bounding Box",
        default=True
    )
    # TODO: We really need this???
    #field_type = EnumProperty(
    #    name="Field Type",
    #    items=[
    #        ('0', "Metaball", "Metaball", 0),
    #        ('1', "Weighted_Isotropic", "Weighted_Isotropic", 1),
    #        ('2', "Weighted_Anisotropic", "Weighted_Anisotropic", 2)
    #    ],
    #    default='1'
    #)
    # Mesh filtering
    filter = BoolProperty(
        name="Enable Mesh Filter"
    )
    relax = FloatProperty(
        name="Relax",
        default=0.0,
        min=0.0, max=1000.0,
        soft_min=0.0, soft_max=1000.0,
        precision=2,
        step=10.0
    )
    tension = FloatProperty(
        name="Tension",
        default=0.0,
        min=0.0, max=1000.0,
        soft_min=0.0, soft_max=1000.0,
        precision=2,
        step=10.0
    )
    thinning = FloatProperty(
        name="Thinning",
        default=0.0,
        min=0.0, max=1.0,
        soft_min=0.0, soft_max=1.0,
        precision=2,
        step=1
    )
    steps = IntProperty(
        name="Steps",
        default=1,
        min=0, max=1000,
        soft_min=0, soft_max=1000,
        subtype='UNSIGNED'
    )
    splash_thinning = BoolProperty(
        name="Splash Thinning",
        default=True,
    )
    splash_threshold = FloatProperty(
        name="Splash Threshold",
        default=0.8,
        min=0.0, max=0.1,
        soft_min=0.0, soft_max=0.1
    )
    thinning_size = FloatProperty(
        name="Thinning Size",
        default=0.5,
        min=0.0, max=1.0,
        soft_min=0.0, soft_max=1.0
    )
    core_smooth = BoolProperty(
        name="Core Smoothing",
        default=True
    )
    core_threshold = FloatProperty(
        name="Core Threshold",
        default=0.4,
        min=0.0, max=1.0
    )
    smooth_steps = IntProperty(
        name="Smoothing Steps",
        default=128,
        min=0, max=1000,
        subtype='UNSIGNED'
    )
    # Mesh Optimization
    optimize = EnumProperty(
        name="Optimization",
        items=[
            ('0', "Disabled", "Disables the optimization process", 0),
            ('1', "Camera", "Camera", 1),
            ('2', "Curvature", "Curvature", 2),
        ],
        default='0'
    )
    distance = FloatProperty(
        name="Camera Distance",
        default=5.0,
        min=0.0, max=1000.0,
        soft_min=0.0, soft_max=1000000.0,
        precision=2,
        step=1000.0
    )
    # System parameters
    threads = IntProperty(
        name="Threads",
        default=0,
        min=0, max=1000,
        soft_min=0, soft_max=1000,
        subtype='UNSIGNED'
    )
    # Export and lazy compute
    action = EnumProperty(
        name="Action",
        items=[
            ('0', "Do Meshing", "Do Meshing", 0),
            ('1', "Do Meshing and Export", "Do Meshing and Export", 1),
            ('2', "Import Mesh", "Import Mesh", 2)
        ],
        default='0'
    )
    md_file = StringProperty(
        name="MD File",
        subtype='FILE_PATH'
    )
    md_offset = IntProperty(
        name="Frame Offset",
        default=0,
        min=-10000, max=10000,
        soft_min=-10000, soft_max=10000
    )
    md_compression = IntProperty(
        name="File Compression",
        min=0, max=9
    )
    md_normals = BoolProperty(
        name="Invert Normals"
    )


class MaxwellVolumetricExtProperties(PropertyGroup):
    enabled = BoolProperty(
        name="Enabled",
        description="Apply Maxwell Volumetric extension"
    )
    hide_emitter = BoolProperty(
        name="Hide emitter"
    )
    material = StringProperty(
        name = "Material"
    )
    field_type = EnumProperty(
        name="Field Type",
        items=[
            #('PARTICLE', "Particle based", "Particle based", 0),
            ('CONSTANT', "Constant", "Constant", 1),
            ('NOISE', "Noise 3D", "Noise 3D", 2)
        ],
        default='CONSTANT'
    )
    field_density = FloatProperty(
        name="Field Density",
        default=1.0,
        min=0.000001, max=10000.0,
        precision=6
    )
    # Noise 3D Field
    seed = IntProperty(
        name="Seed",
        default=4357,
        min=0, max=1000000,
        subtype='UNSIGNED'
    )
    low_value = FloatProperty(
        name="Low Value",
        min=0.0, max=1.0,
        precision=6
    )
    high_value = FloatProperty(
        name="High Value",
        default=1.0,
        min=0.0, max=1.0,
        precision=6
    )
    detail = FloatProperty(
        name="Detail",
        default=2.2,
        min=1.0, max=100.0,
        precision=4
    )
    octaves = IntProperty(
        name="Octaves",
        default=4,
        min=1, max=50,
        subtype='UNSIGNED'
    )
    persistence = FloatProperty(
        name="Persistence",
        default=0.55,
        min=0.0, max=1.0,
        precision=4
    )
    # Particles Field
    load_particles = FloatProperty(
        name="Load particles",
        default=100.0,
        min=0.0, max=100.0,
        subtype='PERCENTAGE'
    )
    radius_factor = FloatProperty(
        name="Radius Multiplier",
        default=1.0,
        min=0.000001, max=1000000.0,
        precision=6
    )
    density_scale = FloatProperty(
        name="Density Multiplier",
        default=1.0,
        min=0.000001, max=1000000.0,
        precision=6
    )
    min_density = FloatProperty(
        name="Min Density",
        description="Min Final Density",
        min=0.1, max=1.0,
        precision=6
    )
    max_density = FloatProperty(
        name="Max Density",
        description="Max Final Density",
        default=1.0,
        min=0.000001, max=10000.0,
        precision=6
    )
    cell_size = FloatProperty(
        name="Simulation Cell Size",
        default=0.1,
        min=0.000001, max=10000.0,
        precision=6
    )
    extra_particles = IntProperty(
        name="Extra Particles",
        description="Extra Particles Per Particle",
        min=0, max=1000000,
        subtype='UNSIGNED'
    )
    extra_dispersion = FloatProperty(
        name="Dispersion",
        description="Extra Particles Dispersion",
        default=1.0,
        min=0.0, max=1000000.0
    )
    extra_deformation = FloatProperty(
        name="Deformation",
        description="Extra Particles Deformation",
        min=0.0, max=1000000.0
    )
    mb_factor = FloatProperty(
        name="Motion Blur Multiplier",
        default=1.0,
        min=0.0, max=1000000.0,
        precision=6
    )
    shutter_speed = FloatProperty(
        name="Shutter Speed",
        default=125.0,
        min=0.000001, max=1000000.0
    )


class MaxwellScatterExtProperties(PropertyGroup):
    enabled = BoolProperty(
        name="Enabled",
        description="Use Maxwell Scatter Extension"
    )
    inherit_oid = BoolProperty(
        name="Inherit Object Id"
    )
    density = FloatProperty(
        name="Density",
        description="Density",
        default=100.0,
        min=0.0, max=100000000.0,
        soft_min=0.0, soft_max=100000.0,
        subtype='UNSIGNED'
    )
    density_map = StringProperty(
        name="Density Map",
        description="Density Map"
    )
    scale = FloatVectorProperty(
        name="Scale",
        description="Scale",
        min=0.0, max=100000.0,
        soft_min=0.0, soft_max=100000.0,
        default=(1.0, 1.0, 1.0),
        precision=4,
        step=100,
    )
    scale_var = FloatVectorProperty(
        name="Scale Variation",
        description="Scale variation",
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0,
        default=(20.0, 20.0, 20.0),
        precision=4,
        step=100,
    )
    scale_map = StringProperty(
        name="Length Map",
        description="Length Map"
    )
    rotation = FloatVectorProperty(
        name="Rotation",
        description="Rotation",
        min=0.0, max=360.0,
        soft_min=0.0, soft_max=360.0,
        default=(0.0, 0.0, 0.0),
        precision=4,
        step=100
    )
    rotation_var = FloatVectorProperty(
        name="Rotation Variation",
        description="Rotation variation",
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0,
        default=(10.0, 10.0, 10.0),
        precision=4,
        step=100
    )
    rotation_map = StringProperty(
        name="Rotation Map",
        description="Rotation Map"
    )
    direction = EnumProperty(
        name="Direction",
        description="Direction Type",
        items=[
            ('0', "Polygon Normal", "Polygon Normal"),
            ('1', "World-Y", "World-Y")
        ],
        default='0'
    )
    angle = FloatProperty(
        name="Initial Angle",
        description="The initial angle at which the blades will grow",
        default=80.0,
        min=0.0, max=90.0,
        soft_min=0.0, soft_max=90.0,
        precision=1
    )
    angle_var = FloatProperty(
        name="Initial Angle Variation",
        description="Randomize the growth angle by a percentage of the initial angle variation",
        default=25.0,
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0,
        subtype='PERCENTAGE',
        precision=1
    )
    angle_map = StringProperty(
        name="Initial Angle Map",
        description="You can use an RGB map which will define the initial angle"
    )
    seed = IntProperty(
        name="Seed",
        description="This number is used to generate the random distribution of the grass blades",
        min=0, max=16300,
        soft_min=0, soft_max=16300,
        subtype='UNSIGNED'
    )


class MaxwellParticlesProperties(PropertyGroup):
    h_type = EnumProperty(
        name="Type",
        items=[
            ('HAIR', "Hair", "Use Maxwell Hair Extension"),
            ('GRASS', "Grass", "Use Maxwell Grass Extension")
        ],
        default='HAIR'
    )
    p_type = EnumProperty(
        name="Type",
        items=[
            ('MESHER', "Mesher", "Use Maxwell Mesher Extension"),
            ('PARTICLES', "Particles", "Use Maxwell Particles Extension"),
            ('VOLUMETRIC', "Volumetric", "Use Maxwell Volumetric Extension")
        ],
        default='PARTICLES'
    )
    hair = PointerProperty(type=MaxwellHairExtProperties)
    grass = PointerProperty(type=MaxwellGrassExtProperties)
    mesher = PointerProperty(type=MaxwellMesherExtProperties)
    particles = PointerProperty(type=MaxwellParticlesExtProperties)
    volumetric = PointerProperty(type=MaxwellVolumetricExtProperties)
    scatter = PointerProperty(type=MaxwellScatterExtProperties)

    @classmethod
    def register(cls):
        ParticleSettings.maxwell = PointerProperty(type=cls)

    @classmethod
    def unregiser(cls):
        del ParticleSettings.maxwell
