 # -*- coding: utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Ildar Nikolaev"
__email__ = "nildar@users.sourceforge.net"

import os

from bpy.types import PropertyGroup, Scene
from bpy.props import (
    BoolProperty,
    EnumProperty,
    IntProperty,
    StringProperty,
    FloatProperty,
    FloatVectorProperty,
    CollectionProperty,
    PointerProperty
)


class MaxwellChannelProperties(PropertyGroup):
    enabled = BoolProperty(
        name="Enabled"
    )
    format = EnumProperty(
        name="Format",
        items=[
            ('png_8', "PNG 8", ""),
            ('png_16', "PNG 16", ""),
            ('tga', "TGA", ""),
            ('tif_8', "TIF 8", ""),
            ('tif_16', "TIF 16", ""),
            ('tif_32', "TIF 32", ""),
            ('exr', "EXR", "")
        ],
        default='png_8'
    )

    def enable(self, value):
        self.enabled = value


class MaxwellChannelsProperties(PropertyGroup):
    output = EnumProperty(
        name="Output Mode",
        items=[
            ('0', "Separate", "Separate"),
            ('1', "Embedded", "Embedded")
        ],
        default='0'
    )
    render_enabled = BoolProperty(
        name="Render",
        default=True,
        description="To access in compositor enable Combined Pass"
    )
    render_type = EnumProperty(
        name="Type",
        items=[
            ('0', "Diffuse + Reflections", "Diffuse + Reflections"),
            ('1', "Diffuse", "Diffuse"),
            ('2', "Reflections", "Reflections")
        ],
        default='0'
    )

    alpha_enabled = BoolProperty(
        name="Alpha",
        description="To access in compositor enable Combined Pass",
        get=lambda s: s.alpha.enabled,
        set=lambda s, v: s.alpha.enable(v)
    )
    alpha = PointerProperty(
        name="Alpha",
        type=MaxwellChannelProperties
    )
    alpha_opaque = BoolProperty(
        name="Opaque"
    )

    z_buffer_enabled = BoolProperty(
        name="Z-Buffer",
        description="To access in compositor enable Z Pass",
        get=lambda s: s.z_buffer.enabled,
        set=lambda s, v: s.z_buffer.enable(v)
    )
    z_buffer = PointerProperty(
        name="Z-Buffer",
        type=MaxwellChannelProperties
    )
    z_range = FloatVectorProperty(
        name="Z-Buffer Range",
        default=(0.0, 1.0),
        min=-100000.0, max=100000.0,
        soft_min=-100000.0, soft_max=100000.0,
        step=100,
        size=2
    )

    shadow_enabled = BoolProperty(
        name="Shadow",
        description="To access in compositor enable Shadow Pass",
        get=lambda s: s.shadow.enabled,
        set=lambda s, v: s.shadow.enable(v)
    )
    shadow = PointerProperty(
        name="Shadow",
        type=MaxwellChannelProperties
    )

    material_id_enabled = BoolProperty(
        name="Material ID",
        description="To access in compositor enable Material Index Pass",
        get=lambda s: s.material_id.enabled,
        set=lambda s, v: s.material_id.enable(v)
    )
    material_id = PointerProperty(
        name="Material ID",
        type=MaxwellChannelProperties
    )

    object_id_enabled = BoolProperty(
        name="Object ID",
        description="To access in compositor enable Object Index Pass",
        get=lambda s: s.object_id.enabled,
        set=lambda s, v: s.object_id.enable(v)
    )
    object_id = PointerProperty(
        name="Object ID",
        type=MaxwellChannelProperties
    )

    motion_enabled = BoolProperty(
        name="Motion Vector",
        description="To access in compositor enable Vector Pass",
        get=lambda s: s.motion.enabled,
        set=lambda s, v: s.motion.enable(v)
    )
    motion = PointerProperty(
        name="Motion Vector",
        type=MaxwellChannelProperties
    )

    roughness = PointerProperty(
        name="Roughness",
        type=MaxwellChannelProperties
    )
    fresnel = PointerProperty(
        name="Fresnel",
        type=MaxwellChannelProperties
    )

    normals_enabled = BoolProperty(
        name="Normals",
        description="To access in compositor enable Normal Pass",
        get=lambda s: s.normals.enabled,
        set=lambda s, v: s.normals.enable(v)
    )
    normals = PointerProperty(
        name="Normals",
        type=MaxwellChannelProperties
    )
    normals_space = EnumProperty(
        name="Normals space",
        items=[
            ('0', "World", ""),
            ('1', "Camera", "")
        ],
        default='0'
    )

    position = PointerProperty(
        name="Position",
        type=MaxwellChannelProperties
    )
    position_space = EnumProperty(
        name="Position space",
        items=[
            ('0', "World", ""),
            ('1', "Camera", "")
        ],
        default='0'
    )
    # TODO: Deep
    #
    uv_enabled = BoolProperty(
        name="UV",
        description="To access in compositor enable UV Pass",
        get=lambda s: s.uv.enabled,
        set=lambda s, v: s.uv.enable(v)
    )
    uv = PointerProperty(
        name="UV",
        type=MaxwellChannelProperties
    )
    # TODO: Custom Alpha
    #


class MaxwellSearchingPath(PropertyGroup):
    name = StringProperty(
        name="Searching path",
        description="Searching path",
        subtype='DIR_PATH'
    )


class MaxwellOverlayTextProperties(PropertyGroup):
    enabled = BoolProperty(
        name="Enable overlay text"
    )
    text = StringProperty(
        name="Text"
    )
    position = EnumProperty(
        name="Position",
        items=[
            ('0', "Bottom", "Bottom"),
            ('1', "Top", "Top")
        ]
    )
    color = FloatVectorProperty(
        name="Color",
        default=(26/255, 26/255, 26/255),
        min=0.0, max=1.0,
        subtype='COLOR'
    )
    background = FloatVectorProperty(
        name="Background",
        default=(178/255, 178/255, 178/255),
        min=0.0, max=1.0,
        subtype='COLOR'
    )
    background_enabled = BoolProperty(
        name="Background enabled",
        description="Enable background",
        default=True
    )


class MaxwellParamsProperties(PropertyGroup):
    quality = EnumProperty(
        name="Quality",
        items=[
            ('RS0', "Draft", "Draft"),
            ('RS1', "Production", "Production")
        ],
        default='RS1'
    )
    threads = IntProperty(
        name="Threads",
        default=0,
        min=0,
        soft_min=0
    )
    time = IntProperty(
        name="Time",
        default=10
    )
    sl = FloatProperty(
        name="SL",
        default=12.0
    )
    multilight = EnumProperty(
        name="Multilight",
        items=[
            ('0', "Disabled", "No multilight"),
            ('1', "Intensity", "Intensity multilight"),
            ('2', "Color", "Intensity + Color multilight")
        ],
        default='0')
    separate = EnumProperty(
        name="Separate Lights",
        items=[
            ('0', "Composite", "Composite"),
            ('1', "Separate", "Separate")
        ],
        default='0'
    )
    # Globals
    globals_collapsed = BoolProperty(
        default=True
    )
    motionblur = BoolProperty(
        name="Motion Blur",
        description="Motion Blur")
    mbsteps = IntProperty(
        name="Steps",
        description="Motion Blur Steps",
        default=1,
        min=1, max=10000,
        soft_min=1, soft_max=100
    )
    displacement = BoolProperty(
        name="Displacement",
        description="Displacement",
        default=True
    )
    dispersion = BoolProperty(
        name="Dispersion",
        description="Dispersion",
        default=True
    )
    # Illumination & Caustics
    ic_collapsed = BoolProperty(
        default=True
    )
    illumination = EnumProperty(
        name="Illumination",
        items=[
            ('0', "None", "No Illumination Layer"),
            ('1', "Direct", "Direct Illumination Layer"),
            ('2', "Indirect", "Indirect Illumination Layer"),
            ('3', "Both", "Direct + Indirect Illumination Layer")
        ],
        default='3'
    )
    reflcaustics = EnumProperty(
        name="Refl.Caustics",
        items=[
            ('0', "None", "No Reflection Caustics Layer"),
            ('1', "Direct", "Direct Reflection Caustics Layer"),
            ('2', "Indirect", "Indirect Reflection Caustics Layer"),
            ('3', "Both", "Direct + Indirect Reflection Caustics Layer")
        ],
        default='3'
    )
    refrcaustics = EnumProperty(
        name="Refr.Caustics",
        items=[
            ('0', "None", "No Refraction Caustics Layer"),
            ('1', "Direct", "Direct Refraction Caustics Layer"),
            ('2', "Indirect", "Indirect Refraction Caustics Layer"),
            ('3', "Both", "Direct + Indirect Refraction Caustics Layer")
        ],
        default='3'
    )
    # Materials
    mats_collapsed = BoolProperty(
        default=True
    )
    ovrmat = StringProperty(
        name="Override Material",
        description="Sets the path of the material that overrides all "
                    "the materials of the scene (except emitter materials)",
        subtype='FILE_PATH'
    )
    defmat = StringProperty(
        name="Default Material",
        default=os.path.join(
            os.environ.get("MAXWELL3_MATERIALS_DATABASE",
                           "Maxwell 3" if os.sys.platform == "darwin" else ""),
            "mxm files", "default.mxm"),
        description="Sets the path of the material that will be used for all the objects/triangles without material",
        subtype='FILE_PATH'
    )
    searching_paths = CollectionProperty(
        type=MaxwellSearchingPath
    )
    active_sp_index = IntProperty(
        name="Active searching path",
        default=-1,
        min=-1
    )
    # Tone Mapping
    colorspace = EnumProperty(
        name="Color Space",
        items=[
            ('0', "sRGB IEC61966-2.1", "sRGB IEC61966-2.1"),
            ('1', "Adobe RGB 98", "Adobe RGB 98"),
            ('2', "Apple RGB / SGI", "Apple RGB / SGI"),
            ('3', "PAL / SECAM (EBU3213)", "PAL / SECAM (EBU3213)"),
            ('4', "NTSC 1953", "NTSC 1953"),
            ('5', "NTSC 1979 (SMPTE-C)", "NTSC 1979 (SMPTE-C)"),
            ('6', "Wide Gamut RGB", "Wide Gamut RGB"),
            ('7', "ProPhoto RGB (ROMM)", "ProPhoto RGB (ROMM)"),
            ('8', "ECI RGB", "ECI RGB"),
            ('9', "CIE 1931", "CIE 1931"),
            ('10', "Bruce RGB", "Bruce RGB"),
            ('11', "ColorMatch RGB", "ColorMatch RGB"),
            ('12', "Best RGB", "Best RGB"),
            ('13', "Don RGB 4", "Don RGB 4"),
            ('14', "HDTV (Rec.709)", "HDTV (Rec.709)")
        ],
        default='0'
    )
    gamma = FloatProperty(
        name="Gamma",
        description="Monitor Gamma",
        default=2.2,
        min=0.1, max=3.5,
        soft_min=0.1, soft_max=3.5,
        precision=1,
        step=0.1
    )
    burn = FloatProperty(
        name="Burn",
        default=0.8,
        min=0.0, max=1.0,
        soft_min=0.0, soft_max=1.0,
        precision=1,
        step=0.1
    )
    doSharpness = BoolProperty(
        name="Enable Sharpness"
    )
    sharpness = FloatProperty(
        name="Sharpness",
        default=60.0,
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0,
        precision=2,
        step=1,
        subtype='PERCENTAGE'
    )
    # Simulens
    doDiffraction = BoolProperty(
        name="Enable Diffraction"
    )
    diffraction = FloatProperty(
        name="Diffraction",
        default=1250.0,
        min=0.0, max=2500.0,
        soft_min=0.0, soft_max=2500.0,
        precision=2,
        step=100
    )
    frequency = FloatProperty(
        name="Frequency",
        default=1250.0,
        min=0.0, max=2500.0,
        soft_min=0.0, soft_max=2500.0,
        precision=2,
        step=100
    )
    aperture = StringProperty(
        name="Aperture",
        description="Aperture Map",
        subtype='FILE_PATH'
    )
    obstacle = StringProperty(
        name="Obstacle",
        description="Obstacle Map",
        subtype='FILE_PATH'
    )
    doScattering = BoolProperty(
        name="Enable Scattering"
    )
    scattering = FloatProperty(
        name="Scattering",
        default=0.0,
        min=0.0, max=2500.0,
        soft_min=0.0, soft_max=2500.0,
        precision=2,
        step=100
    )
    doDevignetting = BoolProperty(
        name="Enable Devignetting"
    )
    devignetting = FloatProperty(
        name="Devignetting",
        default=100.0,
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0,
        precision=2,
        step=100,
        subtype='PERCENTAGE'
    )
    channels = PointerProperty(
        name="Channels",
        type=MaxwellChannelsProperties
    )
    # Overlay Text
    overlay_text = PointerProperty(
        name="Overlay Text",
        type=MaxwellOverlayTextProperties
    )
    # Fire
    fire_sl = FloatProperty(
        name="SL",
        default=7.0
    )
    fire_time = IntProperty(
        name="Time",
        default=20
    )
    fire_floating_shadows = BoolProperty(
        name="Floating Shadows"
    )
    fire_floating_reflections = BoolProperty(
        name="Floating Reflections"
    )
    #fire_quality = IntProperty(
    #    name="Quality",
    #    default=10
    #)
    #fire_p = IntProperty(
    #    name="",  # Unknown
    #)


class MaxwellRenderProperties(PropertyGroup):
    run = EnumProperty(
        name="Open with",
        description="After export open file in program",
        items=[
            ('none', "None", "Only export scene to .mxs file"),
            ('studio', "Maxwell Studio", "After export open .mxs file in Maxwell Studio"),
            ('render', "Maxwell Render", "After export open .mxs file in Maxwell Render"),
            ('network', "Maxwell Network", "After export run Maxwell Network Monitor and add scene to it queue")
        ],
        default='studio'
    )
    run_manager = BoolProperty(
        name="Run Network Manager",
        default=True
    )
    run_node = BoolProperty(
        name="Run Network Node",
        default=True
    )
    manager = StringProperty(
        name="Network Manager"
    )
    cmdline = StringProperty(
        name="Command Line"
    )
    protected = BoolProperty(
        name="Protect MXS",
        default=True
    )
    rp = PointerProperty(
        type=MaxwellParamsProperties
    )

    def _cmdline_get_nogui(self):
        return "-nogui" in self.cmdline.split()

    def _cmdline_set_nogui(self, value):
        v = [c for c in self.cmdline.split() if c != "-nogui"]
        if value:
            v.append("-nogui")
        self.cmdline = " ".join(v)

    def _cmdline_get_nowait(self):
        return "-nowait" in self.cmdline.split()

    def _cmdline_set_nowait(self, value):
        v = [c for c in self.cmdline.split() if c != "-nowait"]
        if value:
            v.append("-nowait")
        self.cmdline = " ".join(v)

    def _cmdline_get_nogui_nowait(self):
        t = self.cmdline.split()
        return "-nogui" in t and "-nowait" in t

    def _cmdline_set_nogui_nowait(self, value):
        v = [c for c in self.cmdline.split() if c != "-nogui" and c != "-nowait"]
        if value:
            v.append("-nogui")
            v.append("-nowait")
        self.cmdline = " ".join(v)

    def _cmdline_get_trytoresume(self):
        return "-trytoresume" in self.cmdline.split()

    def _cmdline_set_trytoresume(self, value):
        v = [c for c in self.cmdline.split() if c != "-trytoresume"]
        if value:
            v.append("-trytoresume")
        self.cmdline = " ".join(v)

    cmdline_nogui = BoolProperty(
        name="No GUI",
        description="Console view on. Display off",
        get=_cmdline_get_nogui,
        set=_cmdline_set_nogui
    )
    cmdline_nowait = BoolProperty(
        name="No Wait",
        description="Close the rendering application after the render finishes and returns control to the console",
        get=_cmdline_get_nowait,
        set=_cmdline_set_nowait
    )
    cmdline_nogui_nowait = BoolProperty(
        name="No GUI & No Wait",
        description="Run the rendering application in console mode and close it after the render finishes",
        get=_cmdline_get_nogui_nowait,
        set=_cmdline_set_nogui_nowait
    )
    cmdline_trytoresume = BoolProperty(
        name="Try to Resume",
        description="Maxwell will automatically resume the render if the MXI file exists",
        get=_cmdline_get_trytoresume,
        set=_cmdline_set_trytoresume
    )

    @classmethod
    def register(cls):
        Scene.maxwell = PointerProperty(type=cls)

    @classmethod
    def unregiser(cls):
        del Scene.maxwell
