# -*- coding: utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Ildar Nikolaev"
__email__ = "nildar@users.sourceforge.net"

import bpy
from bpy.types import PropertyGroup, ImageTexture
from bpy.props import (
    EnumProperty,
    BoolProperty,
    StringProperty,
    IntProperty,
    IntVectorProperty,
    FloatProperty,
    FloatVectorProperty,
    PointerProperty,
    CollectionProperty
)


class MaxwellBakeTextureProperties(PropertyGroup):
    def update_name(self, context):
        mx = context.texture.maxwell
        if self.name:
            if mx.bake_textures.keys().count(self.name) > 1:
                bpy.ops.maxwell.report(
                    type={'WARNING'},
                    text="'%s' already exists in the textures for baking" % self.name)
                self["name"] = ""
            if mx.bake_textures.get(context.texture.name):
                bpy.ops.maxwell.report(
                    type={'WARNING'},
                    text="can't add self to the textures for baking")
                self["name"] = ""

    enabled = BoolProperty(
        name="Enabled",
        default=True
    )
    name = StringProperty(
        name="Texture",
        update=update_name
    )


class Projection:
    channel = IntProperty(
        name="Channel",
        description="Specify the UV set to be used for this texture",
        default=0,
        min=0, max=254,
        soft_min=0, soft_max=254
    )
    tile_x = BoolProperty(
        name="Tile X",
        description="Tile your texture in X axis, Y axis, both the X and Y axis, "
                    "or do not use any tiling at all (no repeat)",
        default=True
    )
    tile_y = BoolProperty(
        name="Tile Y",
        description="Tile your texture in X axis, Y axis, both the X and Y axis, "
                    "or do not use any tiling at all (no repeat)",
        default=True
    )
    units = EnumProperty(
        name="Relative/Meters",
        description="The amount of tiling for a texture can be set in the texture "
                    "coordinates (Relative), or in real scale in meters (Meters)",
        items=[
            ('0', "Relative", "", 0),
            ('1', "Meters", "", 1)
        ],
        default='0'
    )
    mirror_x = BoolProperty(
        name="Mirror X"
    )
    mirror_y = BoolProperty(
        name="Mirror Y"
    )
    repeat = FloatVectorProperty(
        name="Repeat",
        description="Specify the tiling amount (number of repetitions of the texture) in the desired axis",
        default=(1.0, 1.0),
        min=-1000.0, max=1000.0,
        soft_min=-1000.0, soft_max=1000.0,
        precision=3,
        size=2,
        step=100.0,
        subtype = 'XYZ'
    )
    offset = FloatVectorProperty(
        name="Offset",
        description="Select the amount of offset for the X, Y axis",
        default=(0.0, 0.0),
        min=-1000.0, max=1000.0,
        soft_min=-1000.0, soft_max=1000.0,
        precision=3,
        size=2,
        step=10.0,
        subtype='XYZ'
    )
    rotation = FloatProperty(
        name="Rotation",
        default=0,
        min=0.0, max=360.0,
        soft_min=0.0, soft_max=360.0,
        precision=3,
        step=100.0
    )


class MaxwellImageProperties(PropertyGroup, Projection):
    # Projection Properties
    override = BoolProperty(
        name="Use Override Map"
    )
    # Image Properties
    invert = BoolProperty(
        name="Invert",
        description="Invert the loaded texture. This is useful for black & white textures used as a weightmap or mask"
    )
    alpha_only = BoolProperty(
        name="Alpha Only",
        description="On RGBA images (images that contain its Alpha channel embedded), "
                    "this option considers only the Alpha channel, mainly for masking purposes"
    )
    interpolation = BoolProperty(
        name="Interpolation",
        description="Turning this option on applies filtering to a texture, which may be useful "
                    "to avoid pixelization when rendering close-ups of smaller textures"
    )
    brightness = FloatProperty(
        name="Brightness",
        default=0,
        min=-100.0, max=100.0,
        soft_min=-100.0, soft_max=100.0,
        precision=3
    )
    contrast = FloatProperty(
        name="Contrast",
        default=0,
        min=-100.0, max=100.0,
        soft_min=-100.0, soft_max=100.0,
        precision=3
    )
    saturation = FloatProperty(
        name="Saturation",
        default=0,
        min=-100.0, max=100.0,
        soft_min=-100.0, soft_max=100.0,
        precision=3
    )
    hue = FloatProperty(
        name="Hue",
        default=0,
        min=-180.0, max=180.0,
        soft_min=-180.0, soft_max=180.0,
        precision=3
    )
    clamp = IntVectorProperty(
        name="RGB Clamp",
        description="Adjust the levels of your image. This function enables you to specify "
                    "the maximum darkest or brightest values in the texture",
        default=(0, 255),
        min=0, max=255,
        soft_min=0, soft_max=255,
        size=2
    )
    # Normal maps
    flip_x = BoolProperty(
        name="Flip X"
    )
    flip_y = BoolProperty(
        name="Flip Y",
        default=True
    )
    wide = BoolProperty(
        name="Wide"
    )
    # Bake
    bake_textures = CollectionProperty(
        name="Bake textures",
        type=MaxwellBakeTextureProperties
    )
    active_bt_index = IntProperty(
        name="Active Texture",
        default=-1,
        min=-1
    )
    bake_color = FloatVectorProperty(
        name="Base color",
        description="Base color for baking textures",
        default=(0.0, 0.0, 0.0),
        min=0.0, max=1.0,
        soft_min=0.0, soft_max=1.0,
        subtype='COLOR'
    )
    bake_save = BoolProperty(
        name="Save Image",
        description="Save image after baking",
        default=True
    )
    bake_clear = BoolProperty(
        name="Clear Image",
        description="Clear Image before baking",
        default=True,
    )

    @classmethod
    def register(cls):
        ImageTexture.maxwell = PointerProperty(type=cls)

    @classmethod
    def unregiser(cls):
        del ImageTexture.maxwell
