# -*- coding: utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Ildar Nikolaev"
__email__ = "nildar@users.sourceforge.net"

import sys
import time as _tm
from math import pi as _pi

from bpy.types import PropertyGroup, World
from bpy.props import (
    FloatProperty,
    FloatVectorProperty,
    EnumProperty,
    StringProperty,
    IntProperty,
    BoolProperty,
    PointerProperty
)

from ..engine import abspath


class MaxwellConstantSkyProperties(PropertyGroup):
    luminance = FloatProperty(
        name="Intensity",
        description="Intensity of the Sky Dome in cd/m2",
        default=10000.0,
        min=0.1, max=1000000.0,
        soft_min=0.1, soft_max=1000000.0,
        precision=3,
        step=1000,
        subtype='UNSIGNED'
    )
    zenith = FloatVectorProperty(
        name="Zenith",
        description="Color of the dome at the zenith (the highest point above our head)",
        default=(1.0, 1.0, 1.0),
        min=0.0, max=1.0,
        soft_min=0.0, soft_max=1.0,
        subtype='COLOR'
    )
    horizon = FloatVectorProperty(
        name="Horizon",
        description="Color of the dome at the horizon",
        default=(1.0, 1.0, 1.0),
        min=0.0, max=1.0,
        soft_min=0.0, soft_max=1.0,
        subtype='COLOR'
    )
    controlPoint = FloatProperty(
        name="Mid Point",
        description="Transition between the Zenith and Horizon colors (in degrees)",
        default=45.0,
        min=0.0, max=90.0,
        soft_min=0.0, soft_max=90.0,
        precision=1,
        step=100
    )

    def __call__(self):
        return {
            "luminance": self.luminance,
            "zenith": self.zenith,
            "horizon": self.horizon,
            "controlPoint": self.controlPoint
        }


class MaxwellPhysicalSkyProperties(PropertyGroup):
    intensity = FloatProperty(
        name="Intensity",
        default=1.0,
        min=0.0, max=10000.0,
        soft_min=0.0, soft_max=10000,
        precision=2,
        step=1
    )
    planetRefl = FloatProperty(
        name="Planet Reflectance",
        description="Controls the percentage of light reflected from the planet surface back into the atmosphere",
        default=25.0,
        min=1.0, max=100.0,
        subtype='PERCENTAGE',
        precision=2,
        step=10
    )
    ozone = FloatProperty(
        name="Ozone",
        description="The amount of ozone gas in the atmosphere",
        default=0.4,
        min=0.0,  # max = 1000000.0,
        soft_min=0.0, soft_max=1.0,
        precision=3,
        step=10
    )
    water = FloatProperty(
        name="Water",
        description="The amount of water vapor in the atmosphere",
        default=2.0,
        min=0.0,  # max=1000000.0,
        soft_min=0.0, soft_max=10.0,
        precision=3,
        step=10
    )
    angstrom = FloatProperty(
        name="Turbidity Coefficient",
        description="Defines the concentration and amount of particles in the atmosphere",
        default=0.04,
        min=0.0,  # max=1000000.0,
        soft_min=0.0, soft_max=1.0,
        precision=3,
        step=10
    )
    wavelength = FloatProperty(
        name="Wavelength Exponent",
        description="Defines the average size of the particles in the atmosphere",
        default=1.2,
        min=0.0,  # max=1000000.0,
        soft_min=0.0, soft_max=10.0,
        precision=3,
        step=10
    )
    albedo = FloatProperty(
        name="Reflectance",
        description="Refers to the albedo of the aerosols, "
                    "or the rate of energy scattered and absorbed by the aerosols",
        default=80.0,
        min=0.0, max=100.0,
        soft_min=0.0, soft_max=100.0,
        precision=2,
        step=1,
        subtype='PERCENTAGE'
    )
    asymmetry = FloatProperty(
        name="Asymmetry",
        description="This factor control the anisotropy of the particles, "
                    "that is, in which direction most of light will be scattered",
        default=0.7,
        min=-0.99999, max=0.99999,
        soft_min=-0.99999, soft_max=0.99999,
        precision=5,
        step=1
    )

    def __call__(self):
        return {
            "intensity": self.intensity,
            "ozone": self.ozone,
            "water": self.water,
            "angstrom": self.angstrom,
            "wavelength": self.wavelength,
            "albedo": self.albedo / 100,
            "asymmetry": self.asymmetry,
            "reflectance": self.planetRefl / 100
        }


class MaxwellSunProperties(PropertyGroup):
    type = EnumProperty(
        name="Type",
        description="Sun type",
        items=[
            ('0', "Disabled", "Sun disabled"),
            ('1', "Physical", "Physical"),
            ('2', "Custom", "Custom")
        ],
        default='1'
    )
    power = FloatProperty(
        name="Power",
        description="A multiplier that controls the amount of light emitted from the sun",
        default=1.0,
        min=sys.float_info.min,
        soft_min=0.01, soft_max=100.0,
        precision=3,
        step=10
    )
    radius = FloatProperty(
        name="Radius Factor",
        description="Radius factor",
        default=1.0,
        min=0.01, max=1000.0,
        soft_min=0.01, soft_max=1000.0
    )
    temperature = FloatProperty(
        name="Temperature",
        description="Temperature of the sun's spectral radiation",
        default=5777.0,
        min=100.0, max=1000000.0,
        soft_min=100.0, soft_max=10000.0, # 1-10000 studio limits
        precision=3,
        step=1000
    )
    color = FloatVectorProperty(
        name="Color",
        description="Color of the sun",
        default=(1.0, 1.0, 1.0),
        min=0.0, max=1.0,
        soft_min=0.0, soft_max=1.0,
        subtype='COLOR'
    )
    position = EnumProperty(
        name="Position",
        description="Way to set the sun position",
        items=[
            ('0', "Lat/Lon", "Set the earth position to calculate the sky/sun light direction"),
            ('1', "Angles", "Set the sun position based on a spherical angles"),
            ('2', "Direction", "Set the sun position based on a direction")
        ],
        default='0'
    )
    latitude = FloatProperty(
        name="Latitude",
        description="Latitude",
        default=0.0,
        min=-90.0, max=90.0,
        soft_min=-90.0, soft_max=90.0,
        precision=3,
        step=1000,
    )
    longitude = FloatProperty(
        name="Longitude",
        description="Longitude",
        default=0.0,
        min=-180.0, max=180.0,
        soft_min=-180.0, soft_max=180.0,
        precision=3,
        step=1000,
    )

    def _get_time(self):
        return _tm.strftime("%H:%M:%S", _tm.gmtime(round(self.time * 3600)))

    def _set_time(self, value):
        t = _tm.strptime(value, "%H:%M:%S")
        self.time = (t.tm_hour*3600 + t.tm_min*60 + t.tm_sec) / 3600

    def _get_date(self):
        t = (_tm.localtime().tm_year, 1, self.day, 0, 0, 0, 0, 0, -1)
        return _tm.strftime("%Y-%m-%d", _tm.localtime(_tm.mktime(t)))

    def _set_date(self, value):
        self.day = _tm.strptime(value, "%Y-%m-%d").tm_yday

    day = IntProperty(
        default=100
    )
    time = FloatProperty(
        default=17.0
    )
    ui_date = StringProperty(
        name="Date",
        description="Set the date",
        get=_get_date,
        set=_set_date,
        maxlen=10
    )
    ui_time = StringProperty(
        name="Time",
        description="Set the time",
        get=_get_time,
        set=_set_time,
        maxlen=8
    )
    sm = IntProperty(
        name="GMT",
        description="GMT offset",
        default=0,  # -int((_tm.timezone / (60 * 60)))
        min=-12, max=12,
        soft_min=-12, soft_max=12,
    )
    rotation = FloatProperty(
        name="Ground Rotation",
        description="Allow you rotate the north direction (radians)",
        default=0.0,
        min=0.0, max=_pi * 2,
        soft_min=0.0, soft_max=_pi * 2,
        precision=3,
        step=100,
        subtype='ANGLE'
    )
    zenith = FloatProperty(
        name="Zenith",
        default=_pi / 4,
        min=0.0, max=_pi / 2,
        soft_min=0.0, soft_max=_pi / 2,
        precision=4,
        step=1,
        subtype='ANGLE'
    )
    azimuth = FloatProperty(
        name="Azimuth",
        default=_pi / 4,
        min=0.0, max=_pi * 2,
        soft_min=0.0, soft_max=_pi * 2,
        precision=4,
        step=1,
        subtype='ANGLE'
    )
    direction = FloatVectorProperty(
        name="Direction",
        size=3,
        default=(0.0, 0.0, 1.0),
        min=-1000.0, max=1000.0,
        soft_min=-1000.0, soft_max=1000.0,
        precision=4,
        step=1000,
        subtype='XYZ'
    )
    dir_object = StringProperty(
        name="Direction object",
        description="Set the sun position based on a object"
    )


class LayerProperties:
    type = EnumProperty(
        name="Type",
        items=[
            ('0', "Disabled", ""),
            ('1', "HDR Image", ""),
            ('2', "Active Sky", "")
        ],
        default='1'
    )
    map = StringProperty(
        name="Map",
        description="Select the MXI/HDR/EXR map",
        subtype='FILE_PATH'
    )
    intensity = FloatProperty(
        name="Intensity",
        description="Adjust the intensity of the map",
        default=1.0,
        min=0.0, max=1000.0,
        soft_min=0.0, soft_max=1000.0,
        precision=3,
        step=1000,
        subtype='UNSIGNED'
    )
    scale = FloatVectorProperty(
        name="Scale",
        description="Scale the map",
        size=2,
        default=(1.0, 1.0),
        min=-1000.0, max=1000.0,
        soft_min=-1000.0, soft_max=1000.0,
        precision=4,
        step=1000
    )
    offset = FloatVectorProperty(
        name="Offset",
        description="Rotate the spherical environment in the X and Y axis",
        size=2,
        default=(0.0, 0.0),
        min=-360.0, max=360.0,
        soft_min=-360.0, soft_max=360.0,
        precision=3,
        step=1000,
    )

    def __call__(self, interpolate):
        return {
            "filename": abspath(self.map),
            "intensity": self.intensity,
            "interpolate": interpolate,
            "offset": self.offset,
            "scale": self.scale,
            "state": int(self.type)
        }


class MaxwellELProperties(PropertyGroup, LayerProperties):
    type = EnumProperty(
        name="Type",
        items=[
            ('0', "Disabled", ""),
            ('1', "HDR Image", ""),
            ('2', "Active Sky", ""),
            ('3', "Same as Background", "The same map and settings will be used as for the Background channel"),
        ],
        default='3'
    )


class MaxwellELBackgroundProperties(PropertyGroup, LayerProperties):
    pass


class MaxwellIblProperties(PropertyGroup):
    intensity = FloatProperty(
        name="Intensity",
        default=1.0,
        min=0.0, max=1000000.0,
        soft_min=0.0, soft_max=1000000.0,
        precision=1,
        step=1
    )
    interpolation = BoolProperty(
        name="Interpolation",
        description="This option can be useful when you have a low resolution HDR map "
                    "and see pixelisation in the HDR map when it is seen in the background"
    )
    screen_mapping = BoolProperty(
        name="Screen Mapping",
        description="Screen mapping can be used to map the MXI/HDR/EXR image to screen coordinates"
    )
    background = PointerProperty(type=MaxwellELBackgroundProperties)
    reflection = PointerProperty(type=MaxwellELProperties)
    refraction = PointerProperty(type=MaxwellELProperties)
    illumination = PointerProperty(type=MaxwellELProperties)


class MaxwellEnvironmentProperties(PropertyGroup):
    type = EnumProperty(
        name="Environment",
        description="Sky/IBL Environment",
        items=[
            ('none', "None", "No Environment lighting will be used"),
            ('sky', "Sun & Sky", "An sky model will be created"),
            ('ibl', "Image Based Lighting", "Image Based Lighting allows you to light your scene "
                                            "by applying an HDR image to a virtual sphere that encompasses your scene")
        ],
        default='sky'
    )
    sky = EnumProperty(
        name="Sky Type",
        description="Sky Type",
        items=[
            ('physical', "Physical Sky", "Atmosphere model that reproduce the skylight "
                                         "at different locations/times/dates"),
            ('constant', "Constant Dome", "Create a consistent color or gradient dome for lighting your scene")
        ],
        default='physical'
    )
    constant = PointerProperty(type=MaxwellConstantSkyProperties)
    physical = PointerProperty(type=MaxwellPhysicalSkyProperties)
    sun = PointerProperty(type=MaxwellSunProperties)
    ibl = PointerProperty(type=MaxwellIblProperties)

    @classmethod
    def register(cls):
        World.maxwell = PointerProperty(type=cls)

    @classmethod
    def unregister(cls):
        del World.maxwell
