# -*- coding: utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Ildar Nikolaev"
__email__ = "nildar@users.sourceforge.net"

from bpy.types import Panel, Menu
from bl_ui.properties_material import MaterialButtonsPanel

from ..engine import MaxwellRenderEngine


class MaxwellOpaquePresetsMenu(Menu):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_idname = "MAXWELL_MT_material_opaque_presets"
    bl_label = "Preset"

    preset_operator = "script.execute_preset"
    preset_subdir = "bmaxwell/material/opaque"

    draw = Menu.draw_preset


class MaxwellTransparentPresetsMenu(Menu):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_idname = "MAXWELL_MT_material_transparent_presets"
    bl_label = "Preset"

    preset_operator = "script.execute_preset"
    preset_subdir = "bmaxwell/material/transparent"

    draw = Menu.draw_preset


class MaxwellMetalPresetsMenu(Menu):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_idname = "MAXWELL_MT_material_metal_presets"
    bl_label = "Preset"

    preset_operator = "script.execute_preset"
    preset_subdir = "bmaxwell/material/metal"

    draw = Menu.draw_preset


class MaxwellTranslucentPresetsMenu(Menu):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_idname = "MAXWELL_MT_material_translucent_presets"
    bl_label = "Preset"

    preset_operator = "script.execute_preset"
    preset_subdir = "bmaxwell/material/translucent"

    draw = Menu.draw_preset


class MaxwellCarPaintPresetsMenu(Menu):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_idname = "MAXWELL_MT_material_carpaint_presets"
    bl_label = "Preset"

    preset_operator = "script.execute_preset"
    preset_subdir = "bmaxwell/material/carpaint"

    draw = Menu.draw_preset


class MaxwellHairPresetsMenu(Menu):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_idname = "MAXWELL_MT_material_hair_presets"
    bl_label = "Preset"

    preset_operator = "script.execute_preset"
    preset_subdir = "bmaxwell/material/hair"

    draw = Menu.draw_preset


class MaxwellMaterialsPanel(MaterialButtonsPanel, Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_options = {'HIDE_HEADER'}
    bl_label = ""

    @classmethod
    def poll(cls, context):
        # An exception, dont call the parent poll func because
        # this manages materials for all engine types

        engine = context.scene.render.engine
        return (context.material or context.object) and (engine in cls.COMPAT_ENGINES)

    def draw(self, context):
        layout = self.layout

        mat = context.material
        ob = context.object
        slot = context.material_slot
        space = context.space_data

        if ob:
            row = layout.row()

            row.template_list("MATERIAL_UL_matslots", "", ob, "material_slots", ob, "active_material_index", rows=2)

            col = row.column(align=True)
            col.operator("object.material_slot_add", icon='ZOOMIN', text="")
            col.operator("object.material_slot_remove", icon='ZOOMOUT', text="")

            col.menu("MATERIAL_MT_specials", icon='DOWNARROW_HLT', text="")

            if ob.mode == 'EDIT':
                row = layout.row(align=True)
                row.operator("object.material_slot_assign", text="Assign")
                row.operator("object.material_slot_select", text="Select")
                row.operator("object.material_slot_deselect", text="Deselect")

        split = layout.split(percentage=0.7)

        if ob:
            split.template_ID(ob, "active_material", new="material.new")
            row = split.row()

            if slot:
                row.prop(slot, "link", text="")
            else:
                row.label()
        elif mat:
            split.template_ID(space, "pin_id")
            split.separator()


class MaxwellMaterialPanel(MaterialButtonsPanel, Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_label = "Maxwell Material"

    def draw(self, context):
        layout = self.layout

        mat = context.material
        layout.context_pointer_set("maxwell_ui", mat)

        mx = mat.maxwell
        mx_type = mx.type

        layout.prop(mx, "type")
        if mx_type != 'file':
            row = layout.split(0.33)
            row.label("Material ID:")
            row = row.row(align=True)
            if mx.pi_as_color:
                row.prop(mx, "pi_color", text="")
                row.prop(mat, "pass_index")
            else:
                row.prop(mx, "color_id", text="")
                subrow = row.row(align=True)
                subrow.prop(mat, "pass_index")
                subrow.enabled = False
            row.prop(mx, "pi_as_color", text="")

        # Globals
        split = layout.split(0.5, align=True)
        split.prop_search(mx, "bump_tex", mat, "texture_slots", text="", icon='TEXTURE')
        split = split.split(0.85, align=True)
        split.prop(mx, "bump", text="Normal" if mx.bump_normal else "Bump")
        split.prop(mx, "bump_normal", text="N", toggle=True)
        split.enabled = bool(mx.bump_tex)

        row = layout.row()
        col = row.column()
        col.prop(mx, "dispersion")
        col.prop(mx, "shadow")
        col.prop(mx, "matte")
        if mx_type != 'file':
            col = row.column()
            col.prop(mx, "auto_color", toggle=True)
            col.operator("maxwell.material_save_mxm")

        if mx_type == 'file':
            layout.prop(mx, "mxm", text="")

            row = layout.row()
            subrow = row.row(align=True)
            subrow.operator("maxwell.material_browse_mxm")
            if mx.mxm.strip():
                subrow.operator("maxwell.material_edit_mxm")
                row.operator("maxwell.material_read_mxm")
                layout.row().prop(mx, "embed")
        elif mx_type == 'custom':
            # Displacement
            displacement = mx.displacement
            displacement_path = displacement.path_from_id()
            col = layout.column(align=True)
            box = col.box()
            row = box.row()
            subrow = row.row()
            subrow.alignment = 'LEFT'
            icon = 'TRIA_DOWN' if displacement.expanded else 'TRIA_RIGHT'
            op = subrow.operator("maxwell.expand_ui", "Displacement", icon=icon, emboss=False)
            op.path = displacement_path
            op.attr = "expanded"
            row.prop(displacement, "enabled", text="")
            if displacement.expanded:
                box = col.box()
                box.enabled = displacement.enabled

                box.prop_search(displacement, "map", mat, "texture_slots", icon='TEXTURE')
                box.prop(displacement, "type")

                split = box.split(0.925, align=True)
                subsplit = split.split(0.5, align=True)
                row = subsplit.row(align=True)
                row.prop(displacement, "level")
                row.enabled = (displacement.type != '0' or not displacement.adaptive)
                subsplit.prop(displacement, "method", text="")
                split.prop(displacement, "adaptive", text="A", toggle=True)

                row = box.row()
                row.prop(displacement, "smooth")
                subrow = row.row()
                subrow.prop(displacement, "offset")
                #subrow.enabled = displacement.type != '2' # TODO: Inspect this

                box.prop(displacement, "uvi")

                row = box.row()
                row.enabled = (displacement.type != '2')
                split = row.split(0.925, align=True)
                text = "HeightMap, "
                text += "cm" if displacement.absolute else "%"
                split.prop(displacement, "height", text=text)
                split.prop(displacement, "absolute", text="A", toggle=True)

                # Vector 3D Properties
                row = box.row()
                row.alignment = 'LEFT'
                icon = 'DISCLOSURE_TRI_DOWN_VEC' if displacement.v3d_expanded else 'DISCLOSURE_TRI_RIGHT_VEC'
                op = row.operator("maxwell.expand_ui", "Vector 3D:", icon=icon, emboss=False)
                op.path = displacement_path
                op.attr = "v3d_expanded"

                if displacement.v3d_expanded:
                    enabled1 = (displacement.type == '2')
                    row = box.row()
                    row.enabled = enabled1
                    row.prop(displacement, "preset")

                    enabled2 = (enabled1 and displacement.preset == '0')
                    row = box.row()
                    row.enabled = enabled2
                    row.prop(displacement, "transform")

                    row = box.row()
                    row.enabled = enabled2
                    row.prop(displacement, "mapping")

                    row = box.row()
                    row.enabled = enabled1
                    row.prop(displacement, "scale")
        elif mx_type == 'ags':
            props = mx.ags
            ui_collapsed = props.ui_collapsed
            col = layout.column(align=True)
            box = col.box()
            row = box.row()
            row.alignment = 'LEFT'
            icon = 'TRIA_RIGHT' if ui_collapsed else 'TRIA_DOWN'
            op = row.operator("maxwell.expand_ui", "AGS", icon=icon, emboss=False)
            op.path = props.path_from_id()
            op.attr = "ui_collapsed"
            if not ui_collapsed:
                box = col.box()
                row = box.row()
                col = row.column()
                col.prop(props, "color")
                col = row.column(align=True)
                col.label("Reflection:")
                col.prop(props, "reflection", text="")
                box.prop(props, "type")
        elif mx_type == 'opaque':
            props = mx.opaque
            ui_collapsed = props.ui_collapsed
            col = layout.column(align=True)
            box = col.box()
            row = box.row()
            row.alignment = 'LEFT'
            icon = 'TRIA_RIGHT' if ui_collapsed else 'TRIA_DOWN'
            op = row.operator("maxwell.expand_ui", "Opaque", icon=icon, emboss=False)
            op.path = props.path_from_id()
            op.attr = "ui_collapsed"
            if not ui_collapsed:
                box = col.box()
                row = box.row(align=True)
                row.menu("MAXWELL_MT_material_opaque_presets")
                row.operator("maxwell.material_opaque_preset_add", text="", icon='ZOOMIN')
                row.operator("maxwell.material_opaque_preset_add", text="", icon='ZOOMOUT').remove_active = True
                row = box.row(align=True)
                row.prop(props, "color")
                row.prop_search(props, "color_map", mat, "texture_slots", text="", icon='TEXTURE')
                subrow = row.row(align=True)
                subrow.alignment = 'RIGHT'
                subrow.prop(props, "color_map_enabled", text="T", toggle=True)
                row = box.row(align=True)
                row.label("Shininess:")
                row.prop(props, "shininess", text="")
                row.prop_search(props, "shininess_map", mat, "texture_slots", text="", icon='TEXTURE')
                subrow = row.row(align=True)
                subrow.alignment = 'RIGHT'
                subrow.prop(props, "shininess_map_enabled", text="T", toggle=True)
                row = box.row(align=True)
                row.label("Roughness:")
                row.prop(props, "roughness", text="")
                row.prop_search(props, "roughness_map", mat, "texture_slots", text="", icon='TEXTURE')
                subrow = row.row(align=True)
                subrow.alignment = 'RIGHT'
                subrow.prop(props, "roughness_map_enabled", text="T", toggle=True)
                box.prop(props, "clearcoat")
        elif mx_type == 'transparent':
            props = mx.transparent
            ui_collapsed = props.ui_collapsed
            col = layout.column(align=True)
            box = col.box()
            row = box.row()
            row.alignment = 'LEFT'
            icon = 'TRIA_RIGHT' if ui_collapsed else 'TRIA_DOWN'
            op = row.operator("maxwell.expand_ui", "Transparent", icon=icon, emboss=False)
            op.path = props.path_from_id()
            op.attr = "ui_collapsed"
            if not ui_collapsed:
                box = col.box()
                row = box.row(align=True)
                row.menu("MAXWELL_MT_material_transparent_presets")
                row.operator("maxwell.material_transparent_preset_add", text="", icon='ZOOMIN')
                row.operator("maxwell.material_transparent_preset_add", text="", icon='ZOOMOUT').remove_active = True
                row = box.row(align=True)
                row.prop(props, "color")
                row.prop_search(props, "color_map", mat, "texture_slots", text="", icon='TEXTURE')
                subrow = row.row(align=True)
                subrow.alignment = 'RIGHT'
                subrow.prop(props, "color_map_enabled", text="T", toggle=True)
                row = box.row(align=True)
                row.prop(props, "ior")
                row.prop(props, "transparency")
                row = box.row(align=True)
                row.label("Roughness:")
                row.prop(props, "roughness", text="")
                row.prop_search(props, "roughness_map", mat, "texture_slots", text="", icon='TEXTURE')
                subrow = row.row(align=True)
                subrow.alignment = 'RIGHT'
                subrow.prop(props, "roughness_map_enabled", text="T", toggle=True)
                row = box.row(align=True)
                row.prop(props, "specular_tint")
                row.prop(props, "dispersion")
                box.prop(props, "clearcoat")
        elif mx_type == 'metal':
            props = mx.metal
            ui_collapsed = props.ui_collapsed
            col = layout.column(align=True)
            box = col.box()
            row = box.row()
            row.alignment = 'LEFT'
            icon = 'TRIA_RIGHT' if ui_collapsed else 'TRIA_DOWN'
            op = row.operator("maxwell.expand_ui", "Metal", icon=icon, emboss=False)
            op.path = props.path_from_id()
            op.attr = "ui_collapsed"
            if not ui_collapsed:
                box = col.box()
                row = box.row(align=True)
                row.menu("MAXWELL_MT_material_metal_presets")
                row.operator("maxwell.material_metal_preset_add", text="", icon='ZOOMIN')
                row.operator("maxwell.material_metal_preset_add", text="", icon='ZOOMOUT').remove_active = True
                split = box.split(0.25)
                split.label("Metal IOR:")
                row = split.row()
                row.prop(props, "ior", text="")
                row.prop(props, "tint")
                split = box.split(0.25)
                split.label("Color:")
                row = split.row(align=True)
                row.prop(props, "color", text="")
                row.prop_search(props, "color_map", mat, "texture_slots", text="", icon='TEXTURE')
                subrow = row.row(align=True)
                subrow.alignment = 'RIGHT'
                subrow.prop(props, "color_map_enabled", text="T", toggle=True)
                split = box.split(0.25)
                split.label("Roughness:")
                row = split.row(align=True)
                row.prop(props, "roughness", text="")
                row.prop_search(props, "roughness_map", mat, "texture_slots", text="", icon='TEXTURE')
                subrow = row.row(align=True)
                subrow.alignment = 'RIGHT'
                subrow.prop(props, "roughness_map_enabled", text="T", toggle=True)
                split = box.split(0.25)
                split.label("Anisotropy:")
                row = split.row(align=True)
                row.prop(props, "anisotropy", text="")
                row.prop_search(props, "anisotropy_map", mat, "texture_slots", text="", icon='TEXTURE')
                subrow = row.row(align=True)
                subrow.alignment = 'RIGHT'
                subrow.prop(props, "anisotropy_map_enabled", text="T", toggle=True)
                split = box.split(0.25)
                split.label("Angle:")
                row = split.row(align=True)
                row.prop(props, "angle", text="")
                row.prop_search(props, "angle_map", mat, "texture_slots", text="", icon='TEXTURE')
                subrow = row.row(align=True)
                subrow.alignment = 'RIGHT'
                subrow.prop(props, "angle_map_enabled", text="T", toggle=True)
                split = box.split(0.25)
                split.label("Dust & Dirt:")
                row = split.row(align=True)
                row.prop(props, "dust", text="")
                row.prop_search(props, "dust_map", mat, "texture_slots", text="", icon='TEXTURE')
                subrow = row.row(align=True)
                subrow.alignment = 'RIGHT'
                subrow.prop(props, "dust_map_enabled", text="T", toggle=True)
                split = box.split(0.25)
                split.label("Perforation Map:")
                row = split.row(align=True)
                row.prop_search(props, "perforation_map", mat, "texture_slots", text="", icon='TEXTURE')
                subrow = row.row(align=True)
                subrow.alignment = 'RIGHT'
                subrow.prop(props, "perforation_map_enabled", text="T", toggle=True)
        elif mx_type == 'translucent':
            props = mx.translucent
            ui_collapsed = props.ui_collapsed
            col = layout.column(align=True)
            box = col.box()
            row = box.row()
            row.alignment = 'LEFT'
            icon = 'TRIA_RIGHT' if ui_collapsed else 'TRIA_DOWN'
            op = row.operator("maxwell.expand_ui", "Translucent", icon=icon, emboss=False)
            op.path = props.path_from_id()
            op.attr = "ui_collapsed"
            if not ui_collapsed:
                box = col.box()
                row = box.row(align=True)
                row.menu("MAXWELL_MT_material_translucent_presets")
                row.operator("maxwell.material_translucent_preset_add", text="", icon='ZOOMIN')
                row.operator("maxwell.material_translucent_preset_add", text="", icon='ZOOMOUT').remove_active = True
                row = box.row()
                row.prop(props, "scale")
                row.prop(props, "ior")
                split = box.split(0.25)
                split.label("Color:")
                row = split.row(align=True)
                row.prop(props, "color", text="")
                row.prop_search(props, "color_map", mat, "texture_slots", text="", icon='TEXTURE')
                subrow = row.row(align=True)
                subrow.alignment = 'RIGHT'
                subrow.prop(props, "color_map_enabled", text="T", toggle=True)
                row = box.row()
                row.prop(props, "hue_shift")
                row.prop(props, "invert_hue")
                row = box.row()
                row.prop(props, "vibrance")
                row.prop(props, "density")
                box.prop(props, "opacity")
                split = box.split(0.25)
                split.label("Roughness:")
                row = split.row(align=True)
                row.prop(props, "roughness", text="")
                row.prop_search(props, "roughness_map", mat, "texture_slots", text="", icon='TEXTURE')
                subrow = row.row(align=True)
                subrow.alignment = 'RIGHT'
                subrow.prop(props, "roughness_map_enabled", text="T", toggle=True)
                box.prop(props, "specular_tint")
                row = box.row()
                row.prop(props, "clearcoat")
                row.prop(props, "clearcoat_ior")
        elif mx_type == 'carpaint':
            props = mx.carpaint
            ui_collapsed = props.ui_collapsed
            col = layout.column(align=True)
            box = col.box()
            row = box.row()
            row.alignment = 'LEFT'
            icon = 'TRIA_RIGHT' if ui_collapsed else 'TRIA_DOWN'
            op = row.operator("maxwell.expand_ui", "Car Paint", icon=icon, emboss=False)
            op.path = props.path_from_id()
            op.attr = "ui_collapsed"
            if not ui_collapsed:
                box = col.box()
                row = box.row(align=True)
                row.menu("MAXWELL_MT_material_carpaint_presets")
                row.operator("maxwell.material_carpaint_preset_add", text="", icon='ZOOMIN')
                row.operator("maxwell.material_carpaint_preset_add", text="", icon='ZOOMOUT').remove_active = True
                row = box.row()
                row.column().prop(props, "color")
                col = row.column(align=True)
                col.prop(props, "metallic")
                col.prop(props, "topcoat")
        elif mx_type == 'hair':
            props = mx.hair
            ui_collapsed = props.ui_collapsed
            col = layout.column(align=True)
            box = col.box()
            row = box.row()
            row.alignment = 'LEFT'
            icon = 'TRIA_RIGHT' if ui_collapsed else 'TRIA_DOWN'
            op = row.operator("maxwell.expand_ui", "Hair", icon=icon, emboss=False)
            op.path = props.path_from_id()
            op.attr = "ui_collapsed"
            if not ui_collapsed:
                box = col.box()
                row = box.row(align=True)
                row.menu("MAXWELL_MT_material_hair_presets")
                row.operator("maxwell.material_hair_preset_add", text="", icon='ZOOMIN')
                row.operator("maxwell.material_hair_preset_add", text="", icon='ZOOMOUT').remove_active = True
                split = box.split(0.33)
                split.label("Color:")
                row = split.row(align=True)
                row.prop(props, "color", text="")
                row.prop_search(props, "color_map", mat, "texture_slots", text="", icon='TEXTURE')
                subrow = row.row(align=True)
                subrow.alignment = 'RIGHT'
                subrow.prop(props, "color_map_enabled", text="T", toggle=True)
                split = box.split(0.33)
                split.label("Root-Tip Map:")
                split.prop_search(props, "rt_map", mat, "texture_slots", text="", icon='TEXTURE')
                split = box.split(0.33)
                split.label("Root-Tip Weight:")
                row = split.row(align=True)
                row.prop(props, "rt_weight", text="")
                row.prop_search(props, "rt_weight_map", mat, "texture_slots", text="", icon='TEXTURE')
                subrow = row.row(align=True)
                subrow.alignment = 'RIGHT'
                subrow.prop(props, "rt_weight_map_enabled", text="T", toggle=True)
                split = box.split(0.5)
                col = split.column()
                col.label("Primary Highlight:")
                subcol = col.column(align=True)
                subcol.prop(props, "primary_strength")
                subcol.prop(props, "primary_spread")
                col.row().prop(props, "primary_tint")
                col = split.column()
                col.label("Secondary Highlight:")
                subcol = col.column(align=True)
                subcol.prop(props, "secondary_strength")
                subcol.prop(props, "secondary_spread")
                col.row().prop(props, "secondary_tint")


class MaxwellLayersPanel(MaterialButtonsPanel, Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_label = "Maxwell Material Layers"

    @classmethod
    def poll(cls, context):
        return super().poll(context) and context.material.maxwell.type == 'custom'

    def draw(self, context):
        layout = self.layout

        mat = context.material
        layout.context_pointer_set("maxwell_ui", mat)

        mx = mat.maxwell

        row = layout.row()
        row.template_list("UI_UL_maxwell_list", "", mx, "layers", mx, "active_layer_index", rows=2)
        col = row.column(align=True)
        col.operator("maxwell.material_add_layer", icon='ZOOMIN', text="")
        col.operator("maxwell.material_remove_layer", icon='ZOOMOUT', text="")
        col.operator_menu_enum("maxwell.material_layer", "action", text="", icon='DOWNARROW_HLT')

        if mx.layers:
            layer = mx.layers[mx.active_layer_index]
            row = layout.row(align=True)
            row.operator("maxwell.material_up_layer", icon='TRIA_UP', text="")
            row.operator("maxwell.material_down_layer", icon='TRIA_DOWN', text="")
            row.prop(layer, "name", text="")

            if layer.enabled:
                row = layout.row()
                row.label("Blending:")
                row.prop(layer, "blend", expand=True)

                row = layout.row(align=True)
                row.prop(layer, "opacity")
                row.prop_search(layer, "opacity_tex", mat, "texture_slots", text="", icon='TEXTURE')

                # Emitter
                emitter = layer.emitter

                col = layout.column(align=True)
                box = col.box()
                row = box.row()
                subrow = row.row()
                subrow.alignment = 'LEFT'
                icon = 'TRIA_DOWN' if emitter.expanded else 'TRIA_RIGHT'
                op = subrow.operator("maxwell.expand_ui", "Emitter", emboss=False, icon=icon)
                op.path = emitter.path_from_id()
                op.attr = "expanded"
                row.prop(emitter, "enabled", text="")
                if emitter.expanded:
                    box = col.box()
                    box.enabled = emitter.enabled
                    box.row().prop(emitter, "type", expand=True)
                    if emitter.type == '0':
                        # Custom
                        split1 = box.split(0.33)
                        split1.label("Color:")
                        split2 = split1.split(0.9, align=True)
                        split3 = split2.split(0.5, align=True)
                        if emitter.color_as_temp:
                            split3.prop(emitter, "color_tmp", text="")
                            split3.prop(emitter, "color_temp", text="K")
                        else:
                            split3.prop(emitter, "color", text="")
                            row = split3.row(align=True)
                            row.enabled = False
                            row.prop(emitter, "color_temp", text="K")
                        split2.prop(emitter, "color_as_temp", text="T", toggle=True)

                        split = box.split(0.33)
                        split.label("Lobe:")
                        split.row().prop(emitter, "lobe", expand=True)

                        if emitter.lobe == '0':
                            box.prop(emitter, "luminance")
                            if emitter.luminance == '0':
                                row = box.row(align=True)
                                row.prop(emitter, "power")
                                row.prop(emitter, "efficacy")
                            else:
                                box.prop(emitter, "output")
                        else:
                            if emitter.lobe == '1':
                                box.prop(emitter, "ies")
                            elif emitter.lobe == '2':
                                box.prop_search(emitter, "image", mat, "texture_slots", icon='TEXTURE')
                            box.prop(emitter, "intensity")
                    elif emitter.type == '1':
                        # Temperature
                        box.prop(emitter, "temperature")
                    elif emitter.type == '2':
                        # HDRI
                        row = box.split(0.75, align=True)
                        row.prop_search(emitter, "hdri_tex", mat, "texture_slots", icon='TEXTURE')
                        row.prop(emitter, "hdri_intensity", text="")


class MaxwellBSDFsPanel(MaterialButtonsPanel, Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_label = "Maxwell Material BSDFs"

    @classmethod
    def poll(cls, context):
        if super().poll(context):
            mx = context.material.maxwell
            if mx.type == 'custom' and mx.active_layer_index >= 0:
                return mx.layers[mx.active_layer_index].enabled
        return False

    def draw_header(self, context):
        mx = context.material.maxwell
        layer = mx.layers[mx.active_layer_index]
        setattr(self, "text", "%s - [%s]" % (self.bl_label, layer.name))

    def draw(self, context):
        layout = self.layout

        mat = context.material
        layout.context_pointer_set("maxwell_ui", mat)

        mx = mat.maxwell
        layer = mx.layers[mx.active_layer_index]

        row = layout.row()
        row.template_list("UI_UL_maxwell_list", "", layer, "bsdfs", layer, "active_bsdf_index", rows=2)
        col = row.column(align=True)
        col.operator("maxwell.material_add_bsdf", icon='ZOOMIN', text="")
        col.operator("maxwell.material_remove_bsdf", icon='ZOOMOUT', text="")
        col.operator_menu_enum("maxwell.material_bsdf", "action", icon='DOWNARROW_HLT', text="")

        if layer.bsdfs:
            bsdf = layer.bsdfs[layer.active_bsdf_index]

            row = layout.row(align=True)
            row.operator("maxwell.material_up_bsdf", icon='TRIA_UP', text="")
            row.operator("maxwell.material_down_bsdf", icon='TRIA_DOWN', text="")
            row.prop(bsdf, "name", text="")

            row = layout.row(align=True)
            row.prop(bsdf, "weight")
            row.prop_search(bsdf, "weight_tex", mat, "texture_slots", text="", icon='TEXTURE')

            if bsdf.enabled:
                # BSDF Properties
                col = layout.column(align=True)
                box = col.box()
                row = box.row()
                row.alignment = 'LEFT'
                icon = 'TRIA_DOWN' if bsdf.bsdf_expanded else 'TRIA_RIGHT'
                op = row.operator("maxwell.expand_ui", "BSDF Properties", icon=icon, emboss=False)
                op.path = bsdf.path_from_id()
                op.attr = "bsdf_expanded"
                if bsdf.bsdf_expanded:
                    container = col.box()
                    split = container.split(0.33)
                    split.label("IOR")
                    row = split.row()
                    row.prop(bsdf, "ior_type", expand=True)
                    if bsdf.ior_type == '0':
                        split = container.split(0.33)
                        split.label("Reflectance 0")
                        split = split.split(0.7)
                        split.prop_search(bsdf, "refl_0_tex", mat, "texture_slots", text="", icon='TEXTURE')
                        row = split.row()
                        row.prop(bsdf, "refl_0", text="")
                        row.enabled = not bsdf.refl_0_tex

                        split = container.split(0.33)
                        split.label("Reflectance 90")
                        split = split.split(0.7)
                        split.prop_search(bsdf, "refl_90_tex", mat, "texture_slots", text="", icon='TEXTURE')
                        row = split.row()
                        row.prop(bsdf, "refl_90", text="")
                        row.enabled = not bsdf.refl_90_tex

                        split = container.split(0.33)
                        split.label("Transmittance")
                        split = split.split(0.7)
                        split.prop_search(bsdf, "trans_tex", mat, "texture_slots", text="", icon='TEXTURE')
                        row = split.row()
                        row.prop(bsdf, "trans", text="")
                        row.enabled = not bsdf.trans_tex

                        #transmittance_enabled = bool(bsdf.trans.r or bsdf.trans.g or bsdf.trans.b or bsdf.trans_tex)
                        roughness_enabled = bsdf.roughness < 100.0 or bool(bsdf.roughness_tex)

                        row = container.split(0.7, align=True)
                        row.prop(bsdf, "attenuation")
                        row.prop(bsdf, "atten_units", text="")
                        #row.enabled = transmittance_enabled and roughness_enabled and not bsdf.k

                        row = container.row()
                        row.prop(bsdf, "nd")
                        subrow = row.row()
                        subrow.prop(bsdf, "force_fresnel")
                        subrow.enabled = roughness_enabled

                        row = container.row()
                        subrow = row.row()
                        subrow.prop(bsdf, "k")
                        subrow.enabled = roughness_enabled
                        subrow = row.row()
                        subrow.prop(bsdf, "abbe")
                        subrow.enabled = mx.dispersion

                        split = container.split(0.1, align=True)
                        split.prop(bsdf, "cf_enabled", text="R2", toggle=True)
                        row = split.row(align=True)
                        row.prop(bsdf, "cf_angle")
                        row.prop(bsdf, "cf_roughness")
                        row.enabled = bsdf.cf_enabled
                    elif bsdf.ior_type == '1':
                        container.row().prop(bsdf, "ior_file")
                # Surface Properties
                col = layout.column(align=True)
                box = col.box()
                row = box.row()
                row.alignment = 'LEFT'
                icon = 'TRIA_DOWN' if bsdf.surf_expanded else 'TRIA_RIGHT'
                op = row.operator("maxwell.expand_ui", "Surface Properties", icon=icon, emboss=False)
                op.path = bsdf.path_from_id()
                op.attr = "surf_expanded"
                if bsdf.surf_expanded:
                    container = col.box()
                    split = container.split(0.5, align=True)
                    split.prop(bsdf, "roughness")
                    split.prop_search(bsdf, "roughness_tex", mat, "texture_slots", text="", icon='TEXTURE')

                    split = container.split(0.5, align=True)
                    split.prop_search(bsdf, "bump_tex", mat, "texture_slots", text="", icon='TEXTURE')
                    split = split.split(0.825, align=True)
                    split.prop(bsdf, "bump", text="Normal" if bsdf.bump_normal else "Bump")
                    split.prop(bsdf, "bump_normal", text="N", toggle=True)
                    split.enabled = bool(bsdf.bump_tex)

                    split = container.split(align=True)
                    row = split.row(align=True)
                    row.prop(bsdf, "anisotropy")
                    row.enabled = not bsdf.anisotropy_tex
                    split.prop_search(bsdf, "anisotropy_tex", mat, "texture_slots", text="", icon='TEXTURE')

                    split = container.split(0.5, align=True)
                    row = split.row(align=True)
                    row.prop(bsdf, "angle")
                    row.enabled = not bsdf.angle_tex
                    split.prop_search(bsdf, "angle_tex", mat, "texture_slots", text="", icon='TEXTURE')
                    split.enabled = bool(bsdf.anisotropy or bsdf.anisotropy_tex)
                # Subsurface Properties
                col = layout.column(align=True)
                box = col.box()
                row = box.row()
                row.alignment = 'LEFT'
                icon = 'TRIA_DOWN' if bsdf.subs_expanded else 'TRIA_RIGHT'
                op = row.operator("maxwell.expand_ui", "Subsurface Properties", icon=icon, emboss=False)
                op.path = bsdf.path_from_id()
                op.attr = "subs_expanded"
                if bsdf.subs_expanded:
                    container = col.box()

                    row = container.row(align=True)
                    row.prop(bsdf, "scattering")
                    subrow = row.row(align=True)
                    subrow.prop_search(bsdf, "scattering_tex", mat, "texture_slots", text="", icon='TEXTURE')
                    subrow.enabled = bsdf.single_sided

                    split = container.split(0.5, align=True)
                    split.prop(bsdf, "coef")
                    split.prop(bsdf, "asymmetry")

                    row = container.row()
                    row.prop(bsdf, "single_sided")

                    split = container.split(align=True)
                    row = split.row(align=True)
                    row.prop(bsdf, "thickness")
                    row.enabled = not bsdf.thickness_tex
                    split.prop_search(bsdf, "thickness_tex", mat, "texture_slots", text="", icon='TEXTURE')
                    split.enabled = bsdf.single_sided

                    row = container.row(align=True)
                    row.prop(bsdf, "thickness_min")
                    row.prop(bsdf, "thickness_max")
                    row.enabled = bsdf.single_sided  # and bool(bsdf.thickness_tex)
            # Coating
            coating = bsdf.coating
            col = layout.column(align=True)
            box = col.box()
            row = box.row()
            subrow = row.row()
            subrow.alignment = 'LEFT'
            icon = 'TRIA_DOWN' if coating.expanded else 'TRIA_RIGHT'
            op = subrow.operator("maxwell.expand_ui", "Coating", icon=icon, emboss=False)
            op.path = coating.path_from_id()
            op.attr = "expanded"
            row.prop(coating, "enabled", text="")
            if coating.expanded:
                container = col.box()
                container.enabled = coating.enabled

                split = container.split(0.5, align=True)
                row = split.row(align=True)
                row.prop(coating, "thickness")
                row.enabled = not coating.thickness_tex
                split.prop_search(coating, "thickness_tex", mat, "texture_slots", text="", icon='TEXTURE')

                row = container.row(align=True)
                row.prop(coating, "thickness_min")
                row.prop(coating, "thickness_max")
                row.enabled = bool(coating.thickness_tex)

                split = container.split(0.33)
                split.label("IOR:")
                split.row().prop(coating, "ior_type", expand=True)
                if coating.ior_type == '0':
                    split = container.split(0.33)
                    split.label("Reflectance 0:")
                    split = split.split(0.7)
                    split.prop_search(coating, "refl_0_tex", mat, "texture_slots", text="", icon='TEXTURE')
                    row = split.row()
                    row.prop(coating, "refl_0", text="")
                    row.enabled = not coating.refl_0_tex

                    split = container.split(0.33)
                    split.label("Reflectance 90:")
                    split = split.split(0.7)
                    split.prop_search(coating, "refl_90_tex", mat, "texture_slots", text="", icon='TEXTURE')
                    row = split.row()
                    row.prop(coating, "refl_90", text="")
                    row.enabled = not coating.refl_90_tex

                    row = container.row()
                    row.prop(coating, "nd")
                    row.prop(coating, "force_fresnel")

                    split = container.split(0.5)
                    split.prop(coating, "k")
                    subsplit = split.split(0.2, align=True)
                    subsplit.prop(coating, "cf_enabled", text="R2", toggle=True)
                    row = subsplit.row(align=True)
                    row.prop(coating, "cf_angle")
                    row.enabled = coating.cf_enabled
                elif coating.ior_type == '1':
                    container.row().prop(coating, "ior_file")
