# -*- coding: utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Ildar Nikolaev"
__email__ = "nildar@users.sourceforge.net"

import bpy
from bpy.types import Panel, UIList

from ..engine import MaxwellRenderEngine
from ..operators.object import ImportSea

#############################
# Object
#
from bl_ui.properties_object import ObjectButtonsPanel


class MaxwellObjectPanel(ObjectButtonsPanel, Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_label = "Maxwell Object Appearance"

    @classmethod
    def poll(cls, context):
        e = context.scene.render.engine
        return (e in cls.COMPAT_ENGINES) and \
            (context.object.type in ('MESH', 'CURVE', 'SURFACE', 'META', 'FONT', 'EMPTY'))

    def draw(self, context):
        layout = self.layout
        obj = context.object
        mx = context.object.maxwell

        layout.prop(mx, "opacity")

        if obj.type != 'EMPTY':
            layout.prop_search(mx, "backface_material", obj, "material_slots", icon='MATERIAL')

        row = layout.row()
        row.label("Color ID:")
        subrow = row.row()
        if mx.pi_as_color:
            subrow.prop(mx, "pi_color", text="")
        else:
            subrow.prop(mx, "color_id", text="")
        subrow.prop(mx, "pi_as_color")

        split = layout.split(0.33)
        col = split.column()
        col.separator()
        col.label("Hidden from:")
        col = split.box().column(align=True)
        col.prop(mx, "hide_to_camera")
        col.prop(mx, "hide_to_rr")
        col.prop(mx, "hide_to_gi")
        col.prop(mx, "ecp", text="Z-clip Planes")

        split = layout.split(0.33)
        col = split.column()
        col.separator()
        col.label("Coustics:")
        box = split.box()
        row = box.row(align=True)
        row.alignment = 'LEFT'
        # col 1
        col = row.column(align=True)
        col.alignment = 'RIGHT'
        col.enabled = False
        col.label("")
        col.label("Direct")
        col.label("Indirect")
        # col 2
        col = row.column(align=True)
        subrow = col.row(align=True)
        subrow.alignment = 'RIGHT'
        subrow.enabled = False
        subrow.label("Reflection")
        subrow = col.row(align=True)
        subrow.alignment = 'CENTER'
        subrow.prop(mx, "cd_refl", text="")
        subrow = col.row(align=True)
        subrow.alignment = 'CENTER'
        subrow.prop(mx, "ci_refl", text="")
        # col 3
        col = row.column(align=True)
        subrow = col.row(align=True)
        subrow.alignment = 'RIGHT'
        subrow.enabled = False
        subrow.label("Refraction")
        subrow = col.row(align=True)
        subrow.alignment = 'CENTER'
        subrow.prop(mx, "cd_refr", text="")
        subrow = col.row(align=True)
        subrow.alignment = 'CENTER'
        subrow.prop(mx, "ci_refr", text="")


class MaxwellSeaPanel(ObjectButtonsPanel, Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_options = {'DEFAULT_CLOSED'}
    bl_label = "Maxwell Sea"

    @classmethod
    def poll(cls, context):
        e = context.scene.render.engine
        return (e in cls.COMPAT_ENGINES) and context.object.type == 'MESH'

    def draw_header(self, context):
        self.layout.prop(context.object.data.maxwell.sea, "enabled", text="")

    def draw(self, context):
        layout = self.layout
        mesh = context.object.data
        sea = mesh.maxwell.sea

        layout.enabled = sea.enabled
        layout.prop_search(sea, "material", mesh, "materials")

        layout.context_pointer_set("maxwell_ui", mesh)
        sea_path = sea.path_from_id()
        # Geometry
        collapsed = sea.ui_geometry_collapsed
        col = layout.column(align=True)
        box = col.box()
        row = box.row()
        row.alignment = 'LEFT'
        icon = 'TRIA_RIGHT' if collapsed else 'TRIA_DOWN'
        op = row.operator("maxwell.expand_ui", "Geometry", icon=icon, emboss=False)
        op.attr = "ui_geometry_collapsed"
        op.path = sea_path
        if not collapsed:
            box = col.box()
            box.prop(sea, "quality")
            box.prop(sea, "time")
            box.prop(sea, "vertical_scale")
            box.prop(sea, "dimension")
            box.prop(sea, "depth")
            box.prop(sea, "seed")
            box.prop(sea, "choppyness")
            box.prop(sea, "choppy_factor")
        # Wind
        collapsed = sea.ui_wind_collapsed
        col = layout.column(align=True)
        box = col.box()
        row = box.row()
        row.alignment = 'LEFT'
        icon = 'TRIA_RIGHT' if collapsed else 'TRIA_DOWN'
        op = row.operator("maxwell.expand_ui", "Wind", icon=icon, emboss=False)
        op.attr = "ui_wind_collapsed"
        op.path = sea_path
        if not collapsed:
            box = col.box()
            box.prop(sea, "wind_speed")
            box.prop(sea, "wind_direction")
            box.prop(sea, "weight_against_wind")
            box.prop(sea, "wind_alignment")
            box.prop(sea, "min_wave_length")
        # Bake
        collapsed = sea.ui_mc_collapsed
        col = layout.column(align=True)
        box = col.box()
        row = box.row()
        row.alignment = 'LEFT'
        icon = 'TRIA_RIGHT' if collapsed else 'TRIA_DOWN'
        op = row.operator("maxwell.expand_ui", "Bake to Mesh Cache", icon=icon, emboss=False)
        op.attr = "ui_mc_collapsed"
        op.path = sea_path
        if not collapsed:
            box = col.box()
            box.prop(sea, "mc_modifier")
            box.prop(sea, "mc_file")
            box.prop(sea, "mc_quality")
            box.row().prop(sea, "mc_frames")

        row = layout.row(align=True)
        row.operator("maxwell.import_sea", text="Import").action = 'SINGLE_FRAME'
        row.operator("maxwell.import_sea", text="Bake").action = 'FRAMES_RANGE'


#############################
# Empty
#
from bl_ui.properties_data_empty import DataButtonsPanel as EmptyButtonsPanel


class MaxwellEmptyButtonsPanel(EmptyButtonsPanel, Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_label = "Maxwell Object Data"

    @classmethod
    def poll(cls, context):
        e = context.scene.render.engine
        return super().poll(context) and (e in cls.COMPAT_ENGINES)

    #def draw_header(self, context):
    #    self.layout.prop(context.object.maxwell.reference, "enabled", text="")

    def draw(self, context):
        layout = self.layout
        ob = context.object
        mx = ob.maxwell

        layout.prop(mx, "type")
        if mx.type == 'REFERENCE':
            r = mx.reference
            layout.prop(r, "path")
            col = layout.column(align=True)
            col.label("Override Flags:")
            row = col.row()
            row.separator()
            col = row.column()
            col.prop(r, "flag_hidden")
            col.label("Hidden from:")
            row = col.row()
            row.separator()
            col = row.column()
            col.prop(r, "flag_camera")
            col.prop(r, "flag_rr")
            col.prop(r, "flag_gi")
        elif mx.type == 'VOLUMETRIC':
            v = mx.volumetric
            layout.prop(v, "field_type", expand=True)
            if ob.parent:
                layout.prop_search(v, "material", ob.parent.data, "materials")
            layout.prop(v, "field_density")
            if v.field_type == 'NOISE':
                layout.prop(v, "seed")
                layout.prop(v, "low_value")
                layout.prop(v, "high_value")
                layout.prop(v, "detail")
                layout.prop(v, "octaves")
                layout.prop(v, "persistence")


#############################
# UV Projections
#
from bl_ui.properties_data_mesh import DATA_PT_uv_texture

_UVSET_ICON = {
    'PLANAR': 'MATPLANE',
    'SPHERICAL': 'MATSPHERE',
    'CYLINDRICAL': 'MESH_CYLINDER',
    'CUBIC': 'MATCUBE',
    'LOCKED': 'TEXTURE'
}

class MaxwellUVSetsUIList(UIList):
    def draw_item(self, context, layout, data, item, icon, active_data, active_propname):
        if self.layout_type in {'DEFAULT'}:
            label = " %s : %s" % (item.channel, item.type_label)
            t = item.type
            if t == 'LOCKED':
                label += " : %s" % item.name
            layout.label(label, icon=_UVSET_ICON[t])


class MaxwellUVSetsPanel(Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_label = "Maxwell UV Sets"
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "data"

    @classmethod
    def poll(cls, context):
        e = context.scene.render.engine
        return (e in cls.COMPAT_ENGINES) and (context.mesh or context.curve or context.meta_ball)

    def draw_header(self, context):
        self.layout.prop(context.object.data.maxwell, "use_projections", text="")

    def draw(self, context):
        layout = self.layout
        id_data = context.object.data

        mx = id_data.maxwell
        if not mx.use_projections:
            return

        row = layout.row()
        row.template_list("MaxwellUVSetsUIList", "", mx, "projections", mx, "active_projection", rows=2)
        col = row.column(align=True)
        col.operator("maxwell.uv_projection_add", icon='ZOOMIN', text="")
        col.operator("maxwell.uv_projection_remove", icon='ZOOMOUT', text="")

        projections = mx.projections
        active = mx.active_projection
        if 0 <= active < len(projections):
            p = projections[active]
            p_type = p.type
            layout.prop(p, "channel")
            layout.prop(p, "type")
            if p_type == 'LOCKED':
                layout.prop_search(p, "name", context.mesh, "uv_layers", icon='GROUP_UVS')
            else:
                layout.prop_search(p, "object", context.scene, "objects")
                row = layout.row()
                row.enabled = not bool(p.object)
                row.column().prop(p, "position")
                row.column().prop(p, "rotation")
                row.column().prop(p, "scale")
                if p_type == 'SPHERICAL':
                    row = layout.row()
                    row.column().prop(p, "latitude")
                    row.column().prop(p, "longitude")
                elif p_type == 'CYLINDRICAL':
                    row = layout.row()
                    row.prop(p, "angle")

    def draw_panel(self, context):
        if context.scene.render.engine != MaxwellRenderEngine.bl_idname:
            return

        mesh = context.mesh
        uvl = mesh.uv_layers.active
        if not uvl:
            return

        layout = self.layout
        layout.context_pointer_set("maxwell_ui", mesh)

        huvl = mesh.headus_uvlayout
        collapsed = huvl.ui_collapsed

        col = layout.column(align=True)
        box = col.box()
        row = box.row()
        row.alignment = 'LEFT'
        icon = 'TRIA_RIGHT' if collapsed else 'TRIA_DOWN'
        op = row.operator("maxwell.expand_ui", "Headus UVLayout", icon=icon, emboss=False)
        op.path = huvl.path_from_id()
        op.attr = "ui_collapsed"
        if not collapsed:
            box = col.box()
            box.prop(huvl, "uv_file")
            row = box.row()
            subrow = row.row(align=True)
            subrow.operator("headus_uvlayout.new", text="New")
            subrow.operator("headus_uvlayout.edit", text="Edit")
            subrow.operator("headus_uvlayout.update", text="Update")
            row.operator("headus_uvlayout.export", text="Export")

    @classmethod
    def register(cls):
        DATA_PT_uv_texture.append(cls.draw_panel)

    @classmethod
    def unregister(cls):
        DATA_PT_uv_texture.remove(cls.draw_panel)
