# -*- coding: utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Ildar Nikolaev"
__email__ = "nildar@users.sourceforge.net"

from bpy.types import Panel, Menu
from bl_ui.properties_particle import ParticleButtonsPanel

from ..engine import MaxwellRenderEngine


class MaxwellParticlesGrassPresetsMenu(Menu):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_idname = "MAXWELL_MT_particles_grass_presets"
    bl_label = "Preset"

    preset_operator = "script.execute_preset"
    preset_subdir = "bmaxwell/particles/grass"

    draw = Menu.draw_preset


class MaxwellParticlesPanel(ParticleButtonsPanel, Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_label = "Maxwell Extension"

    @classmethod
    def poll(cls, context):
        if super().poll(context):
            pss = context.particle_system.settings
            if pss.type == 'HAIR':
                return pss.render_type == 'PATH'
            elif pss.type == 'EMITTER':
                return pss.render_type in ('HALO', 'LINE', 'PATH')
        return False

    def draw(self, context):
        layout = self.layout
        pss = context.particle_system.settings
        mx = pss.maxwell
        if pss.type == 'HAIR':
            if pss.render_type == 'PATH':
                layout.prop(mx, "h_type", expand=True)
                if mx.h_type == 'HAIR':
                    h = mx.hair
                    row = layout.row(align=True)
                    row.prop(h, "root")
                    row.prop(h, "tip")
                    row = layout.row()
                    row.prop_search(h, "uv_map", context.object.data, "uv_textures", icon='GROUP_UVS')
                elif mx.h_type == 'GRASS':
                    g = mx.grass
                    row = layout.row(align=True)
                    row.menu("MAXWELL_MT_particles_grass_presets")
                    row.operator("maxwell.particles_grass_preset_add", text="", icon='ZOOMIN')
                    row.operator("maxwell.particles_grass_preset_add", text="", icon='ZOOMOUT').remove_active = True
                    # Primitive
                    row = layout.row()
                    row.prop(g, "type")
                    row.prop(g, "blade_points")
                    # Grass Density
                    layout.row().prop_search(g, "density_map", pss, "texture_slots", icon='TEXTURE')
                    row = layout.row(align=True)
                    row.prop(g, "density")
                    row.prop(g, "seed")
                    # Blade Length
                    layout.row().prop_search(g, "length_map", pss, "texture_slots", icon='TEXTURE')
                    row = layout.row(align=True)
                    row.prop(g, "length")
                    row.prop(g, "length_var", text="Variation")
                    # Width
                    row = layout.row()
                    row.label("Width:")
                    subrow = row.row(align=True)
                    subrow.prop(g, "root_width", text="Root")
                    subrow.prop(g, "tip_width", text="Tip")
                    # Angle
                    row = layout.row()
                    row.label("Angle direction:")
                    row.prop(g, "direction", expand=True)
                    layout.row().prop_search(g, "angle_map", pss, "texture_slots", icon='TEXTURE')
                    row = layout.row(align=True)
                    row.prop(g, "angle")
                    row.prop(g, "angle_var", text="Variation")
                    # Bend
                    layout.row().prop_search(g, "start_bend_map", pss, "texture_slots", icon='TEXTURE')
                    row = layout.row(align=True)
                    row.prop(g, "start_bend")
                    row.prop(g, "start_bend_var", text="Variation")
                    layout.row().prop_search(g, "bend_radius_map", pss, "texture_slots", icon='TEXTURE')
                    row = layout.row(align=True)
                    row.prop(g, "bend_radius")
                    row.prop(g, "bend_radius_var", text="Variation")
                    layout.row().prop_search(g, "bend_angle_map", pss, "texture_slots", icon='TEXTURE')
                    row = layout.row(align=True)
                    row.prop(g, "bend_angle")
                    row.prop(g, "bend_angle_var", text="Variation")
                    # Cut off
                    layout.row().prop_search(g, "cut_off_map", pss, "texture_slots", icon='TEXTURE')
                    row = layout.row(align=True)
                    row.prop(g, "cut_off")
                    row.prop(g, "cut_off_var", text="Variation")
                    # Level of Detail
                    row = layout.row()
                    row.prop(g, "lod_enabled")
                    col = row.column(align=True)
                    col.enabled = g.lod_enabled
                    row = col.row(align=True)
                    row.prop(g, "lod_min")
                    row.prop(g, "lod_max")
                    row = col.row(align=True)
                    row.prop(g, "lod_max_density")
                    # Display
                    row = layout.row(align=True)
                    row.prop(g, "display_percent", text="Display")
                    row.prop(g, "display_blades", text="Max. Blades")
                    layout.separator()
                    layout.operator("maxwell.import_grass")
        elif pss.type == 'EMITTER':
            if pss.render_type in ['HALO', 'LINE', 'PATH']:
                layout.prop(mx, "p_type", expand=True)
                if mx.p_type == 'MESHER':
                    m = mx.mesher
                    layout.prop(m, "mesher")

                    # System parameters
                    layout.prop(m, "threads")

                    # Sequence
                    col = layout.column(align=True)
                    row = col.row(align=True)
                    row.prop(m, "radius")
                    row.prop(m, "core")
                    row = col.row(align=True)
                    row.prop(m, "splash")
                    row.prop(m, "velocity")

                    # Mesh Optimization
                    split = layout.split(0.5, align=True)
                    split.prop(m, "optimize", text="Optimization")
                    row = split.row(align=True)
                    row.enabled = m.optimize != '0'
                    row.prop(m, "distance")

                    # Mesh parameters
                    box = layout.box()
                    box.label("Mesh parameters:")
                    flow = box.column_flow(2)
                    col = flow.column(align=True)
                    col.prop(m, "poly")
                    col.prop(m, "smooth")
                    col.prop(m, "iso_level")
                    col = flow.column()
                    col.prop(m, "wn")
                    col.prop(m, "fix_bbox")

                    # Mesh filter
                    box = layout.box()
                    box.prop(m, "filter")
                    enabled = m.filter
                    col = box.column(align=True)
                    col.enabled = enabled
                    row = col.row(align=True)
                    row.prop(m, "relax")
                    row.prop(m, "tension")
                    row = col.row(align=True)
                    row.prop(m, "thinning")
                    row.prop(m, "steps")
                    row = box.row()
                    row.enabled = enabled
                    row.prop(m, "splash_thinning")
                    row = box.row(align=True)
                    row.enabled = enabled
                    row.prop(m, "splash_threshold")
                    row.prop(m, "thinning_size")
                    row = box.row()
                    row.enabled = enabled
                    row.prop(m, "core_smooth")
                    row = box.row(align=True)
                    row.enabled = enabled
                    row.prop(m, "core_threshold")
                    row.prop(m, "smooth_steps")

                    # Export and lazy compute
                    box = layout.box()
                    box.label("Export and lazy compute:")
                    box.prop(m, "action")
                    box.prop(m, "md_file")
                    row = box.row(align=True)
                    row.prop(m, "md_offset")
                    row.prop(m, "md_compression")
                    box.prop(m, "md_normals")
                elif mx.p_type == 'PARTICLES':
                    p = mx.particles
                    layout.prop(p, "radius")
                    layout.prop(p, "motion_blur")
                    layout.prop(p, "shutter_speed")

                    col = layout.column(align=True)
                    col.prop(p, "extra_particles")
                    row = col.row(align=True)
                    row.prop(p, "extra_dispersion")
                    row.prop(p, "extra_deformation")
                elif mx.p_type == 'VOLUMETRIC':
                    v = mx.volumetric
                    col = layout.column(align=True)
                    col.prop(v, "load_particles")
                    col.prop(v, "radius_factor")
                    col.prop(v, "cell_size")
                    col = layout.column(align=True)
                    col.prop(v, "density_scale")
                    row = col.row(align=True)
                    row.prop(v, "min_density", text="Min")
                    row.prop(v, "max_density", text="Max")
                    col = layout.column(align=True)
                    col.prop(v, "extra_particles")
                    row = col.row(align=True)
                    row.prop(v, "extra_dispersion")
                    row.prop(v, "extra_deformation")
                    col = layout.column(align=True)
                    col.prop(v, "mb_factor")
                    col.prop(v, "shutter_speed")


class MaxwellScatterPanel(ParticleButtonsPanel, Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_label = "Maxwell Scatter"

    @classmethod
    def poll(cls, context):
        if super().poll(context):
            pss = context.particle_system.settings
            return pss.render_type == 'OBJECT'
        return False

    def draw_header(self, context):
        mx = context.particle_system.settings.maxwell.scatter
        self.layout.prop(mx, "enabled", text="")

    def draw(self, context):
        pss = context.particle_system.settings
        s = pss.maxwell.scatter

        layout = self.layout
        layout.enabled = s.enabled

        row = layout.row()
        row.prop(s, "inherit_oid")
        row.prop(s, "seed")

        row = layout.row(align=True)
        row.prop(s, "density")
        row.prop_search(s, "density_map", pss, "texture_slots", text="", icon='TEXTURE')

        row = layout.row()
        col = row.column(align=True)
        col.prop(s, "scale")
        col = row.column(align=True)
        col.prop(s, "scale_var")
        layout.prop_search(s, "scale_map", pss, "texture_slots", icon='TEXTURE')

        row = layout.row()
        col = row.column(align=True)
        col.prop(s, "rotation")
        col = row.column(align=True)
        col.prop(s, "rotation_var")
        layout.prop_search(s, "rotation_map", pss, "texture_slots", icon='TEXTURE')

        split = layout.split(0.3333)
        split.label("Direction")
        split.row().prop(s, "direction", expand=True)
