# -*- coding: utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Ildar Nikolaev"
__email__ = "nildar@users.sourceforge.net"

from bpy.types import Panel, Menu
from bl_ui.properties_render import RenderButtonsPanel
from bl_ui.properties_render_layer import RenderLayerButtonsPanel

from ..engine import MaxwellRenderEngine


class MaxwellRenderPresetsMenu(Menu):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_idname = "MAXWELL_MT_render_presets"
    bl_label = "Maxwell Render Presets"
    preset_operator = "script.execute_preset"
    preset_subdir = "bmaxwell/render"

    draw = Menu.draw_preset


class MaxwellRenderPanel(RenderButtonsPanel, Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_label = "Maxwell Render"

    def draw(self, context):
        layout = self.layout
        mx = context.scene.maxwell
        rp = mx.rp

        layout.context_pointer_set("maxwell_ui", context.scene)
        rp_path = rp.path_from_id()

        row = layout.row(align=True)
        row.menu("MAXWELL_MT_render_presets", text=MaxwellRenderPresetsMenu.bl_label)
        row.operator("maxwell.render_preset_add", text="", icon='ZOOMIN')
        row.operator("maxwell.render_preset_add", text="", icon='ZOOMOUT').remove_active = True

        layout.prop(mx, "run")

        if mx.run == 'render':
            box = layout.box()
            col = box.column()
            col.label("Command Line:")
            col.prop(mx, "cmdline", text="")
            row = col.row()
            row.prop(mx, "cmdline_nogui_nowait")
            row.prop(mx, "cmdline_trytoresume")
        elif mx.run == 'network':
            col = layout.column()
            col.prop(mx, "run_node")
            col.prop(mx, "run_manager")
            col.prop(mx, "manager", text="Manager")

        layout.prop(mx, "protected")

        split = layout.split(0.65)
        split.prop(rp, "quality")
        split.prop(rp, "threads")

        row = layout.row(align=True)
        row.prop(rp, "time")
        row.prop(rp, "sl")

        if rp.multilight != '0':
            split = layout.split(0.65)
            split.prop(rp, "multilight")
            split.prop(rp, "separate", text="")
        else:
            layout.prop(rp, "multilight")

        collapsed = rp.globals_collapsed
        col = layout.column(align=True)
        box = col.box()
        row = box.row()
        row.alignment = 'LEFT'
        icon = 'TRIA_RIGHT' if collapsed else 'TRIA_DOWN'
        op = row.operator("maxwell.expand_ui", "Globals", icon=icon, emboss=False)
        op.attr = "globals_collapsed"
        op.path = rp_path
        if not collapsed:
            box = col.box()
            row = box.row()
            row.prop(rp, "motionblur")
            subrow = row.row()
            subrow.prop(rp, "mbsteps")
            subrow.enabled = rp.motionblur

            box.prop(rp, "displacement")
            box.prop(rp, "dispersion")

        collapsed = rp.mats_collapsed
        col = layout.column(align=True)
        box = col.box()
        row = box.row()
        row.alignment = 'LEFT'
        icon = 'TRIA_RIGHT' if collapsed else 'TRIA_DOWN'
        op = row.operator("maxwell.expand_ui", "Materials", icon=icon, emboss=False)
        op.attr = "mats_collapsed"
        op.path = rp_path
        if not collapsed:
            box = col.box()
            box.prop(rp, "ovrmat", text="Override")
            box.prop(rp, "defmat", text="Default")

            box.label(text="Searching paths:")
            row = box.row()
            row.template_list("UI_UL_list", "maxwell.searching_paths", rp, "searching_paths", rp, "active_sp_index", rows=2)
            col = row.column(align=True)
            col.operator("maxwell.add_searching_path", icon='ZOOMIN', text="")
            col.operator("maxwell.remove_searching_path", icon='ZOOMOUT', text="")
            if rp.searching_paths:
                spath = rp.searching_paths[rp.active_sp_index]
                box.prop(spath, "name")

        collapsed = rp.ic_collapsed
        col = layout.column(align=True)
        box = col.box()
        row = box.row()
        row.alignment = 'LEFT'
        icon = 'TRIA_RIGHT' if collapsed else 'TRIA_DOWN'
        op = row.operator("maxwell.expand_ui", "Illumination & Caustics", icon=icon, emboss=False)
        op.attr = "ic_collapsed"
        op.path = rp_path
        if not collapsed:
            box = col.box()
            box.prop(rp, "illumination")
            box.prop(rp, "reflcaustics")
            box.prop(rp, "refrcaustics")


class MaxwellToneMappingPanel(RenderButtonsPanel, Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_options = {'DEFAULT_CLOSED'}
    bl_label = "Maxwell Tone Mapping"

    def draw(self, context):
        layout = self.layout
        rp = context.scene.maxwell.rp

        # layout.label(text = " * Tone Mapping *", icon = 'COLOR')
        layout.row().prop(rp, "colorspace")
        row = layout.row(align=True)
        row.prop(rp, "gamma")
        row.prop(rp, "burn")

        row = layout.row()
        row.prop(rp, "doSharpness")
        subrow = row.row()
        subrow.prop(rp, "sharpness")
        subrow.enabled = rp.doSharpness


class MaxwellSimulensPanel(RenderButtonsPanel, Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_options = {'DEFAULT_CLOSED'}
    bl_label = "Maxwell Simulens"

    def draw(self, context):
        layout = self.layout
        rp = context.scene.maxwell.rp

        # layout.label(text = " * Simulens *", icon = 'CAMERA_DATA')
        layout.row().prop(rp, "doDiffraction")
        enabled = rp.doDiffraction

        row = layout.row()
        row.prop(rp, "aperture")
        row.enabled = enabled

        row = layout.row()
        row.prop(rp, "obstacle")
        row.enabled = enabled

        row = layout.row(align=True)
        row.prop(rp, "diffraction")
        row.prop(rp, "frequency")
        row.enabled = enabled

        row = layout.row()
        row.prop(rp, "doScattering")
        subrow = row.row()
        subrow.prop(rp, "scattering")
        subrow.enabled = rp.doScattering

        row = layout.row()
        row.prop(rp, "doDevignetting")
        subrow = row.row()
        subrow.prop(rp, "devignetting")
        subrow.enabled = rp.doDevignetting


class MaxwellOverlayTextPanel(RenderButtonsPanel, Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_options = {'DEFAULT_CLOSED'}
    bl_label = "Maxwell Overlay Text"

    def draw_header(self, context):
        scene = context.scene
        if scene:
            self.layout.prop(scene.maxwell.rp.overlay_text, "enabled", text="")

    def draw(self, context):
        layout = self.layout
        mx = context.scene.maxwell.rp.overlay_text

        layout.enabled = mx.enabled
        layout.prop(mx, "text")
        layout.prop(mx, "position")
        row = layout.row()
        
        col = row.column(align=True)
        col.prop(mx, "color")

        col = row.column(align=True)
        col.label("Background:")
        row = col.row(align=True)
        subrow = row.row(align=True)
        subrow.prop(mx, "background", text="")
        subrow.enabled = mx.background_enabled
        row.prop(mx, "background_enabled", text="")


class MaxwellFirePanel(RenderButtonsPanel, Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_options = {'DEFAULT_CLOSED'}
    bl_label = "Maxwell Fire"

    def draw(self, context):
        layout = self.layout

        mx = context.scene.maxwell
        rp = mx.rp

        row = layout.row(align=True)
        row.prop(rp, "fire_sl")
        row.prop(rp, "fire_time")
        
        layout.prop(rp, "fire_floating_shadows")
        layout.prop(rp, "fire_floating_reflections")
        
        # TODO: fire settings in blender ui
        #row = layout.row(align=True)
        #row.prop(rp, "fire_quality")
        #row.prop(rp, "fire_p")


class MaxwellChannelsPanel(RenderLayerButtonsPanel, Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_options = {'DEFAULT_CLOSED'}
    bl_label = "Maxwell Channels"

    def draw(self, context):
        layout = self.layout
        mx = context.scene.maxwell
        params = mx.rp
        channels = params.channels

        row = layout.row()
        split = row.split(0.3)
        split.label("OutputMode:")
        split.prop(channels, "output", text="")

        row = layout.row()
        split = row.split(0.3)
        split.prop(channels, "render_enabled")
        subrow = split.row()
        subrow.prop(channels, "render_type", text="")
        subrow.enabled = channels.render_enabled

        row = layout.row()
        split = row.split(0.3)
        #split.prop(channels.alpha, "enabled", text="Alpha")
        split.prop(channels, "alpha_enabled")
        subsplit = split.split(0.35)
        subsplit.prop(channels.alpha, "format", text="")
        subsplit.prop(channels, "alpha_opaque")
        subsplit.enabled = channels.alpha.enabled

        row = layout.row()
        split = row.split(0.3)
        #split.prop(channels.z_buffer, "enabled", text="Z-Buffer")
        split.prop(channels, "z_buffer_enabled")
        subsplit = split.split(0.35)
        subsplit.prop(channels.z_buffer, "format", text="")
        subsplit.row().prop(channels, "z_range", text="")
        subsplit.enabled = channels.z_buffer.enabled

        row = layout.row()
        split = row.split(0.3)
        #split.prop(channels.shadow, "enabled", text="Shadow")
        split.prop(channels, "shadow_enabled")
        subsplit = split.split(0.35)
        subsplit.prop(channels.shadow, "format", text="")
        subsplit.enabled = channels.shadow.enabled

        row = layout.row()
        split = row.split(0.3)
        #split.prop(channels.material_id, "enabled", text="Material ID")
        split.prop(channels, "material_id_enabled")
        subsplit = split.split(0.35)
        subsplit.prop(channels.material_id, "format", text="")
        subsplit.enabled = channels.material_id.enabled

        row = layout.row()
        split = row.split(0.3)
        #split.prop(channels.object_id, "enabled", text="Object ID")
        split.prop(channels, "object_id_enabled")
        subsplit = split.split(0.35)
        subsplit.prop(channels.object_id, "format", text="")
        subsplit.enabled = channels.object_id.enabled

        row = layout.row()
        split = row.split(0.3)
        #split.prop(channels.motion, "enabled", text="Motion Vector")
        split.prop(channels, "motion_enabled")
        subsplit = split.split(0.35)
        subsplit.prop(channels.motion, "format", text="")
        subsplit.enabled = channels.motion.enabled

        row = layout.row()
        split = row.split(0.3)
        split.prop(channels.roughness, "enabled", text="Roughness")
        subsplit = split.split(0.35)
        subsplit.prop(channels.roughness, "format", text="")
        subsplit.enabled = channels.roughness.enabled

        row = layout.row()
        split = row.split(0.3)
        split.prop(channels.fresnel, "enabled", text="Fresnel")
        subsplit = split.split(0.35)
        subsplit.prop(channels.fresnel, "format", text="")
        subsplit.enabled = channels.fresnel.enabled

        row = layout.row()
        split = row.split(0.3)
        #split.prop(channels.normals, "enabled", text="Normals")
        split.prop(channels, "normals_enabled")
        subsplit = split.split(0.35)
        subsplit.prop(channels.normals, "format", text="")
        subsplit.row().prop(channels, "normals_space", expand=True)
        subsplit.enabled = channels.normals.enabled

        row = layout.row()
        split = row.split(0.3)
        split.prop(channels.position, "enabled", text="Position")
        subsplit = split.split(0.35)
        subsplit.prop(channels.position, "format", text="")
        subsplit.row().prop(channels, "position_space", expand=True)
        subsplit.enabled = channels.position.enabled

        row = layout.row()
        split = row.split(0.3)
        #split.prop(channels.uv, "enabled", text="UV")
        split.prop(channels, "uv_enabled")
        subsplit = split.split(0.35)
        subsplit.prop(channels.uv, "format", text="")
        subsplit.enabled = channels.uv.enabled
