# -*- coding: utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Ildar Nikolaev"
__email__ = "nildar@users.sourceforge.net"

from bpy.types import Panel
from bl_ui.properties_texture import TextureButtonsPanel

from ..engine import MaxwellRenderEngine


class MaxwellOverrideMapPanel(TextureButtonsPanel, Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_label = "Maxwell Override Map"
    bl_options = {'DEFAULT_CLOSED'}

    @classmethod
    def poll(cls, context):
        return (
            super().poll(context) and
            context.material and
            hasattr(context.texture, "maxwell")
        )

    def draw(self, context):
        layout = self.layout
        layout.enabled = context.texture.maxwell.override

        mx = context.material.maxwell.override_map

        layout.prop(mx, "channel")

        row = layout.row()

        col = row.column(align=True)
        subrow = col.row(align=True)
        subrow.prop(mx, "tile_x", toggle=True)
        subrow.prop(mx, "tile_y", toggle=True)
        col.prop(mx, "units", text="")

        col = row.column(align=True)
        col.label("Mirror:")
        subrow = col.row(align=True)
        subrow.prop(mx, "mirror_x", text="X", toggle=True)
        subrow.prop(mx, "mirror_y", text="Y", toggle=True)

        row = layout.row()
        row.column().prop(mx, "repeat")
        row.column().prop(mx, "offset")

        layout.prop(mx, "rotation")


class MaxwellProjectionPanel(TextureButtonsPanel, Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_label = "Maxwell Projection"

    @classmethod
    def poll(cls, context):
        return (
            super().poll(context) and
            #context.material and  # also used in particle systems
            hasattr(context.texture, "maxwell")
        )

    def draw(self, context):
        layout = self.layout
        mx = context.texture.maxwell

        enabled = not mx.override or context.particle_system is not None
        row = layout.row()
        subrow = row.row()
        subrow.enabled = bool(context.material)
        subrow.prop(mx, "override")
        subrow = row.row()
        subrow.enabled = enabled
        subrow.prop(mx, "channel")

        row = layout.row()
        row.enabled = enabled

        col = row.column(align=True)
        subrow = col.row(align=True)
        subrow.prop(mx, "tile_x", toggle=True)
        subrow.prop(mx, "tile_y", toggle=True)
        col.prop(mx, "units", text="")

        col = row.column(align=True)
        col.label("Mirror:")
        subrow = col.row(align=True)
        subrow.prop(mx, "mirror_x", text="X", toggle=True)
        subrow.prop(mx, "mirror_y", text="Y", toggle=True)

        row = layout.row()
        row.enabled = enabled
        row.column().prop(mx, "repeat")
        row.column().prop(mx, "offset")

        row = layout.row()
        row.enabled = enabled
        row.prop(mx, "rotation")


class MaxwellImagePanel(TextureButtonsPanel, Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_label = "Maxwell Image"

    @classmethod
    def poll(cls, context):
        return (
            super().poll(context) and
            context.material and
            hasattr(context.texture, "maxwell")
        )

    def draw(self, context):
        layout = self.layout
        mx = context.texture.maxwell

        flow = layout.column_flow(2)

        col = flow.column(align=True)
        col.prop(mx, "invert")
        col.prop(mx, "alpha_only")
        col.prop(mx, "interpolation")
        col.separator()
        col.label("Normal Mapping:")
        col.prop(mx, "flip_x", toggle=True)
        col.prop(mx, "flip_y", toggle=True)
        col.prop(mx, "wide", toggle=True)

        col = flow.column(align=True)
        col.prop(mx, "brightness", slider=True)
        col.prop(mx, "contrast", slider=True)
        col.prop(mx, "saturation", slider=True)
        col.prop(mx, "hue", slider=True)
        col.separator()
        col.prop(mx, "clamp")


class MaxwellBakePanel(TextureButtonsPanel, Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_label = "Maxwell Bake Textures"

    @classmethod
    def poll(cls, context):
        return (
            super().poll(context) and
            context.material and
            hasattr(context.texture, "maxwell")
        )

    def draw(self, context):
        layout = self.layout
        mx = context.texture.maxwell
        id_data = context.texture_slot.id_data

        bt = None

        row = layout.row()
        col1 = row.column()
        col1.template_list("UI_UL_maxwell_list", "", mx, "bake_textures", mx, "active_bt_index", rows=2)
        enabled = mx.active_bt_index != -1
        if enabled:
            bt = mx.bake_textures[mx.active_bt_index]
            col1.prop_search(bt, "name", id_data, "texture_slots", "", icon='TEXTURE')
            #col1.prop_search(bt, "name", context.blend_data, "textures", "", icon='TEXTURE')
        col2 = row.column(align=True)
        col2.operator("maxwell.bake_add_texture", icon='ZOOMIN', text="")
        col2.operator("maxwell.bake_remove_texture", icon='ZOOMOUT', text="")

        if bt and bt.name:
            tex = id_data.texture_slots.get(bt.name)
            if tex:
                box = col1.box()
                box.enabled = bool(bt) and bt.enabled
                split = box.split(0.5)
                col = split.column()
                col.separator()
                row = col.row()
                ss = row.split(0.3)
                ss.label("Blend:")
                ss.prop(tex, "blend_type", text="")
                col.separator()
                row = col.row()
                ss = row.split(0.3)
                ss.label("Color:")
                ss.prop(tex, "color", text="")
                col = split.column()
                col.prop(tex, "use_rgb_to_intensity")
                col.prop(tex, "invert")
                col.prop(tex, "use_stencil")

        split = layout.split(0.5)
        col = split.column()
        col.prop(mx, "bake_color")
        col = split.column()
        col.prop(mx, "bake_clear")
        col.prop(mx, "bake_save")
        split.enabled = enabled

        row = layout.row()
        row.enabled = enabled
        row.operator("maxwell.bake_textures", text="Bake", icon='RENDER_STILL')
