# -*- coding: utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Ildar Nikolaev"
__email__ = "nildar@users.sourceforge.net"

from bpy.types import Panel, Menu
from bl_ui.properties_world import WorldButtonsPanel
from bl_ui.properties_world import WORLD_PT_context_world

from ..engine import MaxwellRenderEngine


def _draw_ibl_panel(layout, layer):
    layout.row().prop(layer, "type")
    enabled = (layer.type == '1')
    row = layout.row()
    row.enabled = enabled
    row.prop(layer, "map")
    row = layout.row()
    row.enabled = enabled
    row.prop(layer, "intensity")
    row = layout.row()
    row.enabled = enabled
    row.prop(layer, "scale")
    row = layout.row()
    row.enabled = enabled
    row.prop(layer, "offset")


class MaxwellEnvironmentPresets(Menu):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_idname = "MAXWELL_MT_environment_presets"
    bl_label = "Maxwell Environment Presets"
    preset_operator = "script.execute_preset"
    preset_subdir = "bmaxwell/environment"

    draw = Menu.draw_preset

    def draw_panel(self, context):
        if context.scene.render.engine == MaxwellRenderEngine.bl_idname:
            layout = self.layout
            if context.world:
                layout.separator()
                row = layout.row(align=True)
                row.menu("MAXWELL_MT_environment_presets")
                row.operator("maxwell.environment_preset_add", text="", icon='ZOOMIN')
                row.operator("maxwell.environment_preset_add", text="", icon='ZOOMOUT').remove_active = True
                layout.prop(context.world.maxwell, "type")

    @classmethod
    def register(cls):
        WORLD_PT_context_world.append(cls.draw_panel)

    @classmethod
    def unregister(cls):
        WORLD_PT_context_world.remove(cls.draw_panel)


class MaxwellSkyPanel(WorldButtonsPanel, Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_label = "Sky"

    @classmethod
    def poll(cls, context):
        return WorldButtonsPanel.poll.__func__(cls, context) and context.world.maxwell.type != 'none'

    def draw(self, context):
        layout = self.layout
        env = context.world.maxwell
        layout.row().prop(env, "sky", expand=True)
        if env.sky == 'physical':
            physical = env.physical
            layout.row().prop(physical, "intensity")
            layout.row().prop(physical, "planetRefl")
            layout.label("Gas:")
            row = layout.row(align=True)
            row.prop(physical, "ozone")
            row.prop(physical, "water")
            layout.label("Aerosol:")
            col = layout.column(align=True)
            col.prop(physical, "angstrom")
            col.prop(physical, "wavelength")
            row = layout.row()
            row.prop(physical, "albedo")
            row.prop(physical, "asymmetry")
        elif env.sky == 'constant':
            constant = env.constant
            layout.row().prop(constant, "luminance")
            row = layout.row()
            row.column().prop(constant, "zenith")
            row.column().prop(constant, "horizon")
            row = layout.row()
            # row.alignment = 'CENTER'
            row.prop(constant, "controlPoint")


class MaxwellSunPanel(WorldButtonsPanel, Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_label = "Sun"

    @classmethod
    def poll(cls, context):
        return WorldButtonsPanel.poll.__func__(cls, context) and context.world.maxwell.type != 'none'

    def draw(self, context):
        sun = context.world.maxwell.sun
        layout = self.layout
        row = layout.row()
        row.prop(sun, "type")
        enabled = (sun.type != '0')
        row = row.row()
        row.prop(sun, "power")
        row.enabled = enabled
        row = layout.row()
        row.prop(sun, "radius")
        row.enabled = enabled
        row = layout.row()
        row.prop(sun, "temperature")
        row.enabled = (sun.type == '1')
        row = layout.row()
        row.prop(sun, "color")
        row.enabled = (sun.type == '2')
        row = layout.row()
        row.prop(sun, "position", expand=True)
        row.enabled = enabled
        if sun.position == '0':
            col = layout.column(align=True)
            col.prop(sun, "latitude")
            col.prop(sun, "longitude")
            col.enabled = enabled
            split = layout.split(0.65)
            col = split.column()
            col.prop(sun, "ui_date")
            col.prop(sun, "ui_time")
            col = split.column()
            col.operator("maxwell.datetime_now")
            col.prop(sun, "sm")
            split.enabled = enabled
            row = layout.row()
            row.prop(sun, "rotation")
            row.enabled = enabled
        elif sun.position == '1':
            col = layout.column(align=True)
            col.prop(sun, "zenith")
            col.prop(sun, "azimuth")
            col.enabled = enabled
        elif sun.position == '2':
            split = layout.split(0.5)
            split.enabled = enabled
            col = split.column()
            col.label("Direction:")
            col.prop_search(sun, "dir_object", context.scene, "objects", text="", icon='LAMP_POINT')
            col = split.column()
            col.prop(sun, "direction", text="")
            col.enabled = enabled and not sun.dir_object


class MaxwellIblPanel(WorldButtonsPanel, Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_label = "Image Based Lighting"

    @classmethod
    def poll(cls, context):
        return WorldButtonsPanel.poll.__func__(cls, context) and context.world.maxwell.type == 'ibl'

    def draw(self, context):
        layout = self.layout
        ibl = context.world.maxwell.ibl
        layout.row().prop(ibl, "intensity")
        row = layout.row()
        row.prop(ibl, "interpolation")
        row.prop(ibl, "screen_mapping")


class MaxwellBackgroundIblPanel(WorldButtonsPanel, Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_label = "IBL Background"

    @classmethod
    def poll(cls, context):
        return WorldButtonsPanel.poll.__func__(cls, context) and context.world.maxwell.type == 'ibl'

    def draw(self, context):
        _draw_ibl_panel(self.layout, context.world.maxwell.ibl.background)


class MaxwellReflectionIblPanel(WorldButtonsPanel, Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_label = "IBL Reflection"

    @classmethod
    def poll(cls, context):
        return WorldButtonsPanel.poll.__func__(cls, context) and context.world.maxwell.type == 'ibl'

    def draw(self, context):
        _draw_ibl_panel(self.layout, context.world.maxwell.ibl.reflection)


class MaxwellRefractionIblPanel(WorldButtonsPanel, Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_label = "IBL Refraction"

    @classmethod
    def poll(cls, context):
        return WorldButtonsPanel.poll.__func__(cls, context) and context.world.maxwell.type == 'ibl'

    def draw(self, context):
        _draw_ibl_panel(self.layout, context.world.maxwell.ibl.refraction)


class MaxwellIlluminationIblPanel(WorldButtonsPanel, Panel):
    COMPAT_ENGINES = {MaxwellRenderEngine.bl_idname}
    bl_label = "IBL Illumination"

    @classmethod
    def poll(cls, context):
        return WorldButtonsPanel.poll.__func__(cls, context) and context.world.maxwell.type == 'ibl'

    def draw(self, context):
        _draw_ibl_panel(self.layout, context.world.maxwell.ibl.illumination)
